#include "SdbComboBoxDelegate.h"

#include <QComboBox>

namespace {
constexpr static const char* UNKNOWN_NAME = "-Unknown-";
typedef QPair<quint32, quint32> SdbAsPair;

std::optional<SdbAsPair> getSdbAsPair(QVariant d) {
    const auto &list = d.toList();
    if(list.size() != 2)
        return {};
    return SdbAsPair{list.first().toULongLong(), list.last().toULongLong()};
}

QString getComboBoxItemName(quint32 id, QString name)
{
    return QString("%1 0x%2").arg(name).arg(id, 8, 16, QChar('0'));
}

int findOrCreateItem(QComboBox *cb, QList<QVariant> data) {
    int index = cb->findData(data);
    if(index == -1) {
        auto sdbPair = getSdbAsPair(data);
        auto seq = sdbPair.value_or(SdbAsPair{}).second;
        QString name = UNKNOWN_NAME;
        if(seq)
            name += QString(" #%1").arg(seq);
        name = getComboBoxItemName(sdbPair.value_or(SdbAsPair{}).first, name);
        cb->addItem(name, data);
        index = cb->findData(data);
    }
    return index;
}
}

SdbComboBoxDelegate::SdbComboBoxDelegate(QObject *parent)
    : QItemDelegate(parent)
{
}

QWidget *SdbComboBoxDelegate::createEditor(QWidget *parent, const QStyleOptionViewItem &option, const QModelIndex &index) const
{
    auto cb = new QComboBox(parent);
    cb->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    auto val = index.model()->data(index, Qt::EditRole).toList();
    updateCbItems(cb, sdbMap, val);
    connect(cb, &QComboBox::currentTextChanged,
            this, &SdbComboBoxDelegate::dataChanged);

    return cb;

}

void SdbComboBoxDelegate::setEditorData(QWidget *editor, const QModelIndex &index) const
{
    auto data = index.model()->data(index, Qt::EditRole).toList();
    auto cb = static_cast<QComboBox*>(editor);
    if(cb->currentData() == data)
        return;
    QSignalBlocker block(cb);
    int i = findOrCreateItem(cb, data);
    cb->setCurrentIndex(i);
}

void SdbComboBoxDelegate::setModelData(QWidget *editor, QAbstractItemModel *model, const QModelIndex &index) const
{
    auto cb = static_cast<QComboBox*>(editor);
    auto value = cb->currentData();
    if(model->data(index, Qt::EditRole).toUInt() != value)
        model->setData(index, value, Qt::EditRole);
}

void SdbComboBoxDelegate::setSdbMap(const QMap<QString, QPair<quint32, quint32>> &_sdbMap)
{
    if(sdbMap == _sdbMap)
        return;
    sdbMap = _sdbMap;
}

void SdbComboBoxDelegate::updateCbItems(QComboBox *cb, const QMap<QString, QPair<quint32, quint32>> &map, QList<QVariant> curData)
{
    QMap<QString, QPair<quint32, quint32>> sdbMap = map;
    static const char* GLOBAL_KEY = "Global (without SDB)";
    sdbMap[GLOBAL_KEY] = {0, 0};
    auto curSdb = getSdbAsPair(curData);

    if(!curSdb && cb->currentData().isValid()) {
        curSdb = getSdbAsPair(cb->currentData());
    }
    auto cbCount = cb->count();

    if(curSdb) {
        bool missing = true;
        for(auto p : sdbMap) {
            if(p == *curSdb) {
                missing = false;
                break;
            }
        }
        if(missing) {
            QString name = UNKNOWN_NAME;
            if(curSdb->second)
                name += QString(" #%1").arg(curSdb->second);
            sdbMap[name] = *curSdb;
        }
    }
    QSignalBlocker block(cb);
    int i=0;

    auto keyList = sdbMap.keys();
    keyList.removeAll(GLOBAL_KEY);
    keyList.prepend(GLOBAL_KEY);
    for(auto name : keyList) {
        auto pair = sdbMap[name];
        QVariant data = QList<QVariant>{pair.first, pair.second};
        if(pair.first)
            name = getComboBoxItemName(pair.first, name);
        if(i>=cbCount) {
            cb->addItem(name, data);
        } else if(name!=cb->itemText(i) || data != cb->itemData(i)) {
            cb->setItemText(i, name);
            cb->setItemData(i, data);
        }
        ++i;
    }

    i = curSdb ? findOrCreateItem(cb, {curSdb->first, curSdb->second}) : 0;
    cb->setCurrentIndex(i);
}

void SdbComboBoxDelegate::dataChanged()
{
    commitData(qobject_cast<QWidget*>(sender()));

//    if(commitDataWidget) {
//        if(commitDataWidget == sender())
//            return;
//        flushCommitData();
//    }
//    commitDataWidget = sender();
//    commitDataTimer->start();
}

void SdbComboBoxDelegate::updateEditorGeometry(QWidget *editor, const QStyleOptionViewItem &option, const QModelIndex &index) const
{
    editor->setGeometry(option.rect);
}
