#include "FileWatcherCore.h"

#include <QJsonDocument>

#include "FileWatcherConfigDb.h"
#include "FileWatcherSignalRouter.h"
#include "FileReader.h"
#include "DataSender.h"
#include "FileWatcherStartupOptions.h"
#include "FileWatcherFsm.h"
#include "ConsoleReader.h"

#include "base/DeviceId.h"
#include "base/SoftwareDevices.h"
#include "base/RunDocument.h"

FileWatcherCore::FileWatcherCore(const FileWatcherStartupOptions &so, QObject *parent)
    : BaseAppCore(parent)
    , swDevIndex({DEVICE_ID_SOFTWARE, SERIAL_TO_CFG})
    , fsm(new FileWatcherFsm())
    , configDb(new FileWatcherConfigDb(so.path, this))
    , fileMonitor(new FileReader(this))
    , dataSender(new DataSender(swDevIndex))
    , consoleReader(new ConsoleReader(this))
{
    qRegisterMetaType<RunDocument>();

    auto &signalRouter = FileWatcherSignalRouter::instance();
    signalRouter.setConfigDb(configDb);
    signalRouter.setCore(this);
    signalRouter.setFsm(fsm);
    signalRouter.setFileMonitor(fileMonitor);
    signalRouter.setRemoteControlServer(remoteControlServer);
    signalRouter.setDataSender(dataSender);
    signalRouter.setConsoleReader(consoleReader);
    signalRouter.connectAll();

    initProgDescr();
}

FileWatcherCore::~FileWatcherCore()
{
    delete fsm;
}

void FileWatcherCore::setConfig(const FileWatcherConfig &newConfig)
{
    if(config == newConfig)
        return;

    config = newConfig;
    updateProgDescrOptions();
}

void FileWatcherCore::remoteStartRequest(const RcRunId &runId_, const QVariant &startEvNum)
{
    Q_UNUSED(startEvNum)
    if(fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
            .arg(fsmStateToString(FsmState::Ready))
            .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(runId_, msg);
        return;
    }

    rcRunId = runId_;
    fsm->setRequestState(FsmState::Run);
}

void FileWatcherCore::remoteStopRequest(const RcRunId &runId_, const QVariant &stopEvNum)
{
    Q_UNUSED(stopEvNum)
    if(rcRunId.runIndex != runId_.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); "
                             "remote control runIndex (%2)")
                     .arg(runId_.runIndex)
                     .arg(runId_.runIndex);
    }

    if(fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(runId_, msg);
    }

    fsm->setRequestState(FsmState::Ready);
}

void FileWatcherCore::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool changed = progDescr.updateProgramInterface(
                ProgramInterfaceRemoteControl, 0,
                status.tcpPort, status.listening, !status.portInUse,
                status.peerHost, status.peerPort);
    if(changed)
        emit updateProgramDescription(progDescr);

    if(status.portInUse) {
//        QString ready4run_str = "Ready for run";
//        for(DeviceIndex &index : devices.keys()){
//            if(!devices[index]->getTqdcStatus().connected){
//                ready4run_str = "Not ready for run";
//            }
//        }
        //        emit sendRemoteState(ready4run_str);
    } else {
        if(lastRcStatus.portInUse) {
            qWarning() << "Remote control disconnected.";

            if(fsm->getState() != FsmState::Ready) {
                qCritical() << "Stopping current run";
                fsm->setRequestState(fsm->getState() == FsmState::Run
                                 ? FsmState::Stopping : FsmState::Error);
            }
        }
    }
    lastRcStatus = status;
}

void FileWatcherCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    BaseAppCore::onFsmStateChange(state, oldState);
    emit enableFileMonitor(state==FsmState::Run);

    switch (state) {
    case FsmState::Error :
        break;
    case FsmState::Ready :
    {
        switch (oldState) {
        case FsmState::Run:
            qInfo() << QString("%1 completed successfully")
                       .arg(fsmStateToString(oldState));
            break;
        default:;
        }
        break;
    }
    case FsmState::Run :
    {
        qInfo() << QString("Start completed successfully");
        sendRunDocument();
        break;
    }
    default:;
    }
}

void FileWatcherCore::printStat() const
{
    if(lastRcStatus.listening) {
        qInfo() << QString("RemoteControlServer: listening on port #%1")
                   .arg(lastRcStatus.tcpPort);
    }
    qInfo() << "config.filePath:" << config.filePath;
    qInfo() << "state:" << fsmStateToString(fsm->getState());
    qInfo() << "rcRunId:" << rcRunId;
    qInfo() << "customBlocks:" << lastMStreamStat.customBlocks;
}

void FileWatcherCore::toggleStatPrint()
{
    periodicalStatPrint = !periodicalStatPrint;
    qInfo() << __PRETTY_FUNCTION__
            << "periodicalStatPrint:" << periodicalStatPrint;
}

void FileWatcherCore::setFilePath(const QString &p)
{
    auto st = fsm->getState();
    if(st!= FsmState::Error && st!= FsmState::Ready) {
        qWarning() << "Can't change file path; FSM state is " << fsmStateToString(st);
        return;
    }
    config.filePath = p;
    emit configChanged(config);
}

void FileWatcherCore::mStreamStatUpdated(const MStreamStat &ms_stat)
{
    bool changed = lastMStreamStat.customBlocks != ms_stat.customBlocks;
    lastMStreamStat = ms_stat;
    if(periodicalStatPrint && changed)
        qInfo() << __func__ << "customBlocks:"
                << lastMStreamStat.customBlocks;
}

bool FileWatcherCore::isRun() const
{
    Q_UNREACHABLE();
}

const BaseFsm *FileWatcherCore::getFsm() const
{
    return fsm;
}

void FileWatcherCore::initProgDescr()
{
    progDescr.setClientIndex(configDb->getClientIndex());

    //todo: transfer it from libs.pro
#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    emit updateProgramDescription(progDescr);
}

void FileWatcherCore::updateProgDescrOptions()
{
    progDescr.options["file"] = config.filePath;
    emit updateProgramDescription(progDescr);
}

void FileWatcherCore::sendRunDocument() const
{
    auto cfgObj = QVariant::fromValue(config).value<QJsonObject>();
    RunDocument rd(rcRunId, configDb->getClientIndex(), cfgObj);
    emit sendRunDocumentToRc(rd);
}

