#include "TrcDeviceChModel.h"

#include <QDebug>
#include <QTimer>

namespace {
quint16 reg_phy_rtig_ch_bit::*getFieldPtr(TrcTreeColumns col) {
    switch(col) {
    case TrcTreeColumns::COL_REDUC_FACTOR:
        return &reg_phy_rtig_ch_bit::reduction_factor;
    case TrcTreeColumns::COL_BEFORE_TIME:
        return &reg_phy_rtig_ch_bit::before_protect_time;
    case TrcTreeColumns::COL_AFTER_TIME:
        return &reg_phy_rtig_ch_bit::after_protect_time;
    default: return nullptr;
    }
}
quint32 PhysicalTriggerConfig::*getMaskPtr(TrcTreeColumns col) {
    switch(col) {
    case TrcTreeColumns::COL_START_MASK_EN:
        return &PhysicalTriggerConfig::startChMask;
    case TrcTreeColumns::COL_READ_MASK_EN:
        return &PhysicalTriggerConfig::readChMask;
    default: return nullptr;
    }
}
}

TrcDeviceChModel::TrcDeviceChModel(QObject *parent)
    : QAbstractTableModel(parent)
    , delayedSendConfigTimer(new QTimer(this))
{
    delayedSendConfigTimer->setInterval(0);
    delayedSendConfigTimer->setSingleShot(true);
    connect(delayedSendConfigTimer, &QTimer::timeout,
            this, &TrcDeviceChModel::postCheckStateChange);
}

QVariant TrcDeviceChModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    if(!config.has_value())
        return QVariant();

    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::ToolTipRole:
        return toolTipData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::CheckStateRole:
        return checkStateData(index);
    case Qt::UserRole+1:
        return itemMaxValue(index);
    case Qt::UserRole+5:
        return itemEditable(index);
    }
    return QVariant();
}

bool TrcDeviceChModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!index.isValid())
        return false;
    if(!config.has_value())
        return false;
    auto &cfg = config.value();

    switch (role) {
    case Qt::CheckStateRole:
    {
        Q_ASSERT(value!=Qt::PartiallyChecked);

        auto col = static_cast<TrcTreeColumns>(index.column());
        Q_ASSERT(cfg.getNCh() > index.row());
        if(cfg.getNCh() <= index.row())
            return false;
        switch(col) {
        case TrcTreeColumns::COL_REDUC_EN:
        case TrcTreeColumns::COL_BEFORE_EN:
        case TrcTreeColumns::COL_AFTER_EN:
        {
            auto &bitFields = cfg.chCfg[index.row()].reg.bit;
            quint8 enVal = value==Qt::Checked;
            switch (getColumnByIndex(index)) {
            case TrcTreeColumns::COL_REDUC_EN:
                bitFields.reduction_en = enVal; break;
            case TrcTreeColumns::COL_BEFORE_EN:
                bitFields.before_protect_en = enVal; break;
            case TrcTreeColumns::COL_AFTER_EN:
                bitFields.after_protect_en = enVal; break;
            default:;
            }
            break;
        }
        case TrcTreeColumns::COL_START_MASK_EN:
        case TrcTreeColumns::COL_READ_MASK_EN:
        {
            auto ptr = getMaskPtr(col);
            if(value == Qt::Checked) {
                cfg.*ptr |= 1<<index.row();
            } else {
                cfg.*ptr &= ~(1<<index.row());
            }
            break;
        }
        default:
            return false;
        }
        if(!delayedSendConfigTimer->isActive())
            delayedSendConfigTimer->start();
        return true;
    }
    case Qt::EditRole:
    {
        const auto &col = getColumnByIndex(index);
        Q_ASSERT(cfg.getNCh() > index.row());
        if(cfg.getNCh() <= index.row())
            return false;
        if(col == TrcTreeColumns::COL_NAME) {
            cfg.chCfg[index.row()].name = value.toString();
        } else {
            auto &cfgBits = cfg.chCfg[index.row()].reg.bit;
            quint16 reg_phy_rtig_ch_bit::*ptr = getFieldPtr(col);
            if(ptr) {
                cfgBits.*ptr = static_cast<quint16>(value.toUInt());
            } else if(col==TrcTreeColumns::COL_DELAY) {
                cfgBits.delay = static_cast<quint8>(value.toUInt());
            } else {
                return false;
            }
        }

        emit phyTrigCfgChanged(cfg);
        emit dataChanged(index, index, {Qt::DisplayRole});
        return true;
    }
    }

    return false;
}

void TrcDeviceChModel::setEnabled(bool enabled){
    this->enable = enabled;
    sendDataChanged();
}

void TrcDeviceChModel::setExpertsSettings(bool _showExpters)
{
    showExpters = _showExpters;
    sendDataChanged();
}


Qt::ItemFlags TrcDeviceChModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);
    auto col = static_cast<TrcTreeColumns>(index.column());
    if(col == TrcTreeColumns::COL_AFTER_EN
            || col == TrcTreeColumns::COL_BEFORE_EN
            || col == TrcTreeColumns::COL_REDUC_EN
            || col == TrcTreeColumns::COL_START_MASK_EN
            || col == TrcTreeColumns::COL_READ_MASK_EN)
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
    if(col == TrcTreeColumns::COL_NAME
            || col == TrcTreeColumns::COL_DELAY
            || col == TrcTreeColumns::COL_REDUC_FACTOR
            || col == TrcTreeColumns::COL_AFTER_TIME
            || col == TrcTreeColumns::COL_BEFORE_TIME)
        f |= Qt::ItemIsEditable;

    if(!enable)
        f = f & (~Qt::ItemIsEnabled);
    if(!showExpters) {
        if(col == TrcTreeColumns::COL_AFTER_EN
                || col == TrcTreeColumns::COL_BEFORE_EN
                || col == TrcTreeColumns::COL_AFTER_TIME
                || col == TrcTreeColumns::COL_BEFORE_TIME) {
            f = f & (~Qt::ItemIsEnabled);
        }
    }
    return f;
}

QVariant TrcDeviceChModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(role != Qt::DisplayRole)
        return QVariant();
    if(orientation == Qt::Vertical)
        return section+1;
    switch (static_cast<TrcTreeColumns>(section)) {
    case TrcTreeColumns::COL_NAME: return "Name";
    case TrcTreeColumns::COL_START_MASK_EN: return "StartBit";
    case TrcTreeColumns::COL_READ_MASK_EN: return "ReadBit";
    case TrcTreeColumns::COL_DELAY: return "Delay";
    case TrcTreeColumns::COL_REDUC_EN: return "Reduc";
    case TrcTreeColumns::COL_REDUC_FACTOR: return "Reduc factor";
    case TrcTreeColumns::COL_HIT_CNT: return "Hits";
    case TrcTreeColumns::COL_BEFORE_EN: return "BeforeEn";
    case TrcTreeColumns::COL_AFTER_EN: return "AfterEn";
    case TrcTreeColumns::COL_BEFORE_TIME: return "BeforeTime";
    case TrcTreeColumns::COL_AFTER_TIME: return "AfterTime";

    default: return QVariant();
    }
}

int TrcDeviceChModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return config.has_value() ? config.value().chCfg.size() : 0;
}

void TrcDeviceChModel::setPhyTrigCfg(const PhysicalTriggerConfig &newConfig)
{
    if(config == newConfig)
        return;
    bool performReset = !config.has_value() ||
            config.value().chCfg.size() != newConfig.chCfg.size();

    config = newConfig;
    if(performReset) {
        beginResetModel();
        endResetModel();
//        void beginInsertRows(const QModelIndex &parent, int first, int last);
//        void endInsertRows();
//        void beginRemoveRows(const QModelIndex &parent, int first, int last);
//        void endRemoveRows();
    } else {
        sendDataChanged();
    }
}

void TrcDeviceChModel::deviceStatusUpdated(const BaseDevCountersMap &st)
{
    rcCounters = st;

    auto col = static_cast<int>(TrcTreeColumns::COL_HIT_CNT);
    auto top = index(0, col);
    auto lastDev = index(rowCount(top)-1, col, top);
    auto lastCh = index(rowCount(lastDev)-1, col, lastDev);
    emit dataChanged(top, lastCh, {Qt::DisplayRole});
}

void TrcDeviceChModel::postCheckStateChange()
{
    sendDataChanged(QVector<int>{Qt::CheckStateRole});
    if(config.has_value())
        emit phyTrigCfgChanged(config.value());
}

void TrcDeviceChModel::sendDataChanged(QVector<int> roles)
{
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh, roles);
}

QVariant TrcDeviceChModel::displayData(const QModelIndex &index) const
{
    auto col = getColumnByIndex(index);
    switch (col) {
    case TrcTreeColumns::COL_NAME:
    {
        const auto &s = config.value().chCfg.value(index.row()).name;
        return s.isEmpty() ? QString("Ch #%1").arg(index.row()+1) : s;
    }
    case TrcTreeColumns::COL_DELAY:
    {
        const auto &val = editData(index);
        if(val.isValid())
            return val;
        break;
    }
    case TrcTreeColumns::COL_HIT_CNT:
    {
        if(!rcCounters.has_value())
            return "-";
        QString allStr;
        if(rcCounters->contains(DevCountersTrcIn)) {
            auto cnt = rcCounters->value(DevCountersTrcIn);
            allStr = QString("%L1").arg(TrcInputCounters::totalCnt(cnt, index.row()));
        } else
            allStr = "-";
        QString reducStr;
        if(rcCounters->contains(DevCountersTrcReduc)) {
            auto cnt = rcCounters->value(DevCountersTrcReduc);
            reducStr = QString("%L1").arg(TrcInputCounters::totalCnt(cnt, index.row()));
        } else
            reducStr = "-";
        return allStr+QChar(0x2192)+reducStr;
    }
    default:
    {
        auto ptr = getFieldPtr(col);
        if(ptr)
            return config.value().chCfg.value(index.row()).reg.bit.*ptr;
    }
    }
    return QVariant();
}

QVariant TrcDeviceChModel::toolTipData(const QModelIndex &index) const
{
    auto col = getColumnByIndex(index);
    switch (col) {
    case TrcTreeColumns::COL_HIT_CNT:
        return QString("<CountAll> ")+QChar(0x2192)+" <CountAfterReduction>";
    default:;
    }
    return QVariant();
}

QVariant TrcDeviceChModel::editData(const QModelIndex &index) const
{
    const auto &col = getColumnByIndex(index);
    if(col==TrcTreeColumns::COL_NAME)
        return config.value().chCfg.value(index.row()).name;
    const auto &cfgBits = config.value().chCfg.value(index.row()).reg.bit;
    const auto &ptr = getFieldPtr(col);
    if(ptr)
        return cfgBits.*ptr;
    if(col==TrcTreeColumns::COL_DELAY)
        return cfgBits.delay;
    return QVariant();
}

QVariant TrcDeviceChModel::itemMaxValue(const QModelIndex &index) const
{
    switch (getColumnByIndex(index)) {
    case TrcTreeColumns::COL_DELAY:
        return 0xFF;
    case TrcTreeColumns::COL_REDUC_FACTOR:
    case TrcTreeColumns::COL_BEFORE_TIME:
        return 0xFFFF;
    case TrcTreeColumns::COL_AFTER_TIME:
        return config.has_value() ? config.value().trigDelay : 0xFFFF;
    default:
        return QVariant();
    }
}

QVariant TrcDeviceChModel::itemEditable(const QModelIndex &index) const
{
    if(!index.isValid())
        return false;
    auto col = static_cast<TrcTreeColumns>(index.column());
    if(!showExpters) {
        if(col == TrcTreeColumns::COL_AFTER_TIME
                || col == TrcTreeColumns::COL_BEFORE_TIME)
            return false;
    }
    if(col == TrcTreeColumns::COL_DELAY
            || col == TrcTreeColumns::COL_REDUC_FACTOR
            || col == TrcTreeColumns::COL_AFTER_TIME
            || col == TrcTreeColumns::COL_BEFORE_TIME) {
        if(config.has_value()) {
            auto &cfg = config.value();
            if(index.row() < cfg.getNCh()) {
                auto bitFields = cfg.chCfg[index.row()].reg.bit;
                switch (getColumnByIndex(index)) {
                case TrcTreeColumns::COL_REDUC_FACTOR:
                    return bitFields.reduction_en;
                case TrcTreeColumns::COL_BEFORE_TIME:
                    return bitFields.before_protect_en;
                case TrcTreeColumns::COL_AFTER_TIME:
                    return bitFields.after_protect_en;
                default:
                    return true;
                }
            }
        }
    }

    return false;
}

QVariant TrcDeviceChModel::checkStateData(const QModelIndex &index) const
{
    switch (auto col = getColumnByIndex(index)) {
    case TrcTreeColumns::COL_REDUC_EN:
    case TrcTreeColumns::COL_BEFORE_EN:
    case TrcTreeColumns::COL_AFTER_EN:
    {
        auto bitFields = config.value().chCfg.value(index.row()).reg.bit;
        bool on;
        switch (getColumnByIndex(index)) {
        case TrcTreeColumns::COL_REDUC_EN: on = bitFields.reduction_en; break;
        case TrcTreeColumns::COL_BEFORE_EN: on = bitFields.before_protect_en; break;
        case TrcTreeColumns::COL_AFTER_EN: on = bitFields.after_protect_en; break;
        default:
            return QVariant();
        }
        return on ? Qt::Checked : Qt::Unchecked;
    }

    case TrcTreeColumns::COL_START_MASK_EN:
    case TrcTreeColumns::COL_READ_MASK_EN:
    {
        auto ptr = getMaskPtr(col);
        const quint32 mask = 1<<index.row();
        return (config.value().*ptr & mask) ? Qt::Checked : Qt::Unchecked;
    }
    default: return QVariant();
    }
}

TrcTreeColumns TrcDeviceChModel::getColumnByIndex(const QModelIndex &index) const
{
    return static_cast<TrcTreeColumns>(index.column());
}
