#include "TrcMainWindow.h"
#include "ui_TrcMainWindow.h"

#include "TrcConfigDispatcher.h"
#include "TrcCountersDialog.h"
#include "TrcDeviceChModel.h"
#include "TrcSignalRouter.h"
#include "TrcTreeColumns.h"
#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "m-stream-output/MStreamStatisticDialog.h"
#include "mldiscover/LldpDiscovery.h"
#include "widgets/DeviceStatusDialog.h"

#include <QBoxLayout>
#include <QFileDialog>
#include <QHostAddress>
#include <QLabel>
#include <QMessageBox>
#include <QMetaType>
#include <QPlainTextEdit>
#include <QScrollBar>
#include <QSemaphore>
#include <QSettings>
#include <QShortcut>
#include <QSpinBox>
#include <QTimer>

Q_DECLARE_METATYPE(QHostAddress)

namespace {
const int STOP_EVENT_TIMEOUT_MS = 10e3;
const int STATUS_UPDATE_DELAY = 1000;
enum {COL_DEV_SERIAL, COL_DEV_SLOT, COL_DEV_FW, COL_DEV_TEMP,
      COL_DEV_EV_NUM, COL_DEV_UNBLOCK, COL_DEV_BLOCK,
      COLUMN_WR_TIME, COLUMN_WR_TIME_OFFSET,
      COLUMN_WR_LINK_ERROR, COLUMN_WR_SYNC_LOST, COLUMN_WR_RX_CODE_ERROR,
      COL_DEV_MS_ERR, COL_DEV_STATUS, COL_DEV_STREAM, COL_DEV_TOTAL};

void setWidgetWarn(QWidget *w, const QStringList &warnList)
{
    bool en = w->isEnabled();
    auto c = en ? (warnList.isEmpty() ? QColor()         : QColor(0xFF8800)) :
                  (warnList.isEmpty() ? QColor(0xbfbfbf) : QColor(0xFFCC44));
    auto p = w->palette();

    auto b = p.brush(w->foregroundRole());
    if(p.color(w->foregroundRole()) != c) {
        p.setColor(w->foregroundRole(), c);
        w->setPalette(p);
    }

    QString tt = warnList.join('\n');
    if(w->toolTip() != tt)
        w->setToolTip(tt);
}
}

class GuardCtrl
{
public:
    GuardCtrl(int *_guardCnt):
    guardCnt(_guardCnt)
    {
        (*guardCnt)++;
    }

    ~GuardCtrl() {(*guardCnt)--;}

    int * const guardCnt;
};

TrcMainWindow::TrcMainWindow(QWidget *parent) :
      ReadoutAppMainWindow(parent)
    , ui(new Ui::TrcMainWindow)
    , trcModel(new TrcDeviceChModel(this))
    , status_update_timer(new QTimer(this))
//    , rcCountersModel(new RcCountersModel(this))
//    , rcView(new QTableView(this))
    , countersDialog(new TrcCountersDialog(this))
{
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);
//    rcView->setModel(rcCountersModel);

    status_update_timer->setSingleShot(true);
    status_update_timer->setInterval(STATUS_UPDATE_DELAY);
    connect(status_update_timer, &QTimer::timeout,
            this, &TrcMainWindow::deviceStatusUpdatedRequest);

    ui->trcTableView->setModel(trcModel);

    init_connections();
    init_table();

    configuration_manager->setDefaultConfig(BaseDeviceAppConfig());

    updateControlState();

    auto signalRouter = &TrcSignalRouter::instance();
    signalRouter->setMainWindow(this);
    signalRouter->setDiscoverDialog(discoverDialog);
    signalRouter->setConfigurationManager(configuration_manager);
    signalRouter->setMStreamStatisticDialog(mStreamStatisticDialog);
    signalRouter->setDeviceChModel(trcModel);
}

TrcMainWindow::~TrcMainWindow()
{
    delete ui;
}

DeviceStatusTableViewFancy* TrcMainWindow::getDeviceStatusTableView()
{
    return nullptr; // TODO
}

void TrcMainWindow::init_connections()
{
    connect(ui->actionPrint_FSM_inputs, &QAction::triggered,
            this, &TrcMainWindow::printFsmInputs);
    connect(countersDialog, &TrcCountersDialog::clearStat,
            this, &TrcMainWindow::softClear);

    auto cfgDisp = &TrcConfigDispatcher::instance();
    cfgDisp->subscribeSpillConfig(ui->widgetSpill);
    cfgDisp->subscribeRcTrigConfig(ui->widgetTrig);
    cfgDisp->subscribePhysicalTriggerConfig(trcModel);
}

void TrcMainWindow::update_mstream_data_rate(const MStreamStat &ms_stat)
{
    int row = getDevRow(ms_stat.index);
    if(row == -1) return;

    double receive_data_rate=0;

    auto prev_stat_it = prev_ms_stat.find(ms_stat.index);
    if(prev_stat_it == prev_ms_stat.end()) {
        prev_ms_stat.insert(ms_stat.index, ms_stat);
    } else {
        qint64 milisecs = 0;
        milisecs = ms_stat.acquiredAtMs - prev_stat_it->acquiredAtMs;
        if(milisecs != 0) {
            receive_data_rate = 1e3*
                    static_cast<double>(ms_stat.recvDataSize - prev_stat_it->recvDataSize)/
                    (static_cast<double>(milisecs)*1e6);
        }
        *prev_stat_it = ms_stat;
    }

    ui->tableWidgetDevice->item(row, COL_DEV_STREAM)
            ->setText(QString("%L1").arg(receive_data_rate));
}

void TrcMainWindow::saveAppGuiSettings(QSettings &s) const
{
    BaseMainWindow::saveAppGuiSettings(s);
    s.beginGroup("MW");
    s.setValue("showExperts", ui->actionShow_experts_settings->isChecked());
    s.endGroup();
}

void TrcMainWindow::loadAppGuiSettings(const QSettings &s)
{
    BaseMainWindow::loadAppGuiSettings(s);
    bool showVar = s.value("MW/showExperts").toBool();
    ui->actionShow_experts_settings->setChecked(showVar);
    on_actionShow_experts_settings_toggled(showVar);
}

void TrcMainWindow::updateControlState()
{
    bool inStateErrReady = curFsmState == FsmState::Ready || curFsmState == FsmState::Error;

    ui->pushButtonStart->setEnabled(curFsmState == FsmState::Ready && !underRemoteCtrl);
    ui->pushButtonStop->setEnabled(curFsmState == FsmState::Run && !underRemoteCtrl);

    controlStateEnable((inStateErrReady || Globals::getDebugMode())
                       && curFsmState != FsmState::Calibrate);

    ui->actionReset_MStream->setEnabled(inStateErrReady);
    ui->actionSend_soft_reset->setEnabled(inStateErrReady);
}

void TrcMainWindow::deviceStatusUpdated(const TrcStatus &status)
{
    dev_status[status.getDeviceIndex()] = status;
//    rcCountersModel->deviceStatusUpdated(status.rcCounters);
    deviceStatusUpdatedRequest();
}

DeviceIndex TrcMainWindow::getIndexByRow(int row)
{
    auto item = ui->tableWidgetDevice->item(row, COL_DEV_SERIAL);
    if(!item)
        return DeviceIndex();
    const auto &data = item->data(Qt::UserRole);
    if(data.isNull())
        return DeviceIndex();
    return data.value<DeviceIndex>();
}

int TrcMainWindow::getDevRow(const DeviceIndex &index)
{
    for(int row=ui->tableWidgetDevice->rowCount()-1; row>=0; --row){
        if(getIndexByRow(row) == index){
            return row;
        }
    }
    return -1;
}

void TrcMainWindow::temperatureSet(DeviceIndex index, QMap<QString, double> temps)
{
    const int row = getDevRow(index);
    if(row == -1 || temps.isEmpty()) return;

    QList<double> tempList = temps.values();
    double minTemp = *std::min_element(tempList.begin(), tempList.end());
    double maxTemp = *std::max_element(tempList.begin(), tempList.end());
    QColor col = (maxTemp>=70 || minTemp<0) ? Qt::red : Qt::black;
    auto tempStr = QString("%1").arg(minTemp,0,'f',1);
    if(!qFuzzyCompare(minTemp, maxTemp))
        tempStr += QString("-%2").arg(maxTemp,0,'f',1);
    tempStr += QChar(0x02da);
    QStringList fullInfo;
    for (auto &tempId : temps.keys()) {
        fullInfo << QString("%1%2 - %3").arg(temps[tempId],0,'f',1)
                        .arg(QChar(0x02da)).arg(tempId);
    }
    QTableWidgetItem * item = ui->tableWidgetDevice->item(row, COL_DEV_TEMP);
    item->setText(tempStr);
    auto fg = item->foreground();
    fg.setColor(col);
    item->setForeground(fg);
    item->setToolTip(fullInfo.join('\n'));
    updateTableSize(ui->tableWidgetDevice);
}

void TrcMainWindow::setTableDevRowCnt(int cnt)
{
    auto t = ui->tableWidgetDevice;
    int oldRowCnt = t->rowCount();
    t->setRowCount(cnt);
    for(int row=oldRowCnt; row<cnt; ++row) {
        for(int col=0; col<COL_DEV_TOTAL; ++col) {
            t->setItem(row, col, new QTableWidgetItem());
            if(col == COL_DEV_MS_ERR) {
                auto item = t->item(row, col);
                auto fg = item->foreground();
                fg.setColor(Qt::red);
                item->setForeground(fg);
            }
        }
    }
}

void TrcMainWindow::deviceStatusUpdatedRequest()
{
    if(status_update_timer == QObject::sender()) {
        if(status_request_flag){
            deviceStatusUpdatedResponce();
            status_request_flag = false;
        }
    }
    else{
        if(status_update_timer->remainingTime() > 0){
            status_request_flag = true;
        }
        else{
            deviceStatusUpdatedResponce();
            status_request_flag = false;
            status_update_timer->start();
        }
    }
}

void TrcMainWindow::deviceStatusUpdatedResponce()
{
    deviceStatusDialog->setMultiStatus(statusMapToBase(dev_status));
    for(auto &s : dev_status.keys()){
        const auto &status = dev_status[s];
        const auto &deviceIndex = status.getDeviceIndex();
        if(!config.knownConfigs.contains(deviceIndex) ||
                !config.knownConfigs.value(deviceIndex).enabled) {
            dev_status.remove(s);
            continue;
        }

        const int row = getDevRow(deviceIndex);
        if(row == -1) continue;

        if(status.connected) {
            temperatureSet(deviceIndex, status.getTemps());
            ui->tableWidgetDevice->item(row, COL_DEV_FW)->setText(status.getFwStr());
            updateDevWrStatus(row, status.wrStatus);
            QString gen_str;
            if (status.globEvNum.has_value())
                gen_str = QString("%L1").arg(status.globEvNum.value());
            ui->tableWidgetDevice->item(row, COL_DEV_EV_NUM)->
                setText(gen_str);
            ui->tableWidgetDevice->item(row, COL_DEV_UNBLOCK)->
                setText(QString("%L1").arg(status.statistic.evNumAccepted));

            QString str;
            if(str.isEmpty()) {
                str = "Ok";
            }

            ui->widgetSpill->setSpillStatus(status.phyTrig.spill);
            ui->tableWidgetDevice->item(row, COL_DEV_STATUS)->setText(str);
        } else {
            auto itemStatus = ui->tableWidgetDevice->item(row, COL_DEV_STATUS);
            itemStatus->setText("Offline");
            auto fg = itemStatus->foreground();
            fg.setColor(Qt::black);
            itemStatus->setForeground(fg);
        }
    }
    updateTableSize(ui->tableWidgetDevice);
}

bool TrcMainWindow::isControlBlocked() const
{
    return blockControl || underRemoteCtrl;
}

void TrcMainWindow::deviceConnected(const DeviceIndex &index)
{
    auto row = getDevRow(index);
    if(row == -1)
        return;
    ui->tableWidgetDevice->item(row, COL_DEV_SERIAL)->setTextColor(Qt::darkGreen);
    updateTableSize(ui->tableWidgetDevice);
}

void TrcMainWindow::deviceDisconnected(const DeviceIndex &index)
{
    auto row = getDevRow(index);
    if(row == -1)
        return;
    ui->tableWidgetDevice->item(row, COL_DEV_SERIAL)->setTextColor(Qt::red);
}

void TrcMainWindow::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if(config == newConfig)
        return;

    auto curDevSet = config.getDeviceIndexEnabledMap().getEnabled();
    auto newDevSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    ReadoutAppMainWindow::setConfig(config_key, newConfig);

    updateGuiVal();
    updateControlState();
    updateTableSize(ui->tableWidgetDevice);

    //add devices
    for(DeviceIndex index : newDevSet-curDevSet) {
        addDevice(index);
    }

    //remove from devices
    for(DeviceIndex index : curDevSet-newDevSet) {
        disconnectFromDev(index);
    }
}

void TrcMainWindow::lldpCacheUpdated(const LldpInfoCache &cache)
{
    lldpCache = cache;

    for(const auto &lldp : lldpCache.values()) {
        const int row = getDevRow(lldp.getDeviceIndex());
        if(row == -1)
            continue;

        QString txt = lldp.isNull() ? "-" :
                                      QString("%1").arg(lldp.ext.chassis_slot, 2, 10, QChar('0'));

        auto itemSlot = ui->tableWidgetDevice->item(row, COL_DEV_SLOT);
        if(itemSlot && itemSlot->text() != txt) {
            itemSlot->setText(txt);
        }
    }
}

void TrcMainWindow::init_table()
{
    ++blockControl;

    QTableWidget * const tableDev = ui->tableWidgetDevice;
    tableDev->verticalHeader()->hide();
    tableDev->setColumnCount(COL_DEV_TOTAL);
    tableDev->setHorizontalHeaderItem(COL_DEV_SERIAL, new QTableWidgetItem("Serial"));
    tableDev->setHorizontalHeaderItem(COL_DEV_SLOT, new QTableWidgetItem("Slot"));
    tableDev->setHorizontalHeaderItem(COL_DEV_FW, new QTableWidgetItem("FPGA f/w"));
    tableDev->setHorizontalHeaderItem(COL_DEV_TEMP, new QTableWidgetItem("temp"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_TIME, new QTableWidgetItem("WR Time"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_TIME_OFFSET, new QTableWidgetItem("Offset"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_LINK_ERROR, new QTableWidgetItem("Link"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_SYNC_LOST, new QTableWidgetItem("Unlock"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_RX_CODE_ERROR, new QTableWidgetItem("RX Err"));
    tableDev->setHorizontalHeaderItem(COL_DEV_EV_NUM, new QTableWidgetItem("Ev. num"));
    tableDev->setHorizontalHeaderItem(COL_DEV_UNBLOCK, new QTableWidgetItem("Acepted"));
    tableDev->setHorizontalHeaderItem(COL_DEV_BLOCK, new QTableWidgetItem("Blocked"));
    tableDev->setHorizontalHeaderItem(COL_DEV_MS_ERR, new QTableWidgetItem("MStream err"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STATUS, new QTableWidgetItem("Status"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STREAM, new QTableWidgetItem("Data Stream, MB/s"));
    tableDev->hideColumn(COLUMN_WR_TIME_OFFSET);
    tableDev->hideColumn(COL_DEV_MS_ERR);
    tableDev->hideColumn(COL_DEV_UNBLOCK);
    tableDev->hideColumn(COL_DEV_BLOCK);


    setTableDevRowCnt(1);
    tableDev->item(0, COL_DEV_SERIAL)->setText("");
    tableDev->item(0, COL_DEV_STATUS)->setText("No device selected");

    tableDev->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    tableDev->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    updateTableSize(tableDev);

    --blockControl;
}

void TrcMainWindow::copyAndSetupDevice()
{
    config.unify_config();
    setupDevice();
}

void TrcMainWindow::setupDevice()
{
    if(blockControl) return;
    emit configChanged(config);
}

void TrcMainWindow::updateGuiVal()
{
    ++blockControl;
    const auto &ds = config.defaultSetup;

    const auto &phyTrig = ds.trcCore.phyTrig;
    ui->spinBoxMatchWin->setValue(phyTrig.matchWin);
    ui->spinBoxMatchWin->setMaximum(phyTrig.trigDelay);
    ui->spinBoxTrigDelay->setValue(phyTrig.trigDelay);

    ui->groupBoxChWaveform->setChecked(phyTrig.oscEn);
    ui->radioButtonOscDelayed->setChecked(phyTrig.oscDelayed);
    ui->spinBoxLatency->setValue(phyTrig.chWfLatency);
    ui->spinBoxReadPoints->setValue(phyTrig.chWfReadPoints);

    ui->groupBoxCalibTrigger->setConfig(phyTrig.calibTrigger);

    --blockControl;
    updateControlState();
}

void TrcMainWindow::controlStateEnable(bool enable)
{
    qWarning() << "Temporary block:" << __PRETTY_FUNCTION__;
//    ui->groupBoxAdc->setEnabled(enable);
//    ui->groupBoxTtl->setEnabled(enable);
//    trcModel->setEnabled(enable);
}

void TrcMainWindow::disconnectFromDev(const DeviceIndex &index)
{
    dev_status.remove(index);
    const int row = getDevRow(index);
    auto t = ui->tableWidgetDevice;
    if(row==0 && t->rowCount() == 1){
        for(int i=0; i<COL_DEV_TOTAL; ++i){
            auto serial_item = t->item(row, i);
            serial_item->setText("");
            serial_item->setData(Qt::UserRole, QVariant());
        }
        t->item(row, COL_DEV_STATUS)->setText("No device selected");
    } else {
        t->removeRow(row);
    }
    updateControlState();
}

void TrcMainWindow::addDevice(const DeviceIndex &index)
{
    auto t = ui->tableWidgetDevice;
    int row = getIndexByRow(0).isNull() ? 0 : t->rowCount();
    setTableDevRowCnt(row+1);

    t->item(row, COL_DEV_SERIAL)->setText(index.getSerialStr());
    t->item(row, COL_DEV_SERIAL)->setData(Qt::UserRole, QVariant::fromValue(index));
    t->item(row, COL_DEV_STATUS)->setText("connecting...");
    const auto &lldpInfo = lldpCache.value(index);
    QString slot_str = lldpInfo.isNull() ? "-"
                                         : QString::number(lldpInfo.ext.chassis_slot);
    t->item(row, COL_DEV_SLOT)->setText(slot_str);
}

void TrcMainWindow::updateTableSize(QTableWidget * const table, bool widthOnly)
{
    table->resizeRowsToContents();
    table->resizeColumnsToContents();

    int width = table->verticalHeader()->width()+5;
    for(int col=0; col<table->columnCount(); ++col) {
        if(table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
    }
    if(width<20)
        width = 20;

    int height = table->horizontalHeader()->height()+5;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if(height<20)
        height = 20;
    if(widthOnly) {
        width += table->verticalScrollBar()->width();
        table->setMinimumWidth(width);
    } else
        table->setFixedSize(width, height);
    table->updateGeometry();
}

// duplicates adc64 DeviceTableWidget::updateWrStatus
void TrcMainWindow::updateDevWrStatus(int row, const std::optional<WrStatus> &wrStatus)
{
    auto table = ui->tableWidgetDevice;

    if (!wrStatus.has_value()) {
        auto cols = {
            COLUMN_WR_LINK_ERROR,
            COLUMN_WR_SYNC_LOST,
            COLUMN_WR_TIME,
            COLUMN_WR_RX_CODE_ERROR,
            COLUMN_WR_TIME_OFFSET
        };
        for (auto i : cols) {
            table->item(row, i)->setText({});
            table->item(row, i)->setBackground(bgColorDefault());
        }
        return;
    }

    table->item(row, COLUMN_WR_LINK_ERROR)->setText(QString::number(wrStatus->linkError));
    table->item(row, COLUMN_WR_SYNC_LOST)->setText(QString::number(wrStatus->syncFail));

    bool wrTimeValid = wrStatus->timeValid();
    const auto &wrTime = QDateTime::fromMSecsSinceEpoch(wrStatus->mSecsSinceEpoch());
    int wrTimeCheck = wrStatus->wrTimeCheck();
    double wrTimeOffsetMs = wrStatus->mSecsSinceEpoch() - wrStatus->acquiredAtMs;
    QColor offsetBgColor = bgColorDefault();
    if (wrTimeCheck == 1)
        offsetBgColor = bgColorWarning();
    if (wrTimeCheck > 1)
        offsetBgColor = bgColorError();

    table->item(row, COLUMN_WR_TIME)->setText(wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss")); // add ".zzz" for ms
    table->item(row, COLUMN_WR_TIME)->setBackground(wrTimeValid ? bgColorDefault() : bgColorError());
    QString strOffset;
    if (wrTimeOffsetMs/1000. > -60 && wrTimeOffsetMs/1000. < 60)
        strOffset = QString(" %1").arg(wrTimeOffsetMs * 1e-3, 6, 'f', 3);
    table->item(row, COLUMN_WR_TIME_OFFSET)->setText(strOffset);
    table->item(row, COLUMN_WR_TIME_OFFSET)->setBackground(offsetBgColor);

    table->item(row, COLUMN_WR_RX_CODE_ERROR)->setText(QString::number(wrStatus->codeError));
    // table->item(row, COLUMN_UPTIME)->setText(QString::number(wrStatus->uptime));
}

// ==== Slots from UI

void TrcMainWindow::on_actionShow_experts_settings_toggled(bool showExperts)
{
    ui->trcTableView->setColumnHidden(static_cast<int>(TrcTreeColumns::COL_START_MASK_EN), !showExperts);
    ui->scrollAreaWidgetContents->setEnabled(showExperts);
    trcModel->setExpertsSettings(showExperts);
}

void TrcMainWindow::on_pushButtonStart_clicked()
{
    if(isControlBlocked()) return;
    if(!config.hasEnabledDevices()) {
        qWarning() << "There is no enabled device";
        return;
    }

    emit performLocalStart();
}

void TrcMainWindow::on_pushButtonStop_clicked()
{
    if(isControlBlocked()) return;
    emit performLocalStop();
}

void TrcMainWindow::on_actionReset_MStream_triggered()
{
    emit mstream_reset_all(true);
}

void TrcMainWindow::on_actionSend_soft_reset_triggered()
{
    auto ret = QMessageBox::warning(this, "Confirm device soft reset...",
                                    "Sending soft reset to device entails "
                                    "resetting device to factory settings "
                                    "followed by writting to current configuration",
                                    {QMessageBox::Yes, QMessageBox::No}, QMessageBox::No);
    if(ret == QMessageBox::Yes)
        emit softReset();
}

void TrcMainWindow::on_actionSend_soft_clear_triggered()
{
    emit softClear();
}

void TrcMainWindow::on_spinBoxMatchWin_valueChanged(int mw)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.matchWin = static_cast<quint16>(mw);
    config.unify_config();
    emit configChanged(config);
}


void TrcMainWindow::on_spinBoxTrigDelay_valueChanged(int delay)
{
    ui->spinBoxMatchWin->setMaximum(delay);
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.trigDelay = static_cast<quint16>(delay);
    config.unify_config();
    emit configChanged(config);
}

void TrcMainWindow::on_radioButtonOscDelayed_toggled(bool checked)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.oscDelayed = checked;
    config.unify_config();
    emit configChanged(config);
}


void TrcMainWindow::on_groupBoxChWaveform_toggled(bool checked)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.oscEn = checked;
    config.unify_config();
    emit configChanged(config);
}

void TrcMainWindow::on_spinBoxLatency_valueChanged(int lat)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.chWfLatency = static_cast<quint16>(lat);
    config.unify_config();
    emit configChanged(config);
}

void TrcMainWindow::on_spinBoxReadPoints_valueChanged(int points)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.chWfReadPoints = static_cast<quint16>(points);
    config.unify_config();
    emit configChanged(config);
}

void TrcMainWindow::on_groupBoxCalibTrigger_cfgUpdated(const CalibrationTriggerConfig &c)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.calibTrigger = c;
    config.defaultSetup.trcCore.phyTrig.unifyCalibTrig();
    config.unify_config();
    emit configChanged(config);
}

void TrcMainWindow::on_actionCounters_dialog_triggered()
{
    countersDialog->show();
}

// FIXME: duplicated code

DeviceIndex TrcMainWindow::getDevIndexByRow(int row) const
{
    const auto &index_data = ui->tableWidgetDevice->item(row, COL_DEV_SERIAL)
                                 ->data(Qt::UserRole);
    return index_data.value<DeviceIndex>();
}

void TrcMainWindow::on_tableWidgetDevice_currentItemChanged(QTableWidgetItem *current, QTableWidgetItem *previous)
{
    if (!current)
        return;
    auto deviceIndex = getDevIndexByRow(current->row());
    deviceStatusDialog->setIndex(deviceIndex);
}

// duplicates DeviceTableWidget
QColor TrcMainWindow::bgColorDefault() const
{
    return style()->standardPalette().light().color();
}

QColor TrcMainWindow::bgColorWarning() const
{
    return Qt::yellow;
}

QColor TrcMainWindow::bgColorError() const
{
    return Qt::red;
}
