/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceIndex.h"

#include <QDataStream>
#include <QDebug>

#include "base/DeviceId.h"

bool operator <(const DeviceIndex &l, const DeviceIndex &r)
{
    if (l.device_id == r.device_id)
        return (l.serial_id < r.serial_id);
    return (l.device_id < r.device_id);
}

QString DeviceIndex::getIdent() const
{
    return QString("%1 %2")
        .arg(getDeviceIdName())
        .arg(getSerialStr());
}

QString DeviceIndex::getDeviceIdName() const
{
    return getDeviceIdName(this->device_id);
}

QString DeviceIndex::getSerialStr() const
{
    return getSerialStr(this->serial_id);
}

QString DeviceIndex::getDeviceIdName(DeviceId device_id)
{
    const auto s = QString(getDeviceTypeName(device_id));
    if (s.isEmpty()) {
        return "0x" +QString("%1").arg(device_id, 2, 16, QChar('0')).toUpper();
    }
    return s;
}

QString DeviceIndex::getSerialStr(quint64 serial_id)
{
    QString serial = QString("%1-%2")
                         .arg((serial_id>>16) & 0xFFFFu, 4, 16, QChar('0'))
                         .arg(serial_id & 0xFFFFu, 4, 16, QChar('0'));
    if(serial_id>>32) {
        serial.prepend(QString("%1-")
                       .arg((serial_id>>32) & 0xFFFFu, 4, 16, QChar('0')));
    }
    if(serial_id>>48) {
        serial.prepend(QString("%1-")
                       .arg((serial_id>>48) & 0xFFFFu, 4, 16, QChar('0')));
    }
    return serial.toUpper();
}

QString DeviceIndex::toString() const
{
    return
        "0x" +
        QString("%1").arg(device_id, 2, 16, QChar('0')).toUpper() +
        ":0x" +
        QString("%1").arg(serial_id, 12, 16, QChar('0')).toUpper();
}

DeviceIndex DeviceIndex::fromString(const QString &r)
{
    const auto list = r.toLower().split(":");
    if (list.size() < 2)
        return DeviceIndex();
    bool ok;
    auto id = list[0].toUInt(&ok, 0);
    Q_ASSERT(ok);
    if (!ok)
        return DeviceIndex();
    auto serial = list[1].toULongLong(&ok, 0);
    Q_ASSERT(ok);
    if (!ok)
        return DeviceIndex();

    return DeviceIndex(static_cast<DeviceId>(id), serial);
}

QDebug operator<<(QDebug dbg, const DeviceIndex &r)
{
    dbg.noquote().nospace() << QString("DeviceIndex(%1, %2)")
                               .arg(r.getDeviceIdName())
                               .arg(r.getSerialStr());
    return dbg;
}

QDataStream &operator<<(QDataStream &out, const DeviceIndex &value)
{
    out << value.device_id  << value.serial_id;
    return out;
}

QDataStream &operator>>(QDataStream &in, DeviceIndex &value)
{
    quint32 raw_id;
    in >> raw_id;
    quint64 serial_id;
    in >> serial_id;
    if (in.status() == QDataStream::Ok) {
        value.device_id = static_cast<DeviceId>(raw_id);
        value.serial_id = serial_id;
    }
    return in;
}

DeviceIndexSet makeDeviceIndexSet(const QList<DeviceIndex> &list)
{
#if (QT_VERSION >= QT_VERSION_CHECK(5, 14, 0))
    return DeviceIndexSet(list.constBegin(), list.constEnd());
#else
    DeviceIndexSet set;
    for (const auto &i: list)
        set << i;
    return set;
#endif
}

DeviceIdSet makeDeviceIdSet(const QList<DeviceId> &list)
{
#if (QT_VERSION >= QT_VERSION_CHECK(5, 14, 0))
    return DeviceIdSet(list.constBegin(), list.constEnd());
#else
    DeviceIdSet set;
    for (const auto &i: list)
        set << i;
    return set;
#endif
}
