/*
**    Copyright 2021 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ServiceConfigUrl.h"

#include <QDebug>
#include <QRegularExpression>
#include <QSettings>
#include "base/Globals.h"

QUrl ServiceConfigUrl::getUrl(QString environment, QString serviceName, QUrl defaultUrl)
{
    const auto &settings = Globals::getSettings();

    QString key = environment + "/" + serviceName;
    if (settings->contains(key)) {
        QString url_str = settings->value(key).toString();
        QUrl u = QUrl::fromUserInput(url_str);
        if (!u.isValid()) {
            qWarning() << QString("%1 service: invalid url '%2'")
                          .arg(serviceName).arg(url_str);
            return defaultUrl;
        }
        if (!url_str.contains(QRegularExpression("^[^:/?#]+://")))
            u.setScheme(defaultUrl.scheme());
        if (u.port() == -1 && defaultUrl.port() != -1)
            u.setPort(defaultUrl.port());
        return u;
    } else {
        QUrl url(defaultUrl);
        const auto &host = settings->value(key + "_" + "host");
        if (!host.isNull()){
            url.setHost(host.toString());
        }
        const auto &port = settings->value(key + "_" + "port");
        if (!port.isNull())
            url.setPort(port.toUInt());
        const auto &username = settings->value(key + "_" + "username");
        if (!username.isNull())
            url.setUserName(username.toString());
        const auto &password = settings->value(key + "_" + "password");
        if (!password.isNull())
            url.setPassword(password.toString());
        return url;
    }

    return defaultUrl;
}

QUrl ServiceConfigUrl::getUrl() const
{
    return getUrl(Globals::getEnvironment(), serviceName, defaultUrl);
}

bool ServiceConfigUrl::isEnabled(QString serviceName, bool defaultEnable)
{
    const auto &settings = Globals::getSettings();
    QString key = Globals::getEnvironment() + "/" + serviceName + "_enable";
    if (settings->contains(key))
        return settings->value(key).toBool();
    return defaultEnable;
}

bool ServiceConfigUrl::isEnabled() const
{
    return isEnabled(serviceName, defaultEnable);
}
