/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ThreadController.h"

#include <QDebug>
#include <QLoggingCategory>
#include <QThread>

#include "util.h"

namespace {

Q_DECLARE_LOGGING_CATEGORY(threads)
Q_LOGGING_CATEGORY(threads, "afi.threads")

enum { USE_THREADS = 1 };

} // namespace

ThreadController::ThreadController(QObject *obj, QObject *parent)
    : ThreadController(obj, QThread::LowestPriority, parent)
{}

ThreadController::ThreadController(QObject *obj, QThread::Priority priority, QObject *parent)
    : QObject(parent),
      obj(obj)
{
    Q_ASSERT(obj);
    if (!obj)
        return;
    formerAncestry = qObjectAncestryStr(obj);
    if (USE_THREADS) {
        objThread = new QThread(this);
        QString s = obj->objectName().isEmpty() ? obj->metaObject()->className() : obj->objectName();
        objThread->setObjectName(s);
        connect(objThread, &QThread::finished, obj,
                &QObject::deleteLater, Qt::DirectConnection); // deleteLater is thread-safe
        connect(obj, &QObject::destroyed,
                objThread, &QThread::quit);
        objThread->start(priority);
        obj->setParent(nullptr);
        obj->moveToThread(objThread);
        qDebug(threads) << QString("Thread started: %1").arg(formerAncestry);
    } else {
        obj->setParent(this);
    }
}

ThreadController::~ThreadController()
{
    if (!objThread || objThread == thread()) {
        delete obj;
        return;
    }
    objThread->quit();

    const quint64 sleepMs = 10;
    const qint64 timeoutMs = 20000;
    for (quint64 i=0; i < timeoutMs / sleepMs; i++) {
        if (objThread->isFinished())
            break;
        thread()->msleep(sleepMs);
    }
    if (!objThread->isFinished()) {
        qCritical(threads) << QString("Thread is still running: %1").arg(formerAncestry);
        objThread->terminate();
        objThread->wait();
        qCritical(threads) << QString("Thread terminated: %1").arg(formerAncestry);
    } else {
        qDebug(threads) << QString("Thread finished: %1").arg(formerAncestry);
    }
}
