/*
**    Merge two table models by rows
**
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MERGECOLUMNSTABLEMODEL_H
#define MERGECOLUMNSTABLEMODEL_H

#include <QAbstractTableModel>


class MergeColumnsTableModel : public QAbstractTableModel
{
    Q_OBJECT
    struct MergeModelInfo {
        MergeModelInfo(QAbstractItemModel *model, int mergeRole)
            : model(model), mergeRole(mergeRole) {}

        QVariant getKeyByRow(int row) const;
        int getRowByKey(const QVariant &key) const;

        QAbstractItemModel *model = nullptr;
        int mergeRole = Qt::DisplayRole;
    };

public:
    struct ModelWithColumn {
        bool isValid() const { return model && col >= 0; }
        QAbstractItemModel *model = nullptr;
        int col = -1;
    };

    MergeColumnsTableModel(QObject *parent = nullptr);

    template <class T>
    int getColumnFromSource(int col) const
    {
        int colOffset = 0;

        for (auto &mi : modelList) {
            if (qobject_cast<T*>(mi.model) != nullptr)
                return col + colOffset;
            colOffset += mi.model->columnCount();
        }

        return -1;
    }
    int getColumnFromSource(QAbstractItemModel *model, int col) const;
    ModelWithColumn mapColumnToSource(int column) const;

    QModelIndex mapFromSource(const QModelIndex &source) const;
    QModelIndex mapToSource(const QModelIndex & proxy) const;

    virtual int rowCount(const QModelIndex &parent = QModelIndex()) const override;
    virtual int columnCount(const QModelIndex &parent = QModelIndex()) const override;

    virtual Qt::ItemFlags flags(const QModelIndex & index) const override;
    virtual QVariant data(const QModelIndex &index, int role) const override;
    virtual QVariant headerData(int section, Qt::Orientation orientation,
                        int role = Qt::DisplayRole) const override;
    bool setData(const QModelIndex& index, const QVariant& value, int role = Qt::EditRole) override;

    void addSourceModel(QAbstractItemModel *model, Qt::ItemDataRole mergeRole);
    void setPrimarySourceModel(QAbstractItemModel *model);
    QList<QAbstractItemModel *> sourceModels() const;

signals:
    void dataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight,
                       const QVector<int> &roles = QVector<int>());
    void modelReset();

private slots:
    void onDataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight,
                       const QVector<int> &roles = QVector<int>());
    void onRowsInserted(const QModelIndex &parent, int first, int last);
//    void onRowsAboutToBeRemoved(const QModelIndex &parent, int first, int last);
    void onRowsRemoved(const QModelIndex &parent, int first, int last);
    void sourceModelReset();

private:
    void connectSourceModel(QAbstractItemModel *model);
    void updateRowKeys();
    bool sourceKeyExists(const QVariant &key, const QAbstractItemModel *excludeModel = nullptr) const;

    QList<MergeModelInfo> modelList;
    QVector<QVariant> rowKeys;
    QAbstractItemModel *primaryModel = nullptr;
};

#endif // MERGECOLUMNSTABLEMODEL_H
