/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceTable.h"

#include <QDebug>

DeviceIndexSet DeviceTable::keys() const
{
    DeviceIndexSet result;
    result.reserve(data.size());
    for (const auto &i : data)
        result.insert(i.index);
    return result;
}

void DeviceTable::insertStatic(const DeviceTableEntry &r)
{
    auto it = find(r.index);
    if(it != end()) {
        it->dynamic = DeviceTableEntry::Static; // make static from dynamic
        it->enabled = r.enabled;
        // emit dataChanged(...)
        return;
    }
    DeviceTableEntry dup(r);
    dup.dynamic = DeviceTableEntry::Static;
    data.append(dup);
}

void DeviceTable::insertDynamic(const DeviceTableEntry &r)
{
    for (auto it = data.begin(); it != data.end(); ++it) {
        if (it->index == r.index) {
            // do not make dynamic from static
            it->lldpInfo = r.lldpInfo;
            return;
        }
    }
    DeviceTableEntry dup(r);
    dup.dynamic = DeviceTableEntry::Dynamic;
    data.append(dup);
}

DeviceIndexSet DeviceTable::getIndices(DeviceTableEntry::DynamicFlag dynamic) const
{
    DeviceIndexSet result;
    for (auto it = data.begin(); it != data.end(); ++it) {
        if (it->dynamic == dynamic)
            result.insert(it->index);
    }
    return result;
}

void DeviceTable::remove(const DeviceIndexSet &r)
{
    if (r.isEmpty())
        return;
    QVector<DeviceTableEntry> result;
    for (auto it = data.begin(); it != data.end(); ++it) {
        if (r.contains(it->index))
            continue;
        result.append(*it);
    }
    data = result;
}

void DeviceTable::removeStatic(const DeviceIndexSet &r)
{
    if (r.isEmpty())
        return;
    const auto toDelete = r & getIndices(DeviceTableEntry::Static);
    if (toDelete.isEmpty())
        return;
    remove(toDelete);
}

int DeviceTable::rowOf(const DeviceIndex &deviceIndex) const
{
    auto it = find(deviceIndex);
    if(it == end())
        return -1;
    auto row = std::distance(begin(), it);
    return static_cast<int>(row);
}

DeviceTable::iterator DeviceTable::find(int row)
{
    if(row < 0 || row >= size())
        return end();
    return begin()+row;
}

DeviceTable::iterator DeviceTable::find(const DeviceIndex &deviceIndex)
{
    auto it = begin();
    while (it != end()) {
        if(it->index == deviceIndex)
            return it;
        ++it;
    }
    return it;
}

DeviceTable::const_iterator DeviceTable::find(int row) const
{
    if(row < 0 || row >= size())
        return end();
    return begin()+row;
}

DeviceTable::const_iterator DeviceTable::find(const DeviceIndex &deviceIndex) const
{
    auto it = begin();
    while (it != end()) {
        if(it->index == deviceIndex)
            return it;
        ++it;
    }
    return it;
}

//DeviceTable DeviceTable::filterDynamic(bool dynamic) const
//{
//    DeviceTable result;
//    for (auto it = data.begin(); it != data.end(); ++it) {
//        if (it->dynamic == dynamic) {
//            result.insert(*it);
//        }
//    }
//    return result;
//}

DeviceTable DeviceTable::filterStaticEnabled(bool enabled) const
{
    DeviceTable result;
    for (auto it = data.begin(); it != data.end(); ++it) {
        if (it->dynamic != DeviceTableEntry::Static)
            continue;
        bool match = enabled
                ? (it->enabled != DeviceTableEntry::Disable)
                : (it->enabled != DeviceTableEntry::Enable);
        if (!match)
            continue;
        if (it->enabled == enabled) {
            result.insertStatic(*it);
        }
    }
    return result;
}

//DeviceIndexEnabledVector DeviceTable::staticToVector() const
//{
//    return staticToVector(*this);
//}

DeviceIndexEnabledMap DeviceTable::staticToMap(const DeviceTable &deviceTable)
{
    DeviceIndexEnabledMap result;
    const auto &data = deviceTable.data;
    for (auto it = data.begin(); it != data.end(); ++it) {
        if (it->dynamic != DeviceTableEntry::Static)
            continue;
        DeviceIndexEnabled die(it->index, it->enabled != DeviceTableEntry::Disable);
        result.insert(it->index, die);
    }
    return result;
}

//DeviceIndexEnabledVector DeviceTable::staticToVector(const DeviceTable &deviceTable)
//{
//    DeviceIndexEnabledVector result;
//    const auto &data = deviceTable.data;
//    for (auto it = data.begin(); it != data.end(); ++it) {
//        if (it->dynamic != DeviceTableEntry::Static)
//            continue;
//        DeviceIndexEnabled die(it->index, it->enabled == DeviceTableEntry::Enable);
//        result.append(die);
//    }
//    return result;
//}

DeviceTable DeviceTable::staticFromMap(const DeviceIndexEnabledMap &v)
{
    DeviceTable result;
    for (auto it = v.begin(); it != v.end(); ++it) {
        DeviceTableEntry ent(it->index,
                             LldpInfo(),
                             DeviceTableEntry::Static,
                             it->enabled ? DeviceTableEntry::Enable : DeviceTableEntry::Disable
                             );
        result.insertStatic(ent);
    }
    return result;
}

DeviceTable DeviceTable::staticFromVector(const DeviceIndexEnabledVector &v)
{
    DeviceTable result;
    for (auto it = v.begin(); it != v.end(); ++it) {
        DeviceTableEntry ent(it->index,
                             LldpInfo(),
                             DeviceTableEntry::Static,
                             it->enabled ? DeviceTableEntry::Enable : DeviceTableEntry::Disable
                             );
        result.insertStatic(ent);
    }
    return result;
}

DeviceTableEntry DeviceTable::value(const DeviceIndex &index) const
{
    auto it = find(index);
    return it==end() ? DeviceTableEntry() : *it;
}

DeviceTableEntry DeviceTable::value(int row) const
{
    if(row < 0 || row >= size())
        return DeviceTableEntry();
    return data[row];
}
