/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceDispatcher.h"

#include <QDebug>
#include <QHostAddress>
#include <QLoggingCategory>
#include <QNetworkInterface>

#include "LldpDiscovery.h"

namespace  {

Q_DECLARE_LOGGING_CATEGORY(dev_dispatch)
Q_LOGGING_CATEGORY(dev_dispatch, "afi.device-dispatcher")

} // namespace

QHostAddress makeIPv4(quint32 address)
{
    if (!address)
        return QHostAddress(); // isNull
    return QHostAddress(address);
}

Q_DECLARE_METATYPE(QHostAddress)

DeviceDispatcher::DeviceDispatcher(QObject *parent)
    : AbstractDispatcher(parent),
      cache(new LldpInfoCache)
{
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<DeviceIndexEnabledMap>();
}

DeviceDispatcher::~DeviceDispatcher()
{
    delete cache;
}

void DeviceDispatcher::lldpCacheUpdated(const LldpInfoCache &new_cache)
{
    const auto old_cache = *cache;
    *cache = new_cache;
    const auto &old_keys = old_cache.keys();
    const DeviceIndexSet old_set = makeDeviceIndexSet(old_keys);
    const auto &new_keys = new_cache.keys();
    const DeviceIndexSet new_set = makeDeviceIndexSet(new_keys);
    // const DeviceIndexSet &common_set = new_set & old_set;
    const DeviceIndexSet &toAdd = new_set - old_set;
    const DeviceIndexSet &toRemove = old_set - new_set;
    for (auto it = toRemove.begin(); it != toRemove.end(); ++it) {
        logDeviceStateChange(*it, false);
    }
    for (auto it = toAdd.begin(); it != toAdd.end(); ++it) {
        // logDeviceStateChange(*it, true);
    }
    for (auto it = new_set.begin(); it != new_set.end(); ++it) {
        const DeviceIndex &index = *it;
        const QHostAddress old_ip(makeIPv4(old_cache.value(index).common.ip_addr));
        const QHostAddress new_ip(makeIPv4(new_cache.value(index).common.ip_addr));
        if (old_ip != new_ip) {
            logIpAddressChange(index, old_ip, new_ip);
        }
        tryConnect(index, new_cache.value(index));
    }
}

bool DeviceDispatcher::isMyAddress(const QHostAddress &address)
{
    return QNetworkInterface::allAddresses().contains(address);
}

QString DeviceDispatcher::deviceIdent(const DeviceIndex &index) const
{
    QString name = cache->value(index).med.modelname_str.trimmed();
    if (name.isEmpty())
        name = index.getDeviceIdName();
    name += " " + index.getSerialStr();
    return name;
}

void DeviceDispatcher::tryConnect(const DeviceIndex &index)
{
    if (cache->contains(index))
        tryConnect(index, cache->value(index));
}

void DeviceDispatcher::tryConnect(const DeviceIndex &index, const LldpInfo &info)
{
    const auto devIt = devices.constFind(index);
    if (devIt == devices.cend() || !devIt->enabled)
        return;

    const QHostAddress ip(makeIPv4(info.common.ip_addr));
#if 0
    // BUG: will not work with NAT
    if (info.master_locked) {
        const auto &master_ip = QHostAddress(makeIPv4(info.master_ip));
        if (!isMyAddress(master_ip)) {
            qInfo(dev_dispatch).noquote()
                << QString("%1 blocked by %2, not connecting")
                       .arg(deviceIdent(index))
                       .arg(master_ip.toString());
            return;
        }
    }
    qInfo(dev_dispatch).noquote()
        << QString("%1 connecting, IP %2")
               .arg(deviceIdent(index))
               .arg(ip.toString());
#endif
    emit setHostAddress(index, ip, info.msToExpire());
}

void DeviceDispatcher::setDeviceList(const DeviceIndexEnabledMap &new_devices)
{
    DeviceIndexEnabledMap filtered_devices(new_devices);
    filtered_devices.remove(DeviceIndex()); // remove invalid index if present
    if (filtered_devices == devices)
        return;
    const auto old_set = DeviceIndexSet(devices);
    const auto new_set = DeviceIndexSet(filtered_devices);
    const auto toAdd = new_set - old_set;
    const auto toRemove = old_set - new_set;
    const auto maybeUpdate = old_set & new_set;
    DeviceIndexSet toUpdate;
    for (auto i : maybeUpdate)
        if (devices[i] != filtered_devices[i])
            toUpdate.insert(i);
    logDevicesChange(toAdd, "add");
    logDevicesChange(toRemove, "remove");
    logDevicesChange(toUpdate, "update");
    devices = filtered_devices;
    emit deviceListChanged(devices);
    for (auto i : qAsConst(toUpdate))
        tryConnect(i);
    for (auto i : toAdd)
        tryConnect(i);
}

void DeviceDispatcher::logDevicesChange(const DeviceIndexSet &set, QString action)
{
    if (set.isEmpty())
        return;
    qDebug(dev_dispatch).noquote() << action << set;
}

bool DeviceDispatcher::isDeviceEnabled(const DeviceIndex &index) const
{
    auto found = devices.constFind(index);
    if (found == devices.cend())
        return false;
    return found->enabled;
}

void DeviceDispatcher::logDeviceStateChange(const DeviceIndex &index, bool discovered)
{
    if (!isDeviceEnabled(index))
        return;
    qDebug(dev_dispatch).noquote()
        << QString("%1 %2")
               .arg(deviceIdent(index))
               .arg(discovered ? tr("discovered") : tr("expired"));
}

void DeviceDispatcher::logIpAddressChange(const DeviceIndex &index, const QHostAddress &old_ip, const QHostAddress &new_ip)
{
    if (!isDeviceEnabled(index))
        return;

    if (!old_ip.isNull() && new_ip.isNull()) {
        qInfo(dev_dispatch).noquote()
            << QString("%1 released IP address %2")
                   .arg(deviceIdent(index))
                   .arg(old_ip.toString());
        return;
    }
    if (old_ip.isNull() && !new_ip.isNull()) {
        QString s;
        const auto &lldpInfo = cache->value(index);
        const auto &master = lldpInfo.ext.master;
        if (lldpInfo.ext.master_locked) {
            const auto &master_ip = master.hostAddress();
            if (!master.isNull() && !isMyAddress(master_ip)) {
                s += QString(" locked by %1:%2").arg(master_ip.toString()).arg(master.port);
            }
        }
        qInfo(dev_dispatch).noquote()
            << QString("%1 discovered, IP address %2")
                    .arg(deviceIdent(index))
                    .arg(new_ip.toString()) + s;
        return;
    }
    qInfo(dev_dispatch).noquote()
        << QString("%1 changed IP address %2 to %3")
               .arg(deviceIdent(index))
               .arg(old_ip.toString())
               .arg(new_ip.toString());
}
