/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "AbstractDeviceModule.h"

#include "MregBusInterface.h"
#include "base/DeviceId.h"
#include "base/DeviceIndex.h"
#include "base/SdbModuleIndex.h"
#include "base/fw_version.h"
#include "mregdev/qtmregdevice.h"
#include "regio/regop.h"
#include "DeviceModuleCommand.h"

AbstractDeviceModule::AbstractDeviceModule(const MregBusInterface &bus) :
    bus(bus)
{}

AbstractDeviceModule::~AbstractDeviceModule()
{}

bool AbstractDeviceModule::isAvailable() const
{
    return bus.isAvailable();
}
quint16 AbstractDeviceModule::baseAddr() const
{
    return bus.baseAddr();
}
quint16 AbstractDeviceModule::lastAddr() const
{
    return bus.lastAddr();
}

uint8_t AbstractDeviceModule::majorVersion() const
{
    return bus.sdbDevice.abi_ver_major;
}

fw_version_t AbstractDeviceModule::getVersion() const
{
    return fw_version_t(bus.sdbDevice.abi_ver_major,
                        bus.sdbDevice.abi_ver_minor, 0);
}

QString AbstractDeviceModule::name() const
{
    return bus.sdbDevice.component.product.name;
}

AFI_SDB_ID AbstractDeviceModule::getSdbId() const
{
    return bus.getSdbId();
}

SdbModuleIndex AbstractDeviceModule::moduleIndex() const
{
    return bus.moduleIndex();
}

DeviceId AbstractDeviceModule::getDeviceId() const
{
    if (!bus.dev)
        return {};
    return bus.dev->getDeviceId();
}

DeviceIndex AbstractDeviceModule::getDeviceIndex() const
{
    if (!bus.dev)
        return {};
    return bus.dev->getDeviceIndex();
}

void AbstractDeviceModule::setBaseAddr(quint16 baseAddr, quint16 mask)
{
    if (isAvailable())
        qWarning() << __PRETTY_FUNCTION__ << getDeviceIndex() << "base address already set";
    bus.setAddrRange(AddrRange(baseAddr, baseAddr | mask));
}

bool AbstractDeviceModule::regOpExec(RegOpVector &data)
{
    Q_ASSERT(bus.dev);
//    Q_ASSERT(bus.isAvailable());
    if (!bus.dev)
        return {};
    return bus.dev->regOpExec(data);
}

bool AbstractDeviceModule::memReadBlk(int startWord, size_t wordCount, std::vector<quint32> *result)
{
    Q_ASSERT(bus.dev);
    Q_ASSERT(bus.isAvailable());
    if (!bus.dev)
        return {};
    return bus.dev->memReadBlk(startWord, wordCount, result);
}

bool AbstractDeviceModule::memWriteBlk(int wordNumber, const std::vector<quint32> &data)
{
    Q_ASSERT(bus.dev);
    Q_ASSERT(bus.isAvailable());
    if (!bus.dev)
        return {};
    return bus.dev->memWriteBlk(wordNumber, data);
}

RegOpVector AbstractDeviceModule::regOpDumpAll(QVector<quint16> *result) const
{
    if (!result)
        return {};
    auto count = lastAddr() + 1 - baseAddr();
    result->resize(count);
    RegOpVector r;
    for (auto i = 0; i < count; i++) {
        quint16 &item = (*result)[i];
        r.RegRead16(i, &item);
    }
    return r;
}

DeviceModuleCommandResult AbstractDeviceModule::runCommand(const DeviceModuleCommand &command)
{
    DeviceModuleCommandResult result;
    result.seq = command.seq;
    result.error = EINVAL;
    Q_ASSERT(false);
    return result;
}
