#include "FeLinkStatusModule.h"

#include "base/DeviceId.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

#include "FeLinkStatus.h"
#include "FeLinkDeviceUtil.h"
#include "FeLinkConfig.h"

FeLinkStatusModule::FeLinkStatusModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
}

QList<AFI_SDB_ID> FeLinkStatusModule::getSupportedSdbIds()
{
    return {AFI_SDB_FE_LINK_CSR};
}

void FeLinkStatusModule::post_init()
{
    BaseDeviceModule::post_init();
    buildParams.reset();

    if(majorVersion()==2) {
        RegOpVector r = getFeLinkBuildParamsROV();
        if (regOpExecRebased(r)) {
            buildParams = getFeLinkBuildParams(r);
            if(buildParams && buildParams->isValid()) {
                auto st = readStatus();
                if(st) {
                    buildParams->chInfo.resize(buildParams->chNum);
                    for(int ch=0; ch<buildParams->chNum ; ++ch) {
                        buildParams->chInfo[ch] = st->value(ch).txInfo;
                    }
                }
            }
        }
    }
}

std::optional<FeLinkStatusMap> FeLinkStatusModule::readStatus()
{
    switch(majorVersion()) {
    case 1: return readStatus_v1();
    case 2: return readStatus_v2();
    default: return {};
    }
}

bool FeLinkStatusModule::writeConfig(const FeLinkConfig &feLink)
{
    RegOpVector regOp;
    switch(majorVersion()) {
    case 1: return true;
    case 2:
    {
        if (!buildParams || !buildParams->isValid())
            return {};
        int ch=0;
        for(auto it=feLink.chCfg.cbegin(); it!=feLink.chCfg.cend(); ++it) {
            if(ch >= buildParams->chNum)
                break;
            quint16 baseAddr = buildParams->startAddr+ch*buildParams->addStep;
            regOp << getWriteChConfigROV_v2(baseAddr, *it);
            ++ch;
        }

        return regOpExecRebased(regOp);
    }
    default:;
    }
    return {};
}

bool FeLinkStatusModule::phyReset(int resetCh)
{
    if(!buildParams || !buildParams->isValid())
        return {};
    switch (majorVersion()) {
    case 2:
    {
        auto regOpRead = getPhyLevelCtrlRegsROV(buildParams.value());
        if(!regOpExecRebased(regOpRead))
            return {};

        auto regs = getPhyLevelCtrlRegs(regOpRead);
        if(regs.size() != buildParams->chNum)
            return {};


        auto regOpReset = selfReleasePhyResetUtil(regs, buildParams.value(), resetCh);

        return regOpExecRebased(regOpReset);
    }
    default:;
    }
    return true;
}

std::optional<FeLinkStatusMap> FeLinkStatusModule::readStatus_v1()
{
    const auto FE_LINK_CNT = getFeLinkCnt();

    mlink::RegOpVector regOp;
    std::optional<int> singleSize;
    for(int ch=0; ch<FE_LINK_CNT ; ++ch) {
        regOp << getFeLinkStatusROV(ch);
        if(singleSize)
            Q_ASSERT((int)regOp.size() == singleSize.value() * (ch+1));
        else
            singleSize = regOp.size();
    }

    if (!regOpExecRebased(regOp))
        return {};

    FeLinkStatusMap stMap;
    for(int ch=0; ch<FE_LINK_CNT ; ++ch) {
        FeLinkStatus st = getFeLinkStatus(regOp, ch*singleSize.value());
        st.readerIndex = getDeviceIndex();
        st.txInfo.ch = ch;
        stMap[ch] = st;
    }

    return stMap;
}

std::optional<FeLinkStatusMap> FeLinkStatusModule::readStatus_v2()
{
    if(!buildParams)
        return {};
    auto regOp = getStatusLockROV(true);
    const int baseSize = regOp.size();
    std::optional<int> singleSize;
    for(int ch=0; ch<buildParams->chNum ; ++ch) {
        quint16 baseAddr = buildParams->startAddr+ch*buildParams->addStep;
        const auto &chROV = getFeLinkStatusROV_v2(baseAddr);
        regOp << chROV;
        if(singleSize)
            Q_ASSERT(chROV.size() == *singleSize);
        else
            singleSize = chROV.size();
    }
    regOp << getStatusLockROV(false);

    if (!regOpExecRebased(regOp))
        return {};

    FeLinkStatusMap stMap;
    for(int ch=0; ch<buildParams->chNum ; ++ch) {
        FeLinkStatus st = getFeLinkStatus_v2(regOp, baseSize+ch*singleSize.value_or(0));
        st.readerIndex = getDeviceIndex();
        st.txInfo.ch = ch;
        stMap[ch] = st;
    }

    return stMap;
}

int FeLinkStatusModule::getFeLinkCnt() const
{
    if(buildParams)
        return buildParams->chNum;
    static const int TTVXS_CH_NUM = 18;
    if(getDeviceId() == DEVICE_ID_CRU16WR)
        return 18;
    return getDeviceId() == DEVICE_ID_TTVXS ? TTVXS_CH_NUM : 1;
}
