/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MODULARMREGDEVICE_H
#define MODULARMREGDEVICE_H

#include "ModularDeviceConfig.h"
#include "MregBusInterface.h"
#include "base/ConfigKey.h"
#include "mregdev/qtmregdevice.h"
#include "sdb/SdbTypes.h"
#include <QVariant>

enum AFI_SDB_ID : uint32_t;

 class AbstractDeviceModule;
 struct AdcBuildParams;
 struct AdcDeserCheckStatus;
 struct AdcLemoMuxBuildParams;
 struct AdcMpdTrigBuildParams;
 class AdcSerDesCalibration;
 class AdcSerDesDelayMap;
 class AdcSpiBaseModule;
 struct ClockBuildParams;
 struct ClockStatus;
 struct ClockDividerStatus;
 struct CrosspointSwitchBuildParams;
 struct DeserializersControlBuildParams;
 class DeviceCommand;
 struct DlnpCalibrationTriggerBuildParams;
 struct FeLinkBuildParams;
 struct FeLinkStatus;
 struct FirFilterBuildParams;
 class ModularDeviceStatus;
 class ModularMregDevicePrivate;
 struct NetworkPortConfig;
 class RunLogicStatus;
 struct SdbModuleIndex;
 struct TdcCoreStatus;
 struct Tlu16SfpTriggerBuildParams;
 struct Tlu40LvdsLuminosityBuildParams;
 struct Tlu40LvdsLuminosityStatus;
 struct Tlu40LvdsTofTriggerBuildParams;
 struct Tlu40LvdsTofTriggerStatus;
 struct TluDecDesBuildParams;
 struct TluDecDesStatus;
 struct TluTtlIoControlBuildParams;

 struct TrigCtrlBuildParams;

class ModularMregDevice : public QtMregDevice
{
    Q_OBJECT
public:
    explicit ModularMregDevice(const DeviceIndex &deviceIndex,
                               ModularDeviceStatus &status,
                               QObject *parent = nullptr);
    ~ModularMregDevice() override;

    template<class T, typename... Arguments>
    QSharedPointer<T> createModule(Arguments... args) {
        auto ptr = QSharedPointer<T>(new T(args..., this));
        insertModuleImpl(ptr);
        return ptr;
    }

    template<class T, typename... Arguments>
    QSharedPointer<T> createSingleModule(Arguments... args) {
        const QList<AFI_SDB_ID> &ids = T::getSupportedSdbIds();
        const QList<SdbDevice> &records = getSdb().filterDevices(ids);

        if(!records.isEmpty()) {
            const auto &sdbDevice = records.first();
            MregBusInterface bus(this, sdbDevice);
            auto module = QSharedPointer<T>(new T(bus, args...));
            insertModuleImpl(module);
            return module;
        }

        return createModule<T>();
    }

    QList<SdbModuleIndex> getModuleIndexes() const;

    QSharedPointer<AbstractDeviceModule> findModule(const SdbModuleIndex &moduleIndex);
    QSharedPointer<const AbstractDeviceModule> findModule(const SdbModuleIndex &moduleIndex) const;
    QList<QSharedPointer<AbstractDeviceModule>> findModules(AFI_SDB_ID id);
    QList<QSharedPointer<const AbstractDeviceModule>> findModules(AFI_SDB_ID id) const;
    QSharedPointer<AbstractDeviceModule> getFirstModule(AFI_SDB_ID id);
    QSharedPointer<const AbstractDeviceModule> getFirstModule(AFI_SDB_ID id) const;

    template<class T>
    QMap<int, QSharedPointer<T>> getModules()  const;

    template<class T>
    QSharedPointer<T> getFirstModule() const
    {
        auto modules = getModules<T>();
        if (modules.empty())
            return {};
        return modules.first();
    }

    virtual QSharedPointer<AdcSpiBaseModule> getAdcSpiModule() const;

signals:
    void deviceSdbListUpdated(const DeviceIndex &index, const QVariantList &sdbList) const;
    void deserializersControlBpUpdated(const DeviceIndex &devIndex, const DeserializersControlBuildParams &bp);
    void dlnpCalibrationTriggerBpUpdated(const DeviceIndex &index, const DlnpCalibrationTriggerBuildParams &bp);
    void feLinkStatusUpdated(const DeviceIndex &index, const QMap<int, FeLinkStatus> &stMap);
    void feLinkBpUpdated(const DeviceIndex &index, const FeLinkBuildParams &bp);
    void firFilterBpUpdated(const DeviceIndex &index, const FirFilterBuildParams &bp);
    void tlu16SfpTrigBpUpdated(const DeviceIndex &index, const Tlu16SfpTriggerBuildParams &bp);
    void tlu40LvdsLuminosityBpUpdated(const DeviceIndex &index, const Tlu40LvdsLuminosityBuildParams &bp);
    void tlu40LvdsLuminosityStatusUpdated(const DeviceIndex &index, const Tlu40LvdsLuminosityStatus &bp);
    void tlu40LvdsTofTriggerStatusUpdated(const DeviceIndex &index, const Tlu40LvdsTofTriggerStatus &bp);
    void tlu40LvdsTofTriggerBpUpdated(const DeviceIndex &index, const Tlu40LvdsTofTriggerBuildParams &bp);
    void tluDecDesBpUpdated(const DeviceIndex &index, const TluDecDesBuildParams &bp);
    void tluDecDesStatusUpdated(const DeviceIndex &index, const TluDecDesStatus &bp);
    void tluTtlIoControlBpUpdated(const DeviceIndex &index, const TluTtlIoControlBuildParams &bp);
    void trigCtrlBpUpdated(const DeviceIndex &index, const TrigCtrlBuildParams &bp);
    void adcBuildParamsUpdated(const DeviceIndex &index, const AdcBuildParams &bp);
    void adcLemoMuxBpUpdated(const DeviceIndex &index, const AdcLemoMuxBuildParams &bp);
    void adcMpdTrigBpUpdated(const DeviceIndex &index, const AdcMpdTrigBuildParams &bp);
    void clockControlBpUpdated(const DeviceIndex &index, const ClockBuildParams &bp);
    void crosspointSwitchBpUpdated(const DeviceIndex &index, const CrosspointSwitchBuildParams &bp);
    void hptdcStatusUpdated(const DeviceIndex &index, const TdcCoreStatus &);
    void devSerDesDelayChanged(const DeviceIndex &index, const AdcSerDesDelayMap &delay);
    void adcSerDesProgressUpdated(const DeviceIndex &index, int percenage) const;
    void adcDeserConfigChecked(const DeviceIndex &, const AdcDeserCheckStatus &st) const;
    void clockControlStatusUpdated(const DeviceIndex &index, const ClockStatus &);
    void clockDividerStatusUpdated(const DeviceIndex &index, const ClockDividerStatus &);
    void runLogicStatusUpdated(const DeviceIndex &index, const RunLogicStatus &);
    void statusUpdated(const ModularDeviceStatus &status);
    void deviceConfigWritten(const DeviceIndex &index, const ConfigKey &key);
    void deviceCommandComplete(const DeviceIndex &index, const DeviceCommand &command);
    void calibrateFinished(const DeviceIndex &index, const ConfigKey &cal_key);
    void evNumResponse(int checkIndex, const DeviceIndex &index, quint64 evNum);

public slots:
    void rebootDevice(const DeviceIndex &index);
    void submitDeviceCommand(const DeviceIndex &index, const DeviceCommand &command);
    void setDeviceConfig(const ConfigKey &key, const ModularDeviceConfig &newConfig);
    virtual void startAdcSrdDesCalibration(const ConfigKey &calibrationKey, bool hardMode, const QString &resDir);
    virtual void evNumRequest(int checkIndex);
    void feLinkPhyReset(const DeviceIndex &dev, int ch);
    void adcClkDivSync();

protected slots:
    void on_deviceConnected() override;
    void on_deviceDisconnected() override;
//    bool writeNetworkPortConfig(const NetworkPortConfig &config);

private slots:
    void on_adcSerDesProgressUpdated(int p) const;
    void checkAdcDeserConfig() const;

protected:
    QMap<QString, double> readTemp();
    bool readStatus() override;
    bool readModulesStatus();
    bool writeConfig() override;
    virtual bool preRunActions(const QVariant &evNum = {});
    virtual bool writeSoftRun(bool runState);
    bool syncConfigDeferred();
    void enforceConfigConstraints();
    virtual void enforceConfigConstraints(ModularDeviceConfig *) {}
    virtual void createModules();
    virtual void initModules();

private:
    void setConfigData(const ConfigKey &key, const ModularDeviceConfig &newConfig);
    QSharedPointer<AbstractDeviceModule> insertModuleImpl(QSharedPointer<AbstractDeviceModule> module);
    bool modWriteConfig(QSharedPointer<AbstractDeviceModule> mod, bool modWriteOk);

    ModularMregDevicePrivate * const d_ptr;
    Q_DECLARE_PRIVATE(ModularMregDevice)
    const DeviceIndex deviceIndex;
    ModularDeviceStatus &status;
    AdcSerDesCalibration * const adcSerDesCal;
    QTimer * const timerAdcDesCheck;

protected:
    void dropConfigCache() { configCache = {}; setupCacheValid = false; }
    void setConfigCache() { configCache = config; setupCacheValid = true; }
    bool isConfigCacheValid() const { return setupCacheValid; }

    ModularDeviceConfig config;
    ModularDeviceConfig configCache;
    ConfigKey config_key;
    bool setupCacheValid = false;
    QSet<QPair<AFI_SDB_ID, quint16>> failedModules;
};

#endif // MODULARMREGDEVICE_H
