/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RunLogicController.h"
#include "AsyncDeviceCommander.h"
#include "DeviceCommand.h"
#include "DeviceCommandType.h"
#include "sdb/AfiSdbId.h"

RunLogicController::RunLogicController(QObject *parent)
    : QObject{parent}
{
    qRegisterMetaType<RunLogicMultiStatus>();
}

void RunLogicController::setDeviceCommander(const AsyncDeviceCommander *deviceCommander)
{
    connect(deviceCommander, &AsyncDeviceCommander::commandComplete,
            this, &RunLogicController::commandComplete);
    connect(this, &RunLogicController::submitCommand,
            deviceCommander, &AsyncDeviceCommander::submitCommand);
}

void RunLogicController::setDeviceDispatcher(DeviceDispatcher *deviceDispatcher)
{
    connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
            this, &RunLogicController::setDeviceList);
}

void RunLogicController::setDeviceList(const DeviceIndexEnabledMap &new_devices)
{
    multiStatus.setDeviceList(new_devices);
    emit runLogicStatusChanged(multiStatus);
}

void RunLogicController::setRunLogicConfig(const RunLogicConfig &newConfig)
{
    if (multiStatus.runLogicConfig() == newConfig)
        return;
    multiStatus.setRunLogicConfig(newConfig);
    emit runLogicStatusChanged(multiStatus);
    emit runLogicConfigChanged(newConfig);
}

void RunLogicController::setRunLogicStatus(const DeviceIndex &deviceIndex, const RunLogicStatus &newStatus)
{
    if (!newStatus.isValid())
        return;
    multiStatus.setRunLogicStatus(deviceIndex, newStatus);
    emit runLogicStatusChanged(multiStatus);
    checkAutoPause();
}

//void RunLogicController::setDeviceStatus(const DeviceStatusBase &newStatus)
//{
//    if (!newStatus.runLogicStatus || !newStatus.runLogicStatus->isValid())
//        return;
//    setRunLogicStatus(newStatus.getDeviceIndex(), *newStatus.runLogicStatus);
//}

bool RunLogicController::setState(RunLogicFsm::State toState)
{
    bool ok = fsm.setState(toState);
    if (ok) {
        emit stateChanged(fsm.getState());
        checkAutoPause();
    }
    return ok;
}

void RunLogicController::commandComplete(const DeviceCommand &command)
{
    if (command.result.seq != fsm.seq) {
        qWarning().noquote() << QString("Unknown sequence %1, expecting %2")
                                    .arg(command.result.seq)
                                    .arg(fsm.seq);
        return;
    }
    RunLogicFsm::State toState = {};
    switch(fsm.getState()) {
    case RunLogicFsm::SentReset:
        toState = RunLogicFsm::Reset;
        break;
    case RunLogicFsm::SentRun:
        toState = RunLogicFsm::Run;
        break;
    case RunLogicFsm::SentPause:
        toState = RunLogicFsm::Pause;
        break;
    default:
        return;
    }
    setState(toState);
}

void RunLogicController::checkAutoPause()
{
    // auto-pause
    switch(fsm.getState()) {
    case RunLogicFsm::Run:
        if (getStatus().allFinished()) {
            performAction(Pause);
        }
        break;
    default:;
    }
}

DeviceCommand RunLogicController::makeCommand(DeviceCommandType type, const QVariant &data)
{
    DeviceCommand command;
    command.moduleIndex.id = AFI_SDB_RUN_LOGIC;
    command.command.seq = (++fsm.seq);
    command.command.type = type;
    command.command.data = data;
    return command;
}

bool RunLogicController::performAction(int action)
{
    bool isRun = fsm.getState() == RunLogicFsm::Run;

    static const QMap<Action, DeviceCommandType> map
        {{Reset, DeviceCommandType::Restart},
         {Run, DeviceCommandType::Start},
         {Pause, DeviceCommandType::Pause},
         {Toggle, isRun ? DeviceCommandType::Pause : DeviceCommandType::Start}};

    static const QMap<DeviceCommandType, RunLogicFsm::State> mapState
        {{DeviceCommandType::Restart, RunLogicFsm::SentReset},
         {DeviceCommandType::Start, RunLogicFsm::SentRun},
         {DeviceCommandType::Pause, RunLogicFsm::SentPause}};

    auto commandType = map.constFind(RunLogicController::Action(action));
    if (commandType ==map.cend())
        return false;
    auto targetState = mapState.constFind(*commandType);
    if (targetState == mapState.cend())
        return false;

    if (! fsm.transitionAllowed(*targetState))
        return false;
    const auto &command = makeCommand(*commandType);
    setState(*targetState);
    emit submitCommand(command);
    return true;
}
