#include "TdcCoreModule.h"

#include "TdcCoreConfig.h"
#include "TdcCoreStatus.h"

#include "base/DeviceId.h"
#include "base/DeviceIndex.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

union reg_measurement_ctrl {
    quint32 all;
    struct {
        unsigned int leading_resolution: 3;  // 2:0
        unsigned int width_select: 4;        // 6:3
        unsigned int unused: 1;              // 7
        unsigned int mode_rc_compression: 1; // 8
        unsigned int mode_rc: 1;             // 9
        unsigned int DLL_Mode: 2;            // 11:10
        unsigned int enable_leading: 1;      // 12
        unsigned int enable_trailing: 1;     // 13
        unsigned int enable_pair: 1;         // 14
        unsigned int core_clock_source: 2;   // 16:15
        unsigned int dll_clock_source: 3;    // 19:17
        unsigned int unused2: 4;             // 23:20
        unsigned int enable_matching: 1;     // 24
        unsigned int unused3: 7;             // 31:25
    } bit;
};

inline bool operator==(const reg_measurement_ctrl &l, const reg_measurement_ctrl &r)
{
    return l.all == r.all;
}

union reg_read_ctrl {
    quint32 all;
    struct {
        unsigned int enable_local_header: 1;      // 0
        unsigned int enable_local_trailer: 1;     // 1
        unsigned int unused: 2;                   // 3:2
        unsigned int enable_error_mark: 1;        // 4
        unsigned int unused2: 3;                  // 7:5
        unsigned int readout_fifo_size: 3;        // 10:8
        unsigned int unused3: 1;                  // 11
        unsigned int max_event_size: 4;           // 15:12
        unsigned int reject_readout_fifo_full: 1; // 16
        unsigned int unused4: 3;                  // 19:17
        unsigned int dead_time: 2;                // 21:20
        unsigned int unused5: 10;                 // 31:22
    } bit;
};

inline bool operator==(const reg_read_ctrl &l, const reg_read_ctrl &r)
{
    return l.all == r.all;
}

// TDC Core Registers
enum {
    REG32_TDC_CH_EN_BASE    = 0x00,
    REG32_TDC_GEN_CTRL      = 0x20,
    REG32_TDC_READ_CTRL     = 0x22,
    REG32_TDC_MEASURE_CTRL  = 0x24,
    REG32_TDC_TRIG_WIN_SETUP= 0x26,
    REG32_TDC_GEN_STATUS    = 0x2c,
//    REG32_TDC_RC_ADJUST_SETUP = 0x28 //unused
};

// HPTDCs Status Registers
enum {
    REG_TDC_HPTDC_STATUS_BASE = 0x80,
//    REG_HPTDC1_STATUS_2 = 0x81,
//    REG_HPTDC1_STATUS_3 = 0x82,
//    REG_HPTDC1_STATUS_4 = 0x83,
    // Reg 0x0084-0x00FF - same for every other HPTDC
    // regs repeated for every HPTDC
};

enum {
    REG32_TDC_GEN_CTRL_BIT_STANDBY = 1
};

enum {
    REG32_TDC_GEN_STAT_BIT_STANDBY = 0x1,
    REG32_TDC_GEN_STAT_BIT_READY = 0x2
};

QDebug operator<<(QDebug dbg, const reg_measurement_ctrl &s)
{
    dbg << "reg_measurement_ctrl{"
        << "leading_resolution:" << s.bit.leading_resolution
        << "width_select:" << s.bit.width_select
        << "unused:" << s.bit.unused
        << "mode_rc_compression:" << s.bit.mode_rc_compression
        << "mode_rc:" << s.bit.mode_rc
        << "DLL_Mode:" << s.bit.DLL_Mode
        << "enable_leading:" << s.bit.enable_leading
        << "enable_trailing:" << s.bit.enable_trailing
        << "enable_pair:" << s.bit.enable_pair
        << "core_clock_source:" << s.bit.core_clock_source
        << "dll_clock_source:" << s.bit.dll_clock_source
        << "unused2:" << s.bit.unused2
        << "enable_matching:" << s.bit.enable_matching
        << "unused3:" << s.bit.unused3
        << "}";
    return dbg;
}

QDebug operator<<(QDebug dbg, const reg_read_ctrl &s)
{
    dbg << "reg_read_ctrl{"
        << "enable_local_header:" << s.bit.enable_local_header
        << "enable_local_trailer:" << s.bit.enable_local_trailer
        << "unused:" << s.bit.unused
        << "enable_error_mark:" << s.bit.enable_error_mark
        << "unused2:" << s.bit.unused2
        << "readout_fifo_size:" << s.bit.readout_fifo_size
        << "unused3:" << s.bit.unused3
        << "max_event_size:" << s.bit.max_event_size
        << "reject_readout_fifo_full:" << s.bit.reject_readout_fifo_full
        << "unused4:" << s.bit.unused4
        << "dead_time:" << s.bit.dead_time
        << "unused5:" << s.bit.unused5
        << "}";
    return dbg;
}

TdcCoreModule::TdcCoreModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
    if (!isAvailable()) {
        setBaseAddr(0x200, 0xFF); // default reg base for versions without SDB
    }
    switch (getDeviceId()) {
    case DEVICE_ID_PHOS_FEC:    chipNum = 1; break;
    case DEVICE_ID_TQDC16VS_E:  chipNum = 2; break;
    case DEVICE_ID_TDC64VHLE:   chipNum = 8; break;
    case DEVICE_ID_TDC64VLE:    chipNum = 2; break;
    case DEVICE_ID_TDC72VXS:    chipNum = 9; break;
    default: chipNum = 0; break;
    }
}

bool TdcCoreModule::vhrOnlyMode() const
{
    return getDeviceId() != DEVICE_ID_TDC64VLE;
}

QList<AFI_SDB_ID> TdcCoreModule::getSupportedSdbIds()
{
    return {AFI_SDB_HPTDC_RegIO_JTAG};
}

void TdcCoreModule::post_init()
{
    setStandby(true); // no config yet
    //    RegOpVector r;
    //    r.RegWrite32(REG32_TDC_MEASURE_CTRL, measurementCtrlVal.all);
    //    r.RegWrite32(REG32_TDC_READ_CTRL, readCtrlVal.all);
    //    regOpExecRebased(r);
}

int TdcCoreModule::getNch() const
{
    bool vHR = vhrOnlyMode() ? true : config.adjustConfig.mode_rc();
    const int chPerChip = vHR ? 8 : 32;
    return chipNum * chPerChip;
}

int TdcCoreModule::getNchPerChip() const
{
    bool vHR = vhrOnlyMode() ? true : config.adjustConfig.mode_rc();
    return vHR ? 8 : 32;
}

std::vector<quint32> TdcCoreModule::getChMask() const
{
    bool vHR = vhrOnlyMode() ? true : config.adjustConfig.mode_rc();
    std::vector<quint32> chMask;
    const auto chPerChip = getNchPerChip();
    const auto nch = getNch();
    chMask.resize(nch / chPerChip, 0);
    for (quint16 ch = 0; ch < nch; ch++) {
        if (!config.chTdcEn.value(ch, false))
            continue;

        auto chInv = nch - 1 - ch;
        quint16 hptdcId = chInv / chPerChip;
        quint32 v = vHR ? 0xF : 1;
        v <<= (chInv % chPerChip) * (vHR ? 4 : 1);
        chMask[hptdcId] |= v;
    }
    return chMask;
}

bool TdcCoreModule::setChEnMask()
{
    auto chMask = getChMask();
    RegOpVector r;
    quint16 addr = REG32_TDC_CH_EN_BASE;
    for (auto mask : chMask) {
        r.RegWrite32(addr, mask);
        addr += 2;
    }
    return regOpExecRebased(r);
}

bool TdcCoreModule::setStandby(bool standBy)
{
    RegOpVector r;
    r.RegWrite32(REG32_TDC_GEN_CTRL, standBy ? REG32_TDC_GEN_CTRL_BIT_STANDBY : 0);
    return regOpExecRebased(r);
}

bool TdcCoreModule::writeHptdcSetup()
{
    const auto &config = this->config;
    auto adjustConfig = config.adjustConfig;
    if (config.autoResolution)
        adjustConfig.resolution = vhrOnlyMode() ? TDC_RES_INDEX_25 : TDC_RES_INDEX_100;

    reg_read_ctrl readCtrlVal;
    readCtrlVal.bit.enable_local_header = config.hptdcHdr;
    readCtrlVal.bit.enable_local_trailer = config.hptdcTrl;
    readCtrlVal.bit.enable_error_mark = adjustConfig.enable_error_mark;
    readCtrlVal.bit.readout_fifo_size = adjustConfig.readout_fifo_size;
    readCtrlVal.bit.max_event_size = adjustConfig.max_event_size;
    readCtrlVal.bit.reject_readout_fifo_full = adjustConfig.reject_readout_fifo_full;
    readCtrlVal.bit.dead_time = adjustConfig.dead_time;

    reg_measurement_ctrl measurementCtrlVal;
    measurementCtrlVal.bit.leading_resolution = adjustConfig.leading_resolution();
    measurementCtrlVal.bit.width_select = 0;
    measurementCtrlVal.bit.mode_rc_compression = adjustConfig.mode_rc_compression();
    measurementCtrlVal.bit.mode_rc = adjustConfig.mode_rc();
    measurementCtrlVal.bit.DLL_Mode = adjustConfig.dll_mode();
    measurementCtrlVal.bit.enable_leading = config.leadEdge;
    measurementCtrlVal.bit.enable_trailing = config.trailEdge;
    measurementCtrlVal.bit.enable_pair = adjustConfig.enable_pair;
    measurementCtrlVal.bit.core_clock_source = adjustConfig.core_clock_source;
    measurementCtrlVal.bit.dll_clock_source = adjustConfig.dll_clock();
    measurementCtrlVal.bit.enable_matching = adjustConfig.enable_matching;

    RegOpVector r;
    r.RegWrite32(REG32_TDC_TRIG_WIN_SETUP, (quint32(config.matchWin) & 0xFFF) << 12 | (config.latency & 0xFFF));
    r.RegWrite32(REG32_TDC_READ_CTRL, readCtrlVal.all);
    r.RegWrite32(REG32_TDC_MEASURE_CTRL, measurementCtrlVal.all);
    return regOpExecRebased(r);
}

bool TdcCoreModule::writeConfig(const TdcCoreConfig &newConfig)
{
    config = newConfig;
    bool ok = true;
    ok &= setChEnMask(); // no reset required

    bool hptdcOk = false;
    for(auto i=0; i<5; ++i) {
        if(!setStandby(true))
            continue;
        if(!waitHptdcInit())
            continue;
        if(!writeHptdcSetup())
            continue;
        if(!setStandby(false))
            continue;
        if(!waitHptdcReady())
            continue;

        hptdcOk = true;
        break;
    }
    ok &= hptdcOk;
    return ok;
}

RegOpVector TdcCoreModule::getStatusRegOp(TdcCoreStatus &result) const
{
    RegOpVector r;
    r.RegRead32(REG32_TDC_GEN_STATUS, &result.generalStatus);
    result.hptdcStatus.resize(chipNum);
    for (quint16 i = 0; i < chipNum; ++i) {
        for (quint16 reg = 0; reg < HptdcStatus::regCount; ++reg) {
            r.RegRead16(REG_TDC_HPTDC_STATUS_BASE + (i * HptdcStatus::regCount) + reg,
                        &result.hptdcStatus[i].regValues[reg]);
        }
    }
    r.rebase(baseAddr());
    return r;
}

std::optional<TdcCoreStatus> TdcCoreModule::readStatus()
{
    TdcCoreStatus st;
    auto regOp = getStatusRegOp(st);
    if (!regOpExec(regOp))
        return {};
    return st;
}

bool TdcCoreModule::waitHptdcInit()
{
    return waitHptdcState(REG32_TDC_GEN_STAT_BIT_STANDBY);
}

bool TdcCoreModule::waitHptdcReady()
{
    return waitHptdcState(REG32_TDC_GEN_STAT_BIT_READY);
}

bool TdcCoreModule::waitHptdcState(quint32 state)
{
    QElapsedTimer timer;
    timer.start();
    constexpr static const qint64 STANDBY_TIMEOUT = 200; // ~100ms for pll reset
    while(true) {
        quint32 genStatus;
        if (!regRead32(REG32_TDC_GEN_STATUS, &genStatus))
            return false;
        if (genStatus & state)
            break;
        if (timer.elapsed() > STANDBY_TIMEOUT) {
            qWarning() << QString("[%1] HPTDC init timeout, %2 TDC_GEN_STATUS: %3 (expected: %4)")
                              .arg(getDeviceIndex().getIdent()).arg(name())
                              .arg(genStatus, 8, 16, QChar('0'))
                              .arg(state, 8, 16, QChar('0'));
            return false;
        }
    }
    return true;
}
