#include "ModularDeviceStatusModel.h"

#include <QColor>
#include <QHostAddress>
#include <QPalette>

#include "ModularDeviceStatus.h"
#include "app/AppStyle.h"
#include "base/DaqModuleTraits.h"
#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"

constexpr const char *BUSY_TOOLTIP = "Current Busy Status\n"
                                     "1xxxx - Dead time busy\n"
                                     "x1xxx - External input busy\n"
                                     "xx1xx - ADC busy\n"
                                     "xxx1x - TDC busy\n"
                                     "xxxx1 - ReadOut busy";

struct DeviceStatusBaseModelPrivateEntry
{
    explicit DeviceStatusBaseModelPrivateEntry(const DeviceIndex &deviceIndex = {})
        : status(deviceIndex) {}
    ModularDeviceStatus status;
    ModularDeviceStatus delta;
};

class ModularDeviceStatusModelPrivate : public QMap<DeviceIndex, DeviceStatusBaseModelPrivateEntry>
{
public:
    int rowOf(const DeviceIndex &deviceIndex) const;
    DeviceIndex indexOf(int row) const;
    ModularDeviceStatusModelPrivate::const_iterator constFind(int row) const;
};

int ModularDeviceStatusModelPrivate::rowOf(const DeviceIndex &deviceIndex) const
{
    return std::distance(begin(), find(deviceIndex));
}

DeviceIndex ModularDeviceStatusModelPrivate::indexOf(int row) const
{
    auto it = constFind(row);
    if(it == cend())
        return {};
    return it.key();
}

ModularDeviceStatusModelPrivate::const_iterator ModularDeviceStatusModelPrivate::constFind(int row) const
{
    int r = 0;
    auto it = cbegin();
    for (; it != cend(); ++it)
        if (r++ == row)
            return it;
    return it;
}


ModularDeviceStatusModel::ModularDeviceStatusModel(QObject *parent)
    : DeviceIndexTableModel(parent)
    , d_ptr(new ModularDeviceStatusModelPrivate())
{
}

ModularDeviceStatusModel::~ModularDeviceStatusModel()
{
    delete d_ptr;
}

DeviceIndex ModularDeviceStatusModel::deviceIndexOf(int row) const
{
    Q_D(const ModularDeviceStatusModel);
    return d->indexOf(row);
}

int ModularDeviceStatusModel::rowOf(const DeviceIndex &sourceIndex) const
{
    Q_D(const ModularDeviceStatusModel);
    return d->rowOf(sourceIndex);
}

void ModularDeviceStatusModel::invalidateRow(const DeviceIndex &index)
{
    int row = rowOf(index);
    if (row == -1)
        return;

    auto indexFrom = createIndex(row, 0);
    auto indexTo = createIndex(row, columnCount()-1);
    emit dataChanged(indexFrom, indexTo);
}

void ModularDeviceStatusModel::setStatus(const ModularDeviceStatus &status)
{
    Q_D(ModularDeviceStatusModel);
    const auto &deviceIndex = status.getDeviceIndex();
    auto it = d->find(deviceIndex);
    if (it == d->end())
        return;
    if (status == it->status)
        return;
    it->delta = status - it->status;
    it->status = status;
    invalidateRow(deviceIndex);
}

void ModularDeviceStatusModel::setMultiStatus(const QMap<DeviceIndex, ModularDeviceStatus> &statusMap)
{
    // TODO: optimize
    for (const auto &i : statusMap)
        setStatus(i);
}

static const double ResetModelRatioThreshold = 0; // always use resetModel

void ModularDeviceStatusModel::setDeviceList(const DeviceIndexEnabledMap &map)
{
    Q_D(ModularDeviceStatusModel);
    const auto &newSet = map.getEnabled();
    const auto &oldSet = makeDeviceIndexSet(d->keys());
    const auto toRemove = oldSet-newSet;
    const auto toAdd = newSet-oldSet;
    bool reset = false;
    reset |= (toRemove.size() >= d->size() * ResetModelRatioThreshold);
    reset |= (toAdd.size() >= d->size() * ResetModelRatioThreshold);
    if (reset) {
        beginResetModel();
        for(auto i : toRemove)
            d->remove(i);
        for(auto i : toAdd)
            d->insert(i, DeviceStatusBaseModelPrivateEntry(i));
        endResetModel();
    } else {
        for(auto i : toRemove) {
            removeRow(rowOf(i));
            d->remove(i);
        }
        for(auto i : toAdd) {
            d->insert(i, DeviceStatusBaseModelPrivateEntry(i));
            insertRow(rowOf(i));
        }
    }
}

int ModularDeviceStatusModel::rowCount(const QModelIndex &parent) const
{
    Q_D(const ModularDeviceStatusModel);
    return  d->size();
}

int ModularDeviceStatusModel::columnCount(const QModelIndex &parent) const
{
    return COL_DSB_TOTAL;
}

Qt::ItemFlags ModularDeviceStatusModel::flags(const QModelIndex & index) const
{
    auto f = QAbstractTableModel::flags(index);
    switch (index.column()) {
//    case COL_DSB_READOUT:
//    case COL_DSB_CHECK_WR:
//        return  f | Qt::ItemIsUserCheckable;
    default:
        return f;
    }
}

QVariant ModularDeviceStatusModel::data(const QModelIndex &index, int role) const
{
    Q_D(const ModularDeviceStatusModel);
    auto column = index.column();
    auto row = index.row();
    auto deviceIt = d->constFind(row);
    if (deviceIt == d->cend())
        return QVariant();

    const auto &device = *deviceIt;

    if(column>=COL_DSB_WR_START && column<=COL_DSB_WR_END) {
        const auto &status = device.status;
        return WrStatusTableHelper::data(status.wrStatus, index.column()-COL_DSB_WR_START, role);
    }

    switch (role) {
    // case Qt::EditRole: return column == COL_DSB_CHECK_WR;
    case Qt::BackgroundRole: return data_BackgroundRole(device, column);
    case Qt::DisplayRole: return data_DisplayRole(device, column);
    case Qt::UserRole: return data_UserRole(deviceIt.key(), column);
    case Qt::ToolTipRole: return data_ToolTipRole(device, column);
    }
    return QVariant();
}

QVariant ModularDeviceStatusModel::data_BackgroundRole(const DeviceStatusBaseModelPrivateEntry &device, int column) const
{
    const auto &status = device.status;
    const auto &delta = device.delta;
    switch (column) {
    case COL_DSB_BMC: if (status.getBmcSeverity() != Severity::Normal) return AppStyle::tableBg(status.getBmcSeverity());
    }
    if (status.desStatus.has_value()) {
        const auto &desStatus = status.desStatus.value();
        switch (column) {
        case COL_DES_STATUS: if (!desStatus.isOk()) return AppStyle::tableBg(Severity::Error);
        }
    }
    if (status.flashProgStatus.has_value()) {
        const auto &flashProgStatus = status.flashProgStatus.value();
        auto severity = flashProgStatus.status();
        switch (column) {
        case COL_DSB_FLASH: if (severity != Severity::Normal) return AppStyle::tableBg(severity);
        }
    }
    if (status.runLogicStatus.has_value()) {
        const auto &value = status.runLogicStatus.value();
        switch (column) {
        case COL_DSB_RUNLOGIC: if (value.isRunning()) return AppStyle::tableBg(Severity::Warning);
        }
    }
    if (status.runStatistic.has_value()) {
        const auto &runStatistic = delta.runStatistic.value();
        switch (column) {
        case COL_TRIG_BLOCK: if (runStatistic.getCounter(RunStatistic::Blocked)) return AppStyle::tableBg(Severity::Warning);
        }
    }
    if (status.tdcCoreStatus.has_value()) {
        const auto &tdcCoreStatus = status.tdcCoreStatus.value();
        switch (column) {
        case COL_HPTDC_STATUS: {
            const auto sev = tdcCoreStatus.getHptdcSeverity();
            if (sev != Severity::Normal) return AppStyle::tableBg(sev);
        }
        }
    }
    if (status.waveRec.has_value()) {
        const auto &value = status.waveRec.value();
        switch (column) {
        case COL_WAVEREC: if (value.hasErrors()) return AppStyle::tableBg(Severity::Error);
        }
    }
    return {};
}

QVariant ModularDeviceStatusModel::data_DisplayRole(const DeviceStatusBaseModelPrivateEntry &device, int column) const
{
    const auto &status = device.status;
    const auto &delta = device.delta;

    switch (column) {
    case COL_DSB_INDEX: return status.getDeviceName();
    case COL_DSB_SERIAL: return status.getSerialStr();
    case COL_DSB_FW: return status.getFwStr();
    case COL_DSB_BMC: return status.getBmcVersionStr();
    case COL_DSB_TEMP: return TempsInfo(status.getTemps()).brief();
    }
    if (status.clockStatus.has_value()) {
        const auto &clockStatus = status.clockStatus.value();
        switch (column) {
        case COL_CLOCK_STATUS: return QString("%L1").arg(clockStatus.getSelectedClockStr());
        }
    }
    if (status.desStatus.has_value()) {
        const auto &desStatus = status.desStatus.value();
        switch (column) {
            case COL_DES_STATUS: return desStatus.isOk()
                    ? "Ok" : QString("%L1").arg(desStatus.toString());
        }
    }
    if (status.flashProgStatus.has_value()) {
        const auto &flashProgStatus = status.flashProgStatus.value();
        switch (column) {
        case COL_DSB_FLASH: return QString("%L1").arg(flashProgStatus.toString());
        }
    }
    if (status.runLogicStatus.has_value()) {
        const auto &value = status.runLogicStatus.value();
        switch (column) {
        case COL_DSB_RUNLOGIC: return QString("%L1").arg(value.getStateStr());
        }
    }
    if (status.runStatistic.has_value()) {
        const auto &st = status.runStatistic.value();
        switch (column) {
        case COL_TRIG_ACCEPT: return QString("%L1").arg(st.getCounter(RunStatistic::Accepted));
        case COL_TRIG_BLOCK: return QString("%L1").arg(st.getCounter(RunStatistic::Blocked));
        case COL_TRIG_DISCARD: return QString("%L1").arg(st.getCounter(RunStatistic::Discarded));
        case COL_TRIG_VETO: return QString("%L1").arg(st.getCounter(RunStatistic::Vetoed));
        }
        const auto &runStatistic = delta.runStatistic.value();
        if (runStatistic.isValid()) {
            switch (column) {
            case COL_DEV_BUSY_TIME: return runStatistic.getText(RunStatistic::Brief);
            case COL_TRIG_ACCEPT_D: return QString("%L1").arg(runStatistic.getRate(RunStatistic::Accepted));
            case COL_TRIG_BLOCK_D: return QString("%L1").arg(runStatistic.getRate(RunStatistic::Blocked));
            case COL_TRIG_DISCARD_D: return QString("%L1").arg(runStatistic.getRate(RunStatistic::Discarded));
            case COL_TRIG_VETO_D: return QString("%L1").arg(runStatistic.getRate(RunStatistic::Vetoed));
            }
        }
    }
    if (status.tdcCoreStatus.has_value()) {
        const auto &tdcCoreStatus = status.tdcCoreStatus.value();
        switch (column) {
        case COL_HPTDC_STATUS: return QString("%L1").arg(tdcCoreStatus.getSummary());
        }
    }
    if (status.trigStatus.has_value()) {
        const auto &trigStatus = status.trigStatus.value();
        if (trigStatus.isValid()) {
            switch (column) {
            case COL_TRIG_QUEUE: return QString("%1").arg(trigStatus.trigQueue);
            case COL_DEV_BUSY_BITS: return QString("%1").arg(trigStatus.busyStatus, 4, 2, QChar('0'));
            }
        }
    }
    if (status.trigInfo.has_value()) {
        const auto &triggerInfo = status.trigInfo.value();
        if (triggerInfo.isValid()) {
            switch (column) {
            case COL_EV_NUM: return QString("%L1").arg(triggerInfo.eventNumber);
            case COL_EV_TS: return QString("%1").arg(triggerInfo.timestamp().toString());
            }
        }
    }
//    if (status.triggerCounters.has_value()) {
//        const auto &triggerCounters = status.triggerCounters.value();
//        switch (column) {
//        case COL_TRIG_ACCEPT: return QString("%L1").arg(triggerCounters.evNumAccepted);
//        case COL_TRIG_BLOCK: return QString("%L1").arg(triggerCounters.blocked());
//        case COL_TRIG_DISCARD: return QString("%L1").arg(triggerCounters.discarded());
//        }
//    }
    if (status.waveRec.has_value()) {
        const auto &value = status.waveRec.value();
        switch (column) {
        case COL_WAVEREC: return QString("%L1").arg(value.getSummary());
        }
    }
    return {};
}

QVariant ModularDeviceStatusModel::data_UserRole(const DeviceIndex &index, int column) const
{
    switch (column) {
    case COL_DSB_INDEX: return QVariant::fromValue(index);
    default: return {};
    }
}

QVariant ModularDeviceStatusModel::data_ToolTipRole(const DeviceStatusBaseModelPrivateEntry &device, int column) const
{
    const auto &status = device.status;
    const auto &delta = device.delta;

    switch (column) {
    case COL_DSB_INDEX: return "0x" + QString("%1").arg(status.getDeviceIndex().getDevId(), 2, 16, QChar('0')).toUpper();
    case COL_DSB_FW: return status.getFwText();
    case COL_DSB_BMC: return  status.getBmcText();
    case COL_DSB_TEMP: return TempsInfo(status.getTemps()).full();
    case COL_HPTDC_STATUS:
    {
        auto &st = status.tdcCoreStatus;
        if(st)
            return st->getSummary();
        break;
    }
    case COL_DEV_BUSY_BITS: if(status.trigStatus) return BUSY_TOOLTIP; break;
    case COL_DEV_BUSY_TIME:
    {
        QStringList list;
        if (status.runStatistic) {
            list << "-- Full --";
            list << status.runStatistic->getText(RunStatistic::Full);
        }
        if (delta.runStatistic) {
            list << "" << "-- Delta --";
            list << delta.runStatistic->getText(RunStatistic::Full);
        }
        return list.join('\n');
    }
    }
    if (status.runLogicStatus.has_value()) {
        const auto &value = status.runLogicStatus.value();
        switch (column) {
        case COL_DSB_RUNLOGIC: return QString("%L1").arg(value.getFullText());
        }
    }
    if (status.waveRec.has_value()) {
        const auto &value = status.waveRec.value();
        switch (column) {
        case COL_WAVEREC: return QString("%L1").arg(value.getFullText());
        }
    }
    return {};
}

QVariant ModularDeviceStatusModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    Q_D(const ModularDeviceStatusModel);
    if (orientation == Qt::Horizontal) {
        if (section>=COL_DSB_WR_START && section<=COL_DSB_WR_END)
            return WrStatusTableHelper::headerData(section-COL_DSB_WR_START, role);
        if (role == Qt::DisplayRole) {
            switch (static_cast<Columns>(section)) {
            case COL_DSB_INDEX: return tr("Device Type");
            case COL_DSB_SERIAL: return tr("Serial");
            case COL_DSB_FW: return tr("FPGA f/w");
            case COL_DSB_FLASH: return tr("Flash");
            case COL_DSB_RUNLOGIC: return tr("Run");
            case COL_DSB_BMC: return tr("BMC f/w");
            case COL_DSB_TEMP: return tr("Temp");
            case COL_CLOCK_STATUS: return "Clock";
            case COL_DES_STATUS: return "Deser.";
            case COL_EV_NUM: return "Ev. #";
            case COL_EV_TS: return "Ev. TS";
            case COL_TRIG_ACCEPT: return "Tr.accepted";
            case COL_TRIG_BLOCK: return "Tr. blocked";
            case COL_TRIG_DISCARD: return tr("Tr. discarded");
            case COL_TRIG_VETO: return tr("Tr. veto");
            case COL_TRIG_ACCEPT_D: return "Tr.accepted (dt)";
            case COL_TRIG_BLOCK_D: return "Tr. blocked (dt)";
            case COL_TRIG_DISCARD_D: return tr("Tr. discarded (dt)");
            case COL_TRIG_VETO_D: return tr("Tr. veto (dt)");
            case COL_TRIG_QUEUE: return "Tr. queue";
            case COL_DEV_BUSY_BITS: return "Busy";
            case COL_DEV_BUSY_TIME: return "Busy Time";
            case COL_HPTDC_STATUS: return "HPTDC St.";
                //    case COL_DSB_BUSY: return tr("Busy");
                //    case COL_DSB_IDLE: return tr("Idle");
                //    case COL_DSB_OCCUPANCY: return tr("Occupancy");
            case COL_WAVEREC: return "WaveRec";
            default: return {};
            }
        }
    }
    if (orientation == Qt::Vertical && role == Qt::DisplayRole) {
        const auto &it = d->constFind(section);
        if (it == d->end())
            return {};
        const auto &status = it->status;
        return status.getDeviceIdent();
    }
    return DeviceIndexTableModel::headerData(section, orientation, role);
}

//bool DeviceStatusBaseModel::setData(const QModelIndex &index, const QVariant &value, int role)
//{
//    auto column = index.column();
//    auto row = index.row();
//    auto deviceIndex = devices->indexOf(row);
//    if (deviceIndex.isNull())
//        return false;

//    auto &device = (*devices)[deviceIndex];
//    auto &config = device.config;

//    if (role != Qt::CheckStateRole)
//        return false;

//    switch (column) {
//    case COL_DSB_READOUT:
//        config.evNumReadReset = (value == Qt::Checked);
//        break;
//    case COL_DSB_CHECK_WR:
//        config.checkWr = (value == Qt::Checked);
//        break;
//    default:
//        return false;
//    }

//    emit dataChanged(index, index);
//    emit trigConfigChanged(deviceIndex, config);
//    return true;
//}

QMap<int, bool> ModularDeviceStatusModel::defaultColumnState() const
{
    QMap<int, bool> columns =
        {
            {COL_EV_TS, false},
            {COL_TRIG_DISCARD, false},
            {COL_TRIG_VETO, false},
            {COL_TRIG_ACCEPT_D, false},
            {COL_TRIG_BLOCK_D, false},
            {COL_TRIG_DISCARD_D, false},
            {COL_TRIG_VETO_D, false},
            {COL_HPTDC_STATUS, false},
            {COL_TRIG_QUEUE, false},
            {COL_DEV_BUSY_BITS, false}
        };
    const auto &wrColumns = WrStatusTableHelper::hiddenColumns();
    for (const auto &col : wrColumns)
        columns[(COL_DSB_WR_START + (int)col)] = false;

    const auto daqModule = Globals::getDaqModuleType();
    const bool isFlashProg = daqModule == DaqModule::Type::DeviceFlashProg;
    columns[COL_DSB_FLASH] = isFlashProg;
    if(isFlashProg) {
        for(auto c : {COL_DSB_RUNLOGIC, COL_CLOCK_STATUS,
            COL_TRIG_ACCEPT, COL_TRIG_BLOCK,
            COL_DEV_BUSY_TIME,
            COL_EV_NUM, COL_WAVEREC}) {
            columns[c] = false;
        }
        for(int i=COL_DSB_WR_START;i<COL_DSB_WR_END; ++i)
            columns[i] = false;
    }
    columns[COL_DES_STATUS] = DaqModuleTraits::hasSerDes(daqModule);
    columns[COL_HPTDC_STATUS] = DaqModuleTraits::hasHptdc(daqModule);

    return columns;
}
