#include "SpillControlWidget.h"
#include "ui_SpillControlWidget.h"

#include <QComboBox>
#include <QDebug>
#include <QLabel>
#include <QSizePolicy>

#include "base/Globals.h"
#include "device-modules/SpillControl/SpillStatus.h"

SpillControlWidget::SpillControlWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::SpillControlWidget)
{
    ui->setupUi(this);

    for(auto sb : {ui->spinBoxSpillDurPhase1,
        ui->spinBoxSpillDurPhase2,
        ui->spinBoxSpillDurPhase3,
        ui->spinBoxSpillDurPhase4,
        ui->spinBoxSpillDurPhase5}) {
        sb->setMaximum(4000000);
        sb->setSingleStep(100);
        sb->setSuffix(" ms");
    }
    ui->spinBoxSpillDurPhase3->setMinimum(100); // 100ms for Readout as mininum
    ui->radioButtonSpillSourceWr->hide();
    ui->lineEditPeriodWr->hide();

    updatePhasesVisibility();
}

SpillControlWidget::~SpillControlWidget()
{
    delete ui;
}

void SpillControlWidget::updateRemoteBlocking(bool remoteConnected) const
{
    ui->groupBoxSoftSpill->setEnabled(!remoteConnected);
}

void SpillControlWidget::on_groupBoxSoftSpill_toggled(bool arg1)
{
    cfg.softSpill = arg1;
    emit spillCfgChanged(cfg);
}

void SpillControlWidget::on_spinBoxPeriodTimer_valueChanged(int arg1)
{
    cfg.timerPeriod = static_cast<uint32_t>(arg1);
    emit spillCfgChanged(cfg);
}

void SpillControlWidget::on_radioButtonSpillSourceAuto_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_AUTO);
}

void SpillControlWidget::on_radioButtonSpillSourceTtl_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_TTL);
}

void SpillControlWidget::on_radioButtonSpillSourceWr_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_WR);
}

void SpillControlWidget::on_radioButtonSpillSourceTimer_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_TIMER);
}

void SpillControlWidget::on_radioButtonSpillSourceGate_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_GATE);
}

void SpillControlWidget::on_radioButtonSpillSourceInfinite_toggled(bool checked)
{
    if (!checked) return;
    setSpillMode(SpillConfig::SPILL_INF);
}

void SpillControlWidget::setSpillMode(const SpillConfig::spill_sel_t &sel)
{
    cfg.source = sel;
    ui->groupBoxSpillPhases->setVisible(!cfg.isSpillGate());
    emit spillCfgChanged(cfg);
}

void SpillControlWidget::on_spinBoxSpillDurPhase1_valueChanged(int arg1)
{
    if(blockSignal) return;
    cfg.durPhase1 = static_cast<uint32_t>(arg1);
    emit spillCfgChanged(cfg);
}

void SpillControlWidget::on_spinBoxSpillDurPhase2_valueChanged(int arg1)
{
    if(blockSignal) return;
    cfg.durPhase2 = static_cast<uint32_t>(arg1);
    emit spillCfgChanged(cfg);
}

void SpillControlWidget::on_spinBoxSpillDurPhase3_valueChanged(int arg1)
{
    if(blockSignal) return;
    cfg.durPhase3 = static_cast<uint32_t>(arg1);
    emit spillCfgChanged(cfg);
}

void SpillControlWidget::on_spinBoxSpillDurPhase4_valueChanged(int arg1)
{
    if(blockSignal) return;
    cfg.durPhase4 = static_cast<uint32_t>(arg1);
    emit spillCfgChanged(cfg);
}

void SpillControlWidget::on_spinBoxSpillDurPhase5_valueChanged(int arg1)
{
    if(blockSignal) return;
    cfg.durPhase5 = static_cast<uint32_t>(arg1);
    emit spillCfgChanged(cfg);
}

void SpillControlWidget::restoreGuiFromConfig()
{
    // trigChannelConfig->updateGui();

    bool oldState1 = ui->spinBoxSpillDurPhase4->blockSignals(true);
    bool oldState2 = ui->spinBoxPeriodTimer->blockSignals(true);
    bool oldState3 = ui->radioButtonSpillSourceAuto->blockSignals(true);

    ui->spinBoxSpillDurPhase1->setValue(static_cast<int>(cfg.durPhase1));
    ui->spinBoxSpillDurPhase2->setValue(static_cast<int>(cfg.durPhase2));
    ui->spinBoxSpillDurPhase3->setValue(static_cast<int>(cfg.durPhase3));
    ui->spinBoxSpillDurPhase4->setValue(static_cast<int>(cfg.durPhase4));
    ui->spinBoxSpillDurPhase5->setValue(static_cast<int>(cfg.durPhase5));
    updatePhasesVisibility();

    ui->groupBoxSoftSpill->setChecked(cfg.softSpill);
    ui->spinBoxPeriodTimer->setValue(static_cast<int>(cfg.timerPeriod));

    ui->radioButtonSpillSourceAuto->setChecked(cfg.source == SpillConfig::SPILL_AUTO);
    ui->radioButtonSpillSourceTimer->setChecked(cfg.source == SpillConfig::SPILL_TIMER);
    ui->radioButtonSpillSourceTtl->setChecked(cfg.source == SpillConfig::SPILL_TTL);
    ui->radioButtonSpillSourceWr->setChecked(cfg.source == SpillConfig::SPILL_WR);
    ui->radioButtonSpillSourceGate->setChecked(cfg.source == SpillConfig::SPILL_GATE);
    ui->radioButtonSpillSourceInfinite->setChecked(cfg.source == SpillConfig::SPILL_INF);

    ui->spinBoxSpillDurPhase4->blockSignals(oldState1);
    ui->spinBoxPeriodTimer->blockSignals(oldState2);
    ui->radioButtonSpillSourceAuto->blockSignals(oldState3);
}

void SpillControlWidget::displaySpillSource(const std::optional<SpillStatus> &st)
{
    if (st && st->ttlLastPeriodMs > 0) {
        ui->lineEditPeriodTtl->setText(QString("%1 ms").arg(st->ttlLastPeriodMs));
    } else {
        ui->lineEditPeriodTtl->clear();
    }

    if (st && st->wrLastPeriodMs > 0) {
        ui->lineEditPeriodWr ->setText(QString("%1 ms").arg(st->wrLastPeriodMs));
    } else {
        ui->lineEditPeriodWr->clear();
    }

    QString spillSourceStr;
    if (st) {
        switch (st->status.reg.bit.sel) {
        case SpillConfig::SPILL_OFF:   spillSourceStr = "OFF"; break;
        case SpillConfig::SPILL_TIMER: spillSourceStr = "Timer"; break;
        case SpillConfig::SPILL_TTL:   spillSourceStr = "TTL"; break;
        case SpillConfig::SPILL_WR:    spillSourceStr = "WR"; break;
        case SpillConfig::SPILL_GATE:   spillSourceStr = "Gate"; break;
        default: spillSourceStr = "AUTO"; break;
        }
    }
    ui->labelSpillSourceSelected->setEnabled(cfg.source == SpillConfig::SPILL_AUTO);
    ui->labelSpillSourceSelected->setText(spillSourceStr);
    if (st) {
    bool spillSourceOk = true;
        if (st->status.reg.bit.sel == SpillConfig::SPILL_TTL
        && !st->status.reg.bit.ttl_ok) spillSourceOk = false;
        if (st->status.reg.bit.sel == SpillConfig::SPILL_WR
            && !st->status.reg.bit.wr_ok) spillSourceOk = false;
    ui->labelSpillSourceSelected->setStyleSheet(spillSourceOk
                                                ? "background-color: green; color: white"
                                                : "background-color: red; color: white");
    } else {
        ui->labelSpillSourceSelected->setStyleSheet({});
    }
}

void SpillControlWidget::updatePhasesVisibility() const
{
    auto l = ui->gridLayoutSpillPhases;
    constexpr int static const COL_SB = 1;
    for(auto r=0; r<l->rowCount(); ++r) {
        auto sb = qobject_cast<QSpinBox*>(l->itemAtPosition(r, COL_SB)->widget());
        if(!sb) {
            Q_ASSERT(sb);
            continue;
        }
        if(sb == ui->spinBoxSpillDurPhase3 || sb == ui->spinBoxSpillDurPhase5)
            continue; // Reaout and Flush phases are always visible
        bool visible = sb->value() || showAllPhases;
        for(auto c=0; c<l->columnCount(); ++c) {
            auto w = l->itemAtPosition(r, c)->widget();
            if(w)
                w->setVisible(visible);
        }
    }
}

void SpillControlWidget::setSpillCfg(const SpillConfig &newCfg)
{
    if(cfg == newCfg)
        return;
    cfg = newCfg;

    restoreGuiFromConfig();
}

void SpillControlWidget::onFsmStateChange(const FsmState &state)
{
    bool en = state==FsmState::Ready || state==FsmState::Error
            || Globals::getDebugMode();
    setEnabled(en);
}

void SpillControlWidget::setSpillStatus(const std::optional<SpillStatus> &st)
{
    const static QString activeStyle = "background-color: green; color:white";
    const static QString passiveStyle;

    QVector<QLabel *> labelList = {
        ui->labelSpillPhase0,
        ui->labelSpillPhase1,
        ui->labelSpillPhase2,
        ui->labelSpillPhase3,
        ui->labelSpillPhase4,
        ui->labelSpillPhase5
    };
    QLabel *curLab = nullptr;
    ui->horizontalFrameSpillPhases->setEnabled(st.has_value());
    if (st) {
        if (st->status.reg.bit.phase) {
            int phase = static_cast<int>(std::log2(st->status.reg.bit.phase))+1;
            if(labelList.size() > phase)
                curLab = labelList[phase];
        } else {
            curLab = ui->labelSpillPhase0;
        }
    }

    for (auto l : qAsConst(labelList)) {
        l->setStyleSheet(l == curLab ? activeStyle : passiveStyle);
        l->setEnabled(l == curLab);
    }

    displaySpillSource(st);
}

void SpillControlWidget::setShowAllPhases(bool show)
{
    showAllPhases = show;
    updatePhasesVisibility();
}
