#include "FlashDev.h"

#include <chrono>
#include <cstdio>
#include <cstring>
#include <iostream>
#include <map>
#include <thread>

#include <QDebug>
#include <QFuture>
#include <QLoggingCategory>

#include "FlashDevTypes.h"

constexpr const int MAX_PROGRAM_RETRY_COUNT = 10;
constexpr const int MIN_MATCH_COUNT = 3;
constexpr const int MAX_VERIFY_RETRY_COUNT = 2;

using namespace std::chrono;

namespace {
Q_DECLARE_LOGGING_CATEGORY(flash)
Q_LOGGING_CATEGORY(flash, "afi.flash")
}

namespace {
std::string formatTime(int64_t ms)
{
    std::string strStr;

    int64_t s = ms / 1000;
    if (ms == 0 || (ms %= 1000) != 0) {
        strStr += std::to_string(ms) + "ms";
    }

    int64_t min = s / 60;
    s %= 60;
    if (s)
        strStr = std::to_string(s) + "sec " + strStr;
    if (min)
        strStr = std::to_string(min) + "min " + strStr;

    return strStr;
}

int64_t elapsedMs(steady_clock::time_point tStart)
{
    auto tNow = steady_clock::now();
    int64_t int_ms = duration_cast<milliseconds>(tNow - tStart).count();
    return int_ms;
}
} // namespace

FlashDev::FlashDev(QObject *parent)
    : QObject(parent)
{
    qRegisterMetaType<FlashDevTask>();
}

QString FlashDev::cmdStatusStr(CmdStatus value)
{
    switch (value) {
    case CmdStatus::Ok: return "ok";
    case CmdStatus::Error: return "I/O error";
    case CmdStatus::Timeout: return "timeout";
    }
    return {};
}

QString FlashDev::commandStr(FlashCommand value)
{
    switch (value) {
    case CMD_WRITE: return "write";
    case CMD_READ: return "read";
    case CMD_SE: return "sector erase";
    case CMD_BE: return "bulk erase";
    case CMD_WRITE_EN: return "write enable";
    case CMD_WRVECR: return "write VECR";
    case CMD_RDSR: return "read SR";
    case CMD_RDID: return "read ID";
    case CMD_ADDR4B_EN: return "enable 4-byte addres";
    case CMD_ADDR4B_DIS: return "disable 4-byte addres";
    }
    return {};
}

bool FlashDev::setState(State newState)
{
    state = newState;
    return true;
}

bool FlashDev::abortTask()
{
    return false; // TODO
}

bool FlashDev::isReady() const
{
    return flashIdData.isValid() && !task.has_value();
}

std::optional<FlashDevTask> FlashDev::getTasks() const
{
    return task;
}

std::optional<FlashDevTask> FlashDev::getFinishedTasks() const
{
    return finishedTask;
}

bool FlashDev::runTasks()
{
    if (!task.has_value())
        return false;
    switch (task->op) {
    case FlashOp::Verify: verifyFlashStep(); break;
    case FlashOp::Write: writeFlashStep(); break;
    case FlashOp::Erase: eraseFlashStep(); break;
    }

    if (task->isAborted()) {
        qWarning(flash).noquote() << task->taskStatusStr();
    }
    if (task->isComplete()) {
        qWarning(flash).noquote() << task->taskStatusStr();
    }
    // remove aborted/complete task
    if (!task->isRunning()) {
        emit taskFinished(*task);
        finishedTask = task;
        task.reset();
    }
    return true;
}

void FlashDev::reverseByteOrder(std::vector<uint16_t> &dataVec)
{
    uint16_t vTmp;
    for (uint32_t i = 0; i < dataVec.size(); i += 2) {
        uint16_t &v1 = dataVec[i];
        uint16_t &v2 = dataVec[i + 1];

        vTmp = ((v2 & 0xFF) << 8) | (v2 >> 8);
        v2 = ((v1 & 0xFF) << 8) | (v1 >> 8);
        v1 = vTmp;
    }
}

FlashDev::CmdStatus FlashDev::waitCmdExecution(int timeoutMs, bool extraWaitCheck)
{
    if (timeoutMs == 0)
        timeoutMs = WRITE_TIMEOUT_MS;
    auto commandTimer = steady_clock::now();
    while (true) {
        uint16_t readBackVal;
        if (!devReadReg(REG_READBACK, &readBackVal))
            return CmdStatus::Error;
        if ((readBackVal >> 14) == 0 && (!extraWaitCheck || (readBackVal & 0xFF) == (useAddr4bMode ? 0x81 : 0x80)))
            break;
        auto elapsedTime = elapsedMs(commandTimer);
        if (elapsedTime > timeoutMs) {
            return CmdStatus::Timeout;
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(1));
    }
    return CmdStatus::Ok;
}

FlashDev::CmdStatus FlashDev::execCmd(FlashCommand cmd, int timeoutMs, bool extraWaitCheck)
{
    if (timeoutMs == 0)
        timeoutMs = WRITE_TIMEOUT_MS;
    ++cmdPerformed;
    CmdStatus ret;
    if (!devWriteReg(REG_CMD, cmd)) {
        ret = CmdStatus::Error;
        goto err;
    }
    ret = waitCmdExecution(std::min(1000, timeoutMs * (2 ^ attemptFactor)), extraWaitCheck);
    if (ret != CmdStatus::Ok)
        goto err;
    return ret;
err:
    qWarning(flash).noquote() << QString("Flash command %1 %2").arg(commandStr(cmd)).arg(cmdStatusStr(ret));
    return ret;
}

void FlashDev::configureFlash(FlashIdData *result, bool supportAddr4bMode)
{
    setState(State::Init);
    useAddr4bMode = false;
    // Configure flash
    // SPI single, default drive strength (see flash manual)
    if (!devWriteReg(REG_FLASH_ADDR, 0xef))
        return;
    if (!devWriteReg(REG_CMD, CMD_WRITE_EN))
        return;
    for (const auto &cmd: {CMD_WRITE_EN, CMD_WRVECR, CMD_RDID}) {
        if (execCmd(cmd) != CmdStatus::Ok)
            return;
    }

    if (!updateFlashIdData())
        return;
    if (flashIdData.isValid()) {
        if(supportAddr4bMode && (getFlashSize() > (1<<24))) {
            useAddr4bMode = true;
            if (!devWriteReg(REG_CTRL, REG_CTRL_BIT_ADDR4B_EN))
                return;

            for (const auto &cmd: {CMD_WRITE_EN, CMD_ADDR4B_EN}) {
                if (execCmd(cmd) != CmdStatus::Ok)
                    return;
            }
        } else {
            if (!devWriteReg(REG_CTRL, 0))
                return;

            for (const auto &cmd: {CMD_WRITE_EN, CMD_ADDR4B_DIS}) {
                if (execCmd(cmd) != CmdStatus::Ok)
                    return;
            }
        }
    } else {
        qWarning().noquote() << "[Configure flash] Invalid flash id data";
        return;
    }

    if (result)
        *result = flashIdData;
    setState(State::Ready);
}

bool FlashDev::eraseFlashStep()
{
    Q_ASSERT(task);
    const auto address = task->startAddr() + task->pos.offset();
    if (!singleSectorErase(address)) {
        task->abort();
        return false;
    }
    task->advance(flashIdData.getSectorSize());
    return true;
}

bool FlashDev::submitTask(const FlashDevTask &newTask)
{
    if (state != State::Ready || task.has_value()) {
        qWarning(flash) << "not ready";
        return false;
    }
    finishedTask.reset();
    task = newTask;
    task->setUseAddr4bMode(useAddr4bMode);
    uint32_t size = 0;
    switch (task->op) {
    case FlashOp::Erase:
        size = task->getMaxImageSize();
        break;
    case FlashOp::Write:
    case FlashOp::Verify:
        size = task->imageSize();
        break;
    }
    task->start(0, size);
    return true;
}

bool FlashDev::singleSectorErase(uint32_t pos)
{
    if (!devWriteReg32(REG_FLASH_ADDR, pos))
        return false;
    if (execCmd(CMD_WRITE_EN) != CmdStatus::Ok) {
        return false;
    }
    if (execCmd(CMD_SE, SECTOR_ERASE_TIMEOUT, true) != CmdStatus::Ok) {
        return false;
    }
    if (!read_rdsr(nullptr)) // delay after sector erase
        return false;

    return true;
}

bool FlashDev::singleBlockWrite(uint32_t pos, std::vector<uint16_t> blockData)
{
    reverseByteOrder(blockData);

    const auto &count = blockData.size();
    std::vector<uint32_t> addr(count);
    for (uint32_t w = 0; w < count; ++w)
        addr[w] = REG_DATA + w;
    if (!devBlkWriteReg(count, addr.data(), blockData.data()))
        return false;

    if (!devWriteReg32(REG32_DATA_ADDR, pos))
        return false;

    if (execCmd(CMD_WRITE_EN) != CmdStatus::Ok) {
        return false;
    }
    if (execCmd(CMD_WRITE, 0, true) != CmdStatus::Ok) {
        return false;
    }
    if (!read_rdsr(nullptr)) // delay after write
        return false;

    return true;
}

bool FlashDev::singleSectorProgram(const uint32_t startPos,
                                          const std::vector<uint16_t> &sectorData)
{
    const auto blockSize = flashIdData.getBlockSize();
    std::vector<uint16_t> readbackData;
    std::vector<uint16_t> blockData(blockSize / sizeof(uint16_t), 0);

    for (int attempt = 0; attempt < MAX_PROGRAM_RETRY_COUNT; ++attempt) {
        attemptFactor = attempt;
        static auto maxAattemptFactor = 0;
        if (attemptFactor > maxAattemptFactor) {
            qDebug() << "maxAattemptFactor updated:" << maxAattemptFactor;
            maxAattemptFactor = attemptFactor;
        }
        if (!singleSectorErase(startPos))
            continue;

        auto sizeToWrite = uint32_t(sectorData.size()) * sizeof(uint16_t);
        uint32_t wroteSize = 0;
        while (wroteSize < sizeToWrite) {
            const auto pos = startPos + wroteSize;
            memcpy(blockData.data(), sectorData.data() + wroteSize / sizeof(uint16_t), blockSize);

            if (!singleBlockWrite(pos, blockData))
                break;

            int rightCnt = singleCheckMode ? 1 : 2;
            readbackData = singleBlockRead(pos, 5, rightCnt);
            if (readbackData != blockData) {
                for (size_t k = 0; k < blockData.size(); ++k) {
                    if (readbackData[k] != blockData[k]) {
                        qDebug().noquote() << QString("err on wroteSize:%1 pos:%2 (readback=%3; orig=%4)")
                                                  .arg(wroteSize)
                                                  .arg(k)
                                                  .arg(readbackData[k])
                                                  .arg(blockData[k]);
                    }
                }
                break;
            }

            wroteSize += blockSize;
        }
        attemptFactor = 0;
        if (wroteSize == sizeToWrite)
            return true;
    }
    attemptFactor = 0;

    return false;
}

std::vector<uint16_t> FlashDev::singleBlockRead(uint32_t pos)
{
    auto readSize = flashIdData.getBlockSize();
    auto end = pos + readSize;
    Q_ASSERT(end <= flashIdData.getFlashSizeBytes());
    if (end > flashIdData.getFlashSizeBytes()) {
        qWarning().noquote() << __func__ << "address out of range:" << pos;
        return {};
    }
    const size_t wordCount = readSize / sizeof(uint16_t); // in 16-bit words
    std::vector<uint16_t> data(wordCount);

    devWriteReg32(REG_FLASH_ADDR, pos);
    if (execCmd(CMD_READ, READ_TIMEOUT_MS) != CmdStatus::Ok) {
        return {};
    }
    ++cmdPerformed;

    uint32_t addr[wordCount];
    for (uint32_t w = 0; w < wordCount; ++w)
        addr[w] = REG_DATA + w;

    bool blkReadOk = devBlkReadReg(wordCount, addr, data.data());
    if (!blkReadOk)
        return {};

    return data;
}

std::vector<uint16_t> FlashDev::singleBlockRead(uint32_t pos, int retries, int rightCnt)
{
    std::map<std::vector<uint16_t>, int> map;
    for (int i = 0; i < retries; ++i) {
        const auto &v = singleBlockRead(pos);
        if (v.empty())
            return {};
        int cnt = ++map[v];
        if (cnt >= rightCnt) {
            static auto maxBlockRead = 0;
            if (i > maxBlockRead) {
                qDebug().noquote() << "maxBlockRead updated:" << i;
                maxBlockRead = i;
            }
            return v;
        }
    }
    return {};
}

bool FlashDev::bulkErase()
{
    qInfo() << "Start bulk erase...";
    bool allOk = true;

    if (execCmd(CMD_WRITE_EN) != CmdStatus::Ok) {
        allOk = false;
    }

    if (allOk) {
        if (execCmd(CMD_BE, BULK_ERASE_TIMEOUT) != CmdStatus::Ok) {
            allOk = false;
        }
    }

    qInfo() << QString("Bulk erase %1").arg(allOk ? "finished" : "failed");
    return allOk;
}

bool FlashDev::writeFlashStep()
{
    Q_ASSERT(task);
    const auto &sectorData = task->getSectorData();
    Q_ASSERT(!sectorData.empty());
    Q_ASSERT(sectorData.size() <= 0x10000); // remove me

    const auto address = task->startAddr() + task->pos.offset();
    bool ok;
    if (task->pos.offset() || task->pos.isWritingFirstSector()) {
        ok = singleSectorProgram(address, sectorData);
    } else {
        ok = singleSectorErase(address);
    }
    if(!ok) {
        task->abort();
        return false;
    }

    const auto bytes = sectorData.size() * sizeof(sectorData[0]);
    task->advance(bytes);
    return true;
}

void FlashDev::dumpVerify(const std::vector<uint16_t> &original, const std::vector<uint16_t> &readback) const
{
    QStringList sl;
    sl << "Error block:";
    sl << "0x<blockOffset>+<blockAddr>: <original> ? <readback>";
    for (uint32_t w = 0; w < std::min(original.size(), readback.size()); w++) {
        const auto &v1 = original[w];
        const auto &v2 = readback[w];
        sl << QString("%1+%2: %3 %4 %5")
                  .arg(task->pos.offset(), 4, 16, QChar('0'))
                  .arg(w, 4, 16, QChar('0'))
                  .arg(v1, 4, 16, QChar('0'))
                  .arg(v1 == v2 ? "==" : "!=")
                  .arg(v2, 4, 16, QChar('0'));
    }
    qDebug().noquote() << sl.join("\n");
}

bool FlashDev::verifyFlashStep()
{
    Q_ASSERT(task);
    const auto &originalData = task->getBlockData();

    for (int i = 0; ; i++) {
        if (i > 0) {
            qDebug().noquote()
                << QString("Verify failed at %1, retry %2").arg(task->getProgressStr()).arg(i);
        }
        const auto address = task->startAddr() + task->pos.offset();
        const auto &readbackData = singleBlockRead(address);
        if (readbackData.empty()) {
            task->abort();
            return false;
        }
        if (originalData == readbackData) {
            break;
        }
        if (i >= MAX_VERIFY_RETRY_COUNT) {
            dumpVerify(originalData, readbackData);
            task->abort();
            return false;
        }
    }

    const auto readBytes = originalData.size() * sizeof(originalData[0]);
    task->advance(readBytes);
    return true;
}

void FlashDev::readStat()
{
    uint16_t statVal1;
    devReadReg(REG_STAT_1, &statVal1);
    uint16_t statVal2;
    devReadReg(REG_STAT_2, &statVal2);

    devWriteReg(REG_CMD, CMD_RDSR);
    uint16_t val;
    devReadReg(REG_READBACK, &val);

    qInfo().noquote() << QString("Statistics: cmdPerformed=%1 reg(%2)=%3; reg(%4)=%5; RDSR=%6")
                             .arg(cmdPerformed)
                             .arg(REG_STAT_1, 4, 16, QChar('0'))
                             .arg(statVal1)
                             .arg(REG_STAT_2, 4, 16, QChar('0'))
                             .arg(statVal2)
                             .arg(val, 4, 16, QChar('0'));
}

bool FlashDev::read_rdsr(uint16_t *result)
{
    if (!devWriteReg(REG_CMD, CMD_RDSR))
        return {};
    uint16_t val;
    if (!devReadReg(REG_READBACK, &val))
        return {};
    if (result)
        *result = val;
    return true;
}

bool FlashDev::updateFlashIdData()
{
    const int data_size = sizeof(FlashIdData) / sizeof(uint16_t);
    uint32_t addr[data_size];
    const uint32_t data_offset = REG_DATA + flashIdData.getBlockSize() / sizeof(uint16_t)
        - data_size;
    for (uint32_t w = 0; w < data_size; ++w)
        addr[w] = data_offset + w;

    bool blkReadOk = devBlkReadReg(data_size, addr, reinterpret_cast<uint16_t *>(&flashIdData));
    if (!blkReadOk) {
        qCritical(flash) << "Error reading Flash ID";
        return {};
    }

    const auto capacity = flashIdData.getFlashCapacity_Mib();
    qDebug(flash).noquote()
        << QString("%1 Flash capacity: %2Mib (%3MiB)")
               .arg(getPrefix())
               .arg(capacity)
               .arg(capacity / 8);
    return true;
}

uint32_t FlashDev::getFlashSize() const
{
    //                              from Mi-bits to bytes
    return flashIdData.getFlashCapacity_Mib() / 8 * (1024*1024);
}

uint32_t FlashDev::getImageOffset(FlashImageType imageType) const
{
    // duplicate of FlashDevTask::startAddr
    return (imageType == FlashImageType::Golden) ? 0 : getFlashSize() / 2;
}

uint32_t FlashDev::getImageSizeLimit(FlashImageType imageType) const
{
    Q_UNUSED(imageType);
    return getFlashSize() / 2;
}
