#include "HistMngrCfg.h"

#include <QDataStream>
#include <QDebug>

#include "mregdev/qtmregdevice.h"

constexpr const int DOUBLE_PULSE_TIME_RANGE = 100;
constexpr const int TDC_WIDTH_MAX_NS = 20e3;

const static QMap<HistId, HistGroupInfo> histGroupInfoMap =
{
    {HistId::tdcTime, HistGroupInfo(HistId::tdcTime, "tdcTime",
     "TDC Hit Time, ns", 51200, 0, 51200,
     "Histogram fills with hit time from HPTDC part of event")},

    {HistId::tdcDoublePulseTime, HistGroupInfo(HistId::tdcDoublePulseTime, "tdcDoublePulseTime",
     "TDC Double-Pulse Time, ns",
     2*DOUBLE_PULSE_TIME_RANGE*1024/24, -1*DOUBLE_PULSE_TIME_RANGE, DOUBLE_PULSE_TIME_RANGE,
     "Histogram fills with hit time differance relative to each hit from 'StopCh'")},

    // maximum value is <sum of all channels>
    {HistId::tdcCh, HistGroupInfo(HistId::tdcCh, "tdcCh",
     "TDC Hits by channel", 0, 0, 0,
     "Histogram fills with channel number of HPTDC hit")},

    // maximum value is 100*<sum of all channels>
    {HistId::tdcHitEv, HistGroupInfo(HistId::tdcHitEv, "tdcHitEv",
     "TDC Hits per event", 0, 0, 0,
     "Histogram fills with number of HPTDC hits from all channels per event")},

    {HistId::errors, HistGroupInfo(HistId::errors, "errors",
     "TDC Errors bits", 15, 1, 100,
     "Histogram fills with error bits from HPTDC event")},

    {HistId::tdcTaiTime, HistGroupInfo(HistId::tdcTaiTime, "tdcTaiTime",
     "TDC TAI Time, ns", 0, 0, 0,
     "Histogram fills with trigger time atomic interval ")},

    {HistId::mscPlot, HistGroupInfo(HistId::mscPlot, "mscPlot",
     "MSC hist. plot, sec", 0, 0, 0,
     "Histogram fills with MSC histogram plot")},

    {HistId::tdcWidth, HistGroupInfo(HistId::tdcWidth, "tdcWidth",
     "TDC Hit Width, ns",
     TDC_WIDTH_MAX_NS*(1024/24/4), 0, TDC_WIDTH_MAX_NS, // 100ps bins
     "Histogram fills with hit width from HPTDC part of event")},

    {HistId::adcTime, HistGroupInfo(HistId::adcTime, "adcTime",
     "ADC Hit Time, ns", 10240, 0, 10240,
     "Histogram fills with hit time from ADC part of event")},

    {HistId::tdcAdcDt, HistGroupInfo(HistId::tdcAdcDt, "tdcAdcDt",
     "dTime(TDC-ADC), ns", 2048, -1024, 1024,
     "Histogram fills with time differance between last HPTDC hit and last waveform")},

    {HistId::adcCh, HistGroupInfo(HistId::adcCh, "adcCh",
     "ADC Hits by channel", 72, 1, 73,
     "Histogram fills with channel number on every ADC waveform")},

    {HistId::adcHitEv, HistGroupInfo(HistId::adcHitEv, "adcHitEv",
     "ADC Hits per event", 100, 1, 100,
     "Histogram fills with number of separated ADC waveform from all channels per event")},

    {HistId::chSamples, HistGroupInfo(HistId::chSamples, "chSamples",
     "ADC Samples", 65536/4, -32768, 32768,
     "Histogram fills with every sample from ADC waveform")},

    {HistId::chAmpl, HistGroupInfo(HistId::chAmpl, "chAmpl",
     "ADC Amplitude", 65536/4, -32768, 32768,
     "Histogram fills with either maximum or minimum sample of waveform")},

    {HistId::inlCheck, HistGroupInfo(HistId::inlCheck, "inlCheck",
     "INL Check", 2*1024*40, -1024, 1024,
     "Hit time differance relative to all hits in 'StopCh'")},

    {HistId::inl, HistGroupInfo(HistId::inl, "inl",
     "INL", 1024, 0, 1024,
     "Code density distribution or INL correction")}
};

uint qHash(const HistId &key, uint seed) {
    return qHash(static_cast<int>(key), seed);
}

QString getStrKeyForHistId(HistId id)
{
    return histGroupInfoMap.value(id).strKey;
}

HistId getHistIdForStrKey(QString key)
{
    for(auto it=histGroupInfoMap.cbegin(); it!=histGroupInfoMap.cend(); ++it)
        if(it->strKey == key)
            return it.key();
    return HistId::unknown;
}

bool HistGroupCfg::isBaseHist(HistId histGroupId)
{
    return true
            && (histGroupId != HistId::inl)
            && (histGroupId != HistId::inlCheck)
            && (histGroupId != HistId::chSamples)
            && (histGroupId != HistId::chAmpl)
            && (histGroupId != HistId::mscPlot)
            ;
}

HistGroupInfo HistGroupCfg::getHistGroupInfo(HistId id)
{
    return histGroupInfoMap.value(id);
}

void HistMngrCfg::init(const DaqModule &daq_module_, bool is_inl_enabled_)
{
    daq_module = daq_module_;
    is_inl_enabled = is_inl_enabled_;

    addHistGroup(HistId::tdcTime);

    addHistGroup(HistId::tdcDoublePulseTime);
    addHistGroup(HistId::tdcCh);
    addHistGroup(HistId::tdcHitEv);
    addHistGroup(HistId::errors);
    if(daq_module == DaqModule::Type::AfiViewer){
        addHistGroup(HistId::tdcTaiTime);
        addHistGroup(HistId::mscPlot);
        addHistGroup(HistId::tdcWidth);
    }
    if(daq_module == DaqModule::Type::TQDC){
        addHistGroup(HistId::adcTime);
        addHistGroup(HistId::tdcAdcDt);
        addHistGroup(HistId::adcCh);
        addHistGroup(HistId::adcHitEv);
        addHistGroup(HistId::chSamples);
        addHistGroup(HistId::chAmpl);
    }

    if(is_inl_enabled) {
        addHistGroup(HistId::inl);
        addHistGroup(HistId::inlCheck);
    }
}

HistKey HistMngrCfg::get_cur_hist_key() const
{
    auto groupCfgIt = hist_groups.constFind(cur_hist_id);
    if(groupCfgIt == hist_groups.cend())
        return HistKey();
    return HistGroupCfg::isBaseHist(cur_hist_id)
            ? HistKey(cur_hist_id) : HistKey(cur_hist_id, cur_dev_index, cur_ch);
}

HistGroupInfo HistMngrCfg::getHistGroupInfo(HistId id) const
{
    auto ret = HistGroupCfg::getHistGroupInfo(id);
    if(id == HistId::tdcTime) {
        ret.min = mw_min;
        ret.max = mw_max;
        ret.n_bins = (mw_max - mw_min)*1024/24;
    }
    return ret;
}

void HistMngrCfg::makeConfigValid()
{
    auto max = std::max(mw_max, HIST_MATCHWIN_MIN);
    max = std::min(max, HIST_MATCHWIN_MAX);
    auto min = std::max(mw_min, HIST_MATCHWIN_MIN);
    min = std::min(min, HIST_MATCHWIN_MAX);
    if(max < min)
        max = min = (max + min) / 2;
    if(mw_max != max) {
        qWarning("%s mw_max field was changed from %d to %d",
                 __PRETTY_FUNCTION__, mw_max, max);
        mw_max = max;
    }
    if(mw_min != min) {
        qWarning("%s mw_min field was changed from %d to %d",
                 __PRETTY_FUNCTION__, mw_min, min);
        mw_min = min;
    }

    for(auto hId : hist_groups.keys()) {
        if(hId == HistId::unknown) {
            qWarning("%s remove unknown group from config",
                     __PRETTY_FUNCTION__);
            hist_groups.remove(hId);
            continue;
        }
        if(hId != hist_groups[hId].hist_id) {
            qWarning("%s remove %s group from config",
                     __PRETTY_FUNCTION__,
                     getStrKeyForHistId(hId).toStdString().c_str());
            hist_groups.remove(hId);
            continue;
        }
    }
}

void HistMngrCfg::addHistGroup(HistId id)
{
    if(!hist_groups.contains(id))
        hist_groups[id] = HistGroupCfg(id);
}

void HistKey::print_hist_key() const {
    qInfo().noquote().nospace() << "hist_id: " << hist_id;
    qInfo().noquote().nospace() << "index: " << index.getSerialStr();
    qInfo().noquote().nospace() << "channel: " << ch;
}

QDebug operator<<(QDebug dbg, const HistId &r)
{
    dbg << (getStrKeyForHistId(r).prepend("HistId::"));
    return dbg;
}

QDebug operator<<(QDebug dbg, const HistKey &r)
{
    dbg << "HistKey{"
        << "hist_id:" << r.hist_id << ";"
        << "index:" << r.index << ";"
        << "channel: " << r.ch << "}";
    return dbg;
}

QDebug operator<<(QDebug dbg, const HistGroupCfg &c)
{
    dbg << "title:" << histGroupInfoMap.value(c.hist_id).title << "; "
        << "hist_id:" << c.hist_id << "; "
        << "hist_group_en:" << c.hist_group_en << "; "
        << "show_fit:" << c.show_fit;
    return dbg;
}

QDebug operator<<(QDebug dbg, const HistMngrCfg &c)
{
    dbg << "daq_module:" << c.daq_module << "; "
        << "is_inl_enabled:" << c.is_inl_enabled << "; "
        << "cur_hist_id:" << c.cur_hist_id << "; "
        << "dev_index:" << c.cur_dev_index << "; "
        << "dev_ch:" << c.cur_ch << "; "
        << "curEdgeLead:" << c.curEdgeLead << "; "
        << "curEdgeTrail:" << c.curEdgeTrail << "; "
        << "use_stop_ch:" << c.use_stop_ch << "; "
        << "stop_dev_index:" << c.stop_dev_index << "; "
        << "stop_ch:" << c.stop_ch << "; "
        << "stopEdgeLead:" << c.stopEdgeLead << "; "
        << "stopEdgeTrail:" << c.stopEdgeTrail << "; "
        << "inl_hist_code_dens:" << c.inl_hist_code_dens << "; "
        << "global_enabled:" << c.global_enabled << "; "
        << "tai_ts_window:" << c.tai_ts_window << "; "
        << "mw_min:" << c.mw_min << "; "
        << "mw_max:" << c.mw_max << "; "
        << "hitTimeEdgeLead:" << c.hitTimeEdgeLead << "; "
        << "hitTimeEdgeTrail:" << c.hitTimeEdgeTrail << "; "
        << "log_y_axis:" << c.log_y_axis << "; "
        << "hide_zero_bin:" << c.hide_zero_bin << "; "
        << "ampl_hist_with_max:" << c.ampl_hist_with_max << "; "
        << "hist_groups:" << c.hist_groups << "; "
        << "histExportPath:" << c.histExportPath;

    return dbg;
}

QDataStream &operator<<(QDataStream &out, const HistGroupCfg &r)
{
    out
            << r.hist_id
            << r.hist_group_en
            << r.show_fit;
    return out;
}

QDataStream &operator>>(QDataStream &in, HistGroupCfg &r)
{
    in
            >> r.hist_id
            >> r.hist_group_en
            >> r.show_fit;
    return in;
}

QDataStream &operator<<(QDataStream &out, const HistMngrCfg &r)
{
    out
            << r.daq_module
            << r.is_inl_enabled
            << r.cur_hist_id
            << r.cur_dev_index
            << r.cur_ch
            << r.curEdgeLead
            << r.curEdgeTrail
            << r.use_stop_ch
            << r.stop_dev_index
            << r.stop_ch
            << r.stopEdgeLead
            << r.stopEdgeTrail
            << r.inl_hist_code_dens
            << r.global_enabled
            << r.tai_ts_window
            << r.mw_min
            << r.mw_max
            << r.hitTimeEdgeLead
            << r.hitTimeEdgeTrail
            << r.log_y_axis
            << r.hide_zero_bin
            << r.ampl_hist_with_max
            << r.hist_groups
            << r.histExportPath;

    return out;
}

QDataStream &operator>>(QDataStream &in, HistMngrCfg &r)
{
    in
            >> r.daq_module
            >> r.is_inl_enabled
            >> r.cur_hist_id
            >> r.cur_dev_index
            >> r.cur_ch
            >> r.curEdgeLead
            >> r.curEdgeTrail
            >> r.use_stop_ch
            >> r.stop_dev_index
            >> r.stop_ch
            >> r.stopEdgeLead
            >> r.stopEdgeTrail
            >> r.inl_hist_code_dens
            >> r.global_enabled
            >> r.tai_ts_window
            >> r.mw_min
            >> r.mw_max
            >> r.hitTimeEdgeLead
            >> r.hitTimeEdgeTrail
            >> r.log_y_axis
            >> r.hide_zero_bin
            >> r.ampl_hist_with_max
            >> r.hist_groups
            >> r.histExportPath;
    return in;
}
