#include "HistWorker.h"
#include "DeviceData.h"
#include "DeviceHelper.h"
#include "HistMngrCfg.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDebug>
#include <QSemaphore>
#include <QTimer>

namespace {
void shiftHistIfEmpty(Histogramm *hist, double baseTime) {
    if(hist->getEntries())
        return;
    double size = (hist->max - hist->min)/2;
    hist->resize(hist->n_bins, baseTime-size, baseTime+size);
}
}

HistWorker::HistWorker(QSemaphore *_histSem) :
    QObject(nullptr),
    histSem(_histSem),
    refreshHistTimer(new QTimer(this))
{
    qRegisterMetaType<WorkerStat>();
    qRegisterMetaType<InlResults>();
    qRegisterMetaType<DevChMap>();
    connect(refreshHistTimer, &QTimer::timeout, [this](){ refreshCurrentPlot(); });
    refreshHistTimer->start(500);
}

HistWorker::~HistWorker()
{
    destroy_flg = true;
    trashAllData();
}

void HistWorker::requestInlResults(bool allRes) const
{
    if(allRes){
        emit inlResultsUpdated(inlResults, allRes);
        return;
    }
    InlResults inlRes;
    if(!config.is_inl_enabled) return;

    QVector<double> codeDensData(1024);

    QMap<DeviceIndex, DeviceCalibration> devicesCal;
    for(const DeviceIndex &index : tdcChMap.keys()) {
        devicesCal[index] = inlResults.value(index, DeviceCalibration(index));
    }

    for(HistKey hist_key: histogramms.keys()){
        if(hist_key.hist_id == HistId::inl){
            DeviceIndex index = hist_key.index;
            int ch = hist_key.ch;

            const Histogramm &curHist = histogramms[hist_key];
            const int nBins = curHist.n_bins;
            if(nBins != 1024) continue;
            for (int bin = 0; bin < nBins; ++bin) {
                codeDensData[bin] = curHist.bins[bin];
            }
            // inlRes[index][ch] = makeDNLCorrectionTable(codeDensData);
            devicesCal[index].inl[ch] = codeDensData;
            codeDensData.clear();
            codeDensData.resize(1024);
        }
    }

    for(const DeviceIndex &index : tdcChMap.keys()) {
        if(!devicesCal[index].inl.isEmpty()) {
            if(devErrors.contains(index))
                devicesCal[index].errMask = devErrors[index];
            inlRes[index] = devicesCal[index];
        }
    }

    emit inlResultsUpdated(inlRes);
}

void HistWorker::tdcTtvxMapUpdated(TdcTtvxsMapType ttMap)
{
    tdcTtvxsMap = ttMap;
}

void HistWorker::inlConfigUpdated(InlResults inl)
{
    fullDevCorrMap = inl;
    corrFillCurDevMap();
}

void HistWorker::corrFillCurDevMap()
{
    // Fill curDevCorrMap from fullDevCorrMap according to devList
    for(auto devIt=tdcChMap.cbegin(); devIt!=tdcChMap.cend(); ++devIt) {
        auto &dev_cal = curDevCorrMap[devIt.key()];
        for(int ch=0; ch<=devIt->second; ++ch) {
            auto &ch_cal = dev_cal.inl[ch];
            if(!DeviceCalibration::isValidChInlCal(ch_cal)) {
                ch_cal.clear();
                for(auto it=fullDevCorrMap.cbegin(); it!=fullDevCorrMap.cend(); ++it){
                    if(it.key() != devIt.key() && it.key().getSerial() != devIt.key().getSerial())
                        continue;
                    // update ch_cal with whole match of DeviceIndex or with Serial match
                    // (DEVICE_ID_AnyDevice fix for vmedaq cal-files)
                    if(it->hasValidChInlCal(ch)) {
                        ch_cal = it.value().inl.value(ch);
                        if(it.key() == devIt.key())
                            break;
                    }
                }

                if(ch_cal.isEmpty())
                    ch_cal = QVector<double>(1024, 0);
            }
        }
    }
}

void HistWorker::clearAll()
{
    curDevCorrMap.clear();
    tdcChMap.clear();
    adcChMap.clear();

    stat.clear();
    for(auto key : histogramms.keys()) {
        if(HistGroupCfg::isBaseHist(key.hist_id)) {
            if(key.hist_id == HistId::tdcCh)
                histogramms[key].resize(0,0,0);
            if(key.hist_id == HistId::adcCh)
                histogramms[key].resize(0,0,0);
            histogramms[key].reset();
        } else {
            histogramms.remove(key);
        }
    }

    clearTaiHist(HistId::tdcTaiTime);
    clearTaiHist(HistId::mscPlot);
    inlResults.clear();
}

void HistWorker::clearHist(const HistKey &hist_key)
{
    for(auto it=histogramms.begin(); it!=histogramms.end(); ++it) {
        if(!hist_key.isEmpty() && !(it.key() == hist_key))
            continue;
        if(it.key().hist_id == HistId::tdcTaiTime
                || it.key().hist_id == HistId::mscPlot)
            clearTaiHist(it.key().hist_id);
        else
            it->reset();
    }

    refreshCurrentPlot(true);
    inlResults.clear();
}

void HistWorker::handle_data()
{
    emit workerStatUpdated(stat);
    data_processing = true;
    while(!data_list.isEmpty()) {
        if(destroy_flg || !config.global_enabled) {
            trashAllData();
            break;
        }
        DecodedData data = data_list.takeFirst();
        histSem->release();

        int tdcHits;
        int adcHits;
        int ev_cnt=0;
        auto evIt = data.events.begin();
        const auto checkExternSignals = [&ev_cnt]() {
            if(++ev_cnt>1000) {
                ev_cnt = 0;
                // for quick handle clean & destroy signals
                QCoreApplication::processEvents(QEventLoop::AllEvents, 10);
            }
        };

        while(evIt != data.events.end()) {
            CompleteEvent &ev = evIt.value();
            checkExternSignals();
            if(destroy_flg)
                break;
            tdcHits=0;
            adcHits=0;
            dataProcessingSingleEv(ev, tdcHits, adcHits);

            auto tdcHitEvIt = histogramms.find(HistKey(HistId::tdcHitEv));
            if(tdcHitEvIt != histogramms.end()) {
                if(isHistGroupEn(HistId::tdcHitEv)) {
                    tdcHitEvIt->fill(tdcHits);
                }
            }
            auto adcHitEvIt = histogramms.find(HistKey(HistId::adcHitEv));
            if(adcHitEvIt != histogramms.end()) {
                if(isHistGroupEn(HistId::adcHitEv)) {
                    adcHitEvIt->fill(adcHits);
                }
            }
            ++evIt;
        }

        for(auto mscIt=data.mscData.cbegin(); mscIt!=data.mscData.cend(); ++mscIt) {
            checkExternSignals();
            if(destroy_flg)
                break;
            dataProcessingMsc(mscIt.key(), mscIt.value());
        }
    }
    if(!destroy_flg){
        refreshCurrentPlot();
    }
    data_processing = false;
}

void HistWorker::dataUpdated(const DecodedData &data)
{
    if(destroy_flg) {
        histSem->release();
        return;
    }

    data_list << data;
    if(data_processing)
        return;

    QElapsedTimer worker_timer;
    worker_timer.start();

    handle_data();

    stat.decodeLoop++;
    stat.workerTimeMs += worker_timer.elapsed();
}

void HistWorker::setHistConfig(HistMngrCfg config_)
{
    if(config.hist_groups == config_.hist_groups &&
            config.tai_ts_window == config_.tai_ts_window &&
            config.mw_min == config_.mw_min &&
            config.mw_max == config_.mw_max) {
        config = config_;
        return;
    }

    // At this moment nBins/min/max can't be changed (no need to resize)
    // for all histograms except tdcTime/tdcCh
//    for(auto hCfgIt=config_.hist_groups.cbegin(); hCfgIt!=config_.hist_groups.cbegin(); ++hCfgIt) {
//        const auto &histId = hCfgIt.key();
//        if(histId == HistId::tdcCh)
//            continue;
//        auto hCfgOldIt=config.hist_groups.constFind(histId);
//        if(hCfgOldIt==config.hist_groups.cend())
//            continue;
//        const auto &oldInfo = config.getHistGroupInfo(histId);
//        const auto &newInfo = config_.getHistGroupInfo(histId);
//        if((newInfo.n_bins == oldInfo.n_bins) &&
//                (newInfo.min == oldInfo.min) &&
//                (newInfo.max == oldInfo.max))
//            continue;
//        for(auto hIt=histogramms.begin(); hIt!=histogramms.end(); ++hIt) {
//            if(hIt.key().hist_id == histId)
//                hIt->resize(newInfo.n_bins, newInfo.min, newInfo.max);
//        }
//    }

    config = config_;

    for(auto hCfg : config.hist_groups) {
        if(hCfg.isBaseHist()) {
            addBaseHist(hCfg.hist_id);
        }
    }
    for(auto histKey : histogramms.keys()) {
        if(!config.hist_groups.contains(histKey.hist_id))
            histogramms.remove(histKey);
    }

//    // Save state
//    auto devChCnt = getDevChCntMap();
//    // Restore state
//    auto chHistIt = histogramms.find(HistKey(HistId::tdcCh));
//    if(chHistIt != histogramms.end()) {
//        int maxChNum = tdcChMap.getMaxChNum();
//        chHistIt->resize(maxChNum, 0, maxChNum);

//        fillDevChHist(devChCnt);
//    }
    for(auto hIt=histogramms.begin(); hIt!=histogramms.end(); ++hIt) {
        auto hId = hIt.key().hist_id;
        if(hId==HistId::tdcTaiTime ||
                hId==HistId::mscPlot) {
            if(hIt->n_bins)
                recreateTaiHist(hId, taiReperNew[hIt.key()]);
        }
    }
    auto tdcTimeIt = histogramms.find(HistKey(HistId::tdcTime));
    if(tdcTimeIt != histogramms.end()) {
        const auto &hInfo = config.getHistGroupInfo(HistId::tdcTime);
        tdcTimeIt->resize(hInfo.n_bins, hInfo.min, hInfo.max);
    }
    emit devListUpdated(tdcChMap, histogramms);
}

void HistWorker::refreshCurrentPlot(bool force)
{
    emit hist_updated(histogramms, force);
}

int HistWorker::updateDevMaxCh(const HistId &hId, const DeviceIndex &index, int ch)
{
    const auto &prevDevChCnt = getDevChCntMap(hId);

    int maxChNum = 0;
    auto &devChMap = (hId==HistId::tdcCh) ? tdcChMap : adcChMap;
    switch (hId) {
    case HistId::tdcCh:
        ch = std::max(ch, getTdcChNum(index)-1); break;
    case HistId::adcCh:
        ch = std::max(ch, getAdcChNum(index)-1); break;
    default:;
    }

    auto myDevIt = devChMap.find(index);
    if(myDevIt == devChMap.end()) {
        for(auto p : devChMap)
            maxChNum = std::max(maxChNum, p.first+p.second+1);
        myDevIt = devChMap.insert(index, {maxChNum, ch});
        maxChNum += ch+1;
    } else {
        Q_ASSERT(myDevIt->second < ch);
        if(myDevIt->second >= ch) {
            // no changes needed
            return myDevIt->first;
        }
        auto chDiff = ch - myDevIt->second;
        auto reperCh = myDevIt->first;
        myDevIt->second = ch;
        maxChNum = std::max(maxChNum, myDevIt->first+myDevIt->second+1);


        // Shift chOffset to subsequent devices
        for(auto i=devChMap.begin(); i!=devChMap.end(); ++i) {
            if(i->first > reperCh) {
                i->first += chDiff;
                maxChNum = std::max(maxChNum, i->first+i->second+1);
            }
        }
    }

    auto chHistIt = histogramms.find(hId);
    if(chHistIt != histogramms.end()) {
        chHistIt->resize(maxChNum, 0, maxChNum);
        fillDevChHist(hId, prevDevChCnt);
    }

    if(hId==HistId::tdcCh) {
        auto hitHistIt = histogramms.find(HistKey(HistId::tdcHitEv));
        if(hitHistIt != histogramms.end()) {
            auto &bins = hitHistIt->bins;
            auto old_bins = bins;
            int maxHit = maxChNum ? maxChNum*100 : 100;
            hitHistIt->resize(maxHit, 0, maxHit);
            old_bins.resize(bins.size());
            bins = old_bins;
            // Remember that hitHistIt->over_flow can be not empty
        }

        for(int c=0; c<=ch; ++c) {
            addDeviceChHist(HistKey(HistId::inl, index, c));
            addDeviceChHist(HistKey(HistId::inlCheck, index, c));
        }

        corrFillCurDevMap();
    }

    emit devListUpdated(tdcChMap, histogramms);
    return myDevIt->first;
}

int HistWorker::getDevChOffset(const HistId &hId, const DeviceIndex &index, int curCh)
{
    const auto &map = hId==HistId::tdcCh ? tdcChMap : adcChMap;
    auto chOffsetIt = map.constFind(index);
    if(chOffsetIt == map.cend() || chOffsetIt->second<curCh) {
        updateDevMaxCh(hId, index, curCh);
        chOffsetIt = map.constFind(index);
        Q_ASSERT(chOffsetIt != map.cend());
    }
    return chOffsetIt->first;
}

void HistWorker::addBaseHist(HistId id)
{
    const auto &key = HistKey(id);
    if(histogramms.contains(key))
        return;

    auto cfgIt = config.hist_groups.constFind(id);
    if(cfgIt == config.hist_groups.cend())
        return;

    const auto &info = config.getHistGroupInfo(id);
    histogramms.insert(key, Histogramm(info.title, info.n_bins, info.min, info.max));
}

void HistWorker::addDeviceChHist(HistKey key)
{
    addDeviceChHist(key.hist_id, key.index, key.ch);
}

void HistWorker::addDeviceChHist(HistId id, const DeviceIndex &index, int ch)
{
    const auto &key = HistKey(id, index, ch);
    if(histogramms.contains(key))
        return;

    auto cfgIt = config.hist_groups.constFind(id);
    if(cfgIt == config.hist_groups.cend())
        return;

    const auto &info = config.getHistGroupInfo(id);
    const QString &title = QString("%1 %2 ch:%3")
            .arg(info.title).arg(index.getIdent()).arg(ch);

    if(id==HistId::mscPlot) {
        histogramms.insert(key, Histogramm(title, config.mscPlotWin*10, 0, config.mscPlotWin));
    } else {
        histogramms.insert(key, Histogramm(title, info.n_bins, info.min, info.max));
    }
}

void HistWorker::dataProcessingSingleEv(CompleteEvent &ev,
                                        int &tdcHitsPerEv,
                                        int &adcHitsPerEv)
{
    curEv = &ev;
    TqdcEvent tqdcStopEvent;

    if(!config.stop_dev_index.isNull() && ev.tqdcEv.contains(config.stop_dev_index)) {
        TqdcEvent tqdcEvent = ev.tqdcEv[config.stop_dev_index];
        if(tqdcEvent.hits.contains(config.stop_ch) &&
                !tqdcEvent.hits[config.stop_ch].tdcHits.isEmpty()) {
            tqdcStopEvent = tqdcEvent;
        }
    }

    TqdcChHits &stopHits = tqdcStopEvent.hits[config.stop_ch];
    auto tqdcEvIt = ev.tqdcEv.begin();
    while(tqdcEvIt != ev.tqdcEv.end()) {
        TqdcEvent &tqdcEv = tqdcEvIt.value();
        dataProcessingTqdc(tqdcEv,
                           tdcHitsPerEv,
                           adcHitsPerEv,
                           tqdcStopEvent,
                           stopHits.tdcHits);
        ++tqdcEvIt;
    }
    curEv = nullptr;
}

void HistWorker::dataProcessingTqdc(TqdcEvent &ev,
                                    int &tdcHitsPerEv,
                                    int &adcHitsPerEv,
                                    TqdcEvent &tqdcStopEvent,
                                    QVector<TdcChHit> &stopHits)
{
    // TQDC processing
    auto tqdcChIt = ev.hits.begin();
    while(tqdcChIt != ev.hits.end()) {
        const auto &ch = tqdcChIt.key();

        auto &tdcHits = tqdcChIt.value().tdcHits;
        if(!tdcHits.isEmpty()) {
            dataProcessingTdc(ev, ch, tdcHits, tdcHitsPerEv);

            if(isHistGroupEn(HistId::inlCheck) &&
               (ev.index != config.stop_dev_index || ch != config.stop_ch))
                fillDoublePulse(HistKey(HistId::tdcDoublePulseTime),
                                tqdcStopEvent,
                                stopHits,
                                ev,
                                tdcHits);
        }

        auto &adcHits = tqdcChIt.value().adcHits;
        if(!adcHits.isEmpty())
            dataProcessingAdc(ev, ch, adcHits, adcHitsPerEv);

        ++tqdcChIt;
    }

//    dataProcessingTai(ev);
    grabStatInfo(ev);
    errorsProcessing(ev);
}

void HistWorker::dataProcessingTdc(TqdcEvent &ev,
                                   const int &ch,
                                   QVector<TdcChHit> &tdcHits,
                                   int &tdcHitsPerEv)
{
    tdcHitsPerEv += tdcHits.size();

    // to updateDevMaxCh(...) even on disabled HistId::tdcCh
    const auto &devChOffset = getDevChOffset(HistId::tdcCh, ev.index, ch);
    if(isHistGroupEn(HistId::tdcCh)) {
        const auto &absCh = ch + devChOffset;
        histogramms[HistKey(HistId::tdcCh)].fill(absCh, tdcHits.size());
    }

    for (auto hitIt = tdcHits.begin(); hitIt != tdcHits.end(); ++hitIt) {
        TdcChHit &tdcHit = *hitIt;

        if(isHistGroupEn(HistId::tdcTime)) {
            if(config.hitTimeEdgeLead && tdcHit.leadingTime) {
                double t = tdcHit.hasLeadTimeInl
                        ? tdcHit.leadTimeInl
                        : tdcHit.getLeadTimeNs(getInl(ev, tdcHit));
                histogramms[HistKey(HistId::tdcTime)].fill(t);
            }
            if(config.hitTimeEdgeTrail && tdcHit.trailingTime) {
                double t = tdcHit.hasTrailTimeInl
                        ? tdcHit.trailTimeInl
                        : tdcHit.getTrailTimeNs(getInl(ev, tdcHit));
                histogramms[HistKey(HistId::tdcTime)].fill(t);
            }
        }

        if(isHistGroupEn(HistId::tdcWidth) &&
           tdcHit.leadingTime &&
           tdcHit.trailingTime)
            histogramms[HistKey(HistId::tdcWidth)]
                .fill(tdcHit.hasWidthTimeInl ?
                      tdcHit.widthTimeInl :
                      tdcHit.getWidthNs(getInl(ev, tdcHit)));

        if(isHistGroupEn(HistId::inl)) {
            auto &h = histogramms[HistKey(HistId::inl, ev.index, ch)];
            if(tdcHit.leadingTime)
                h.fill(tdcHit.leadingTime%1024);
//            if(tdcHit.trailingTime)
//                h.fill(tdcHit.trailingTime%1024);
        }
    }
}

void HistWorker::dataProcessingAdc(TqdcEvent &ev,
                                   const int &ch,
                                   QVector<AdcChHit> &adcHits,
                                   int &adcHitsPerEv)
{
    adcHitsPerEv += adcHits.size();

    Histogramm *adcChHist = isHistGroupEn(HistId::adcCh)
            ? &histogramms[HistKey(HistId::adcCh)]
            : nullptr;
    Histogramm *adcTimeHist = isHistGroupEn(HistId::adcTime)
            ? &histogramms[HistKey(HistId::adcTime)]
            : nullptr;
    Histogramm *adcChSamplesHist = isHistGroupEn(HistId::chSamples)
            ? &histogramms[HistKey(HistId::chSamples, ev.index, ch)]
            : nullptr;
    Histogramm *adcChAmplHist = isHistGroupEn(HistId::chAmpl)
            ? &histogramms[HistKey(HistId::chAmpl, ev.index, ch)]
            : nullptr;

    std::optional<int> devChOffset;
    if(adcChHist) {
        auto chOffsetIt = adcChMap.constFind(ev.index);
        if(chOffsetIt == adcChMap.cend() || chOffsetIt->second<ch) {
            updateDevMaxCh(HistId::adcCh, ev.index, ch);
            chOffsetIt = adcChMap.constFind(ev.index);
            Q_ASSERT(chOffsetIt != adcChMap.cend());
        }
        devChOffset = chOffsetIt->first;
    }

    for(auto hitIt = adcHits.begin(); hitIt != adcHits.end(); ++hitIt) {
        AdcChHit &adcChHit = *hitIt;
        if(adcChHist)
            adcChHist->fill(ch + devChOffset.value());
        if(adcTimeHist)
            adcTimeHist->fill(adcChHit.adcTs*8);

        double ampl =0;
        bool has_any_sample = false;
        for(const auto &sample : adcChHit.wf) {
            if(has_any_sample) {
                ampl = config.ampl_hist_with_max
                    ? std::max(sample, ampl) : std::min(sample, ampl);
            } else {
                has_any_sample = true;
                ampl = sample;
            }
            if(adcChSamplesHist) {
                adcChSamplesHist->fill(sample);
            }
        }
        if(has_any_sample && adcChAmplHist)
            adcChAmplHist->fill(ampl);
    }
}


void HistWorker::dataProcessingTai(const TqdcEvent &ev)
{
    constexpr const auto hId = HistId::tdcTaiTime;
    if(ev.tai.valid)
        return;
    if(!isHistGroupEn(hId))
        return;
    auto hIt = histogramms.find(hId);
    if(hIt == histogramms.end())
        return;

    // TAI processing
    auto &reper = taiReperNew[hId];
    if(hIt->n_bins == 0 || !reper.has_value()) {
        recreateTaiHist(hId, ev.tai);
    }
    auto dt = double(ev.tai.toNs()-reper->toNs());
    dt /= 1e9;
    hIt->fill(dt);
}

void HistWorker::dataProcessingMsc(const DeviceIndex &index, const MscData &ev)
{
    const auto hId = HistId::mscPlot;
    if(!isHistGroupEn(hId))
        return;

    QHash<int, QPair<std::optional<TaiTime>*, Histogramm*>> chHists;
    for(auto hIt = histogramms.begin(); hIt != histogramms.end(); ++hIt) {
        if(hIt.key().hist_id == hId && hIt.key().index == index) {
            chHists[hIt.key().ch] = {&taiReperNew[hIt.key()], &hIt.value()};
        }
    }

    // MSC processing
    for(auto sl : ev.slices) {
        if(!sl.tai.isValid())
            continue;

        for(auto chPair : sl.ch) {
            const auto &ch = chPair.first;
            const auto &count = chPair.second;
            const HistKey hKey{hId,index,ch};
            if(!count)
                continue;

            //TODEL: filter only ch=0;
            if(ch!=2)
                continue;
            //TODEL: filter only ch=0;


            auto myHistIt = chHists.find(ch);
            if(myHistIt==chHists.end()) {
                addDeviceChHist(hKey);
                emit devListUpdated(tdcChMap, histogramms);
                Q_ASSERT(histogramms.contains(hKey));
                myHistIt = chHists.insert(ch, {&taiReperNew[hKey], &histogramms[hKey]});
            }

            auto *tRep = myHistIt->first;
            auto *hPtr = myHistIt->second;

            if(!tRep->has_value())
                recreateTaiHist(hKey, sl.tai);

            Q_ASSERT(tRep->has_value());
            double dt = 1e-9 * double(sl.tai.toNs()-tRep->value().toNs());
    #ifdef QT_DEBUG
            static double dtt = -1;
    //        qInfo() << __PRETTY_FUNCTION__ << dt;
//            if(dt<dtt)
//                qInfo() << __PRETTY_FUNCTION__ << "got previous time" << dt;
            dtt = dt;
    #endif

//            Histogramm *hIt = chPair.first<hCache.size() ? hCache[ch]
//                    : nullptr;
//            if(!hIt) {
//                for(int c=hCache.size(); c<=ch; ++ch) {
//                    qInfo() << __PRETTY_FUNCTION__ << "add msc hist for ch:" << c;
//                    HistKey hKey = {hId, index, c};
//                    addDeviceChHist(hId, index, c);
//                    Q_ASSERT(histogramms.contains(hKey));
//                    hCache.push_back(&histogramms[hKey]);
//                }
//                hIt = hCache[chPair.first];
//            }

//            if(hPtr->n_bins != config.mscPlotWin*10)
//                qInfo() << __PRETTY_FUNCTION__ << "hIt->n_bins:" << hPtr->n_bins;
            if(dt > hPtr->max) {
                recreateTaiHist(hKey, sl.tai);
                dt = 1e-9 * double(sl.tai.toNs()-tRep->value().toNs());
                if(false) {
                    auto shift = dt - hPtr->max;
                    uint shiftSec = static_cast<uint>(std::max(static_cast<int>(shift), 1));
                    shiftTaiHist(hKey, shiftSec);
                    dt -= shiftSec;
                }
            }
//            if(hPtr->n_bins != config.mscPlotWin*10)
//                qInfo() << __PRETTY_FUNCTION__ << "hIt->n_bins:" << hPtr->n_bins;

//            static quint64 v1 = 0;
//            if(sl.tai.sec == 0x6207a798) {
//                v1 += count;
//            }
//            if(sl.tai.sec == 0x6207a799) {
//                static bool vvv = true;
//                if(vvv)
//                qInfo()  << __PRETTY_FUNCTION__ << "time:0x6207a798 cnt:" << v1;
//                vvv = false;
//            }
            hPtr->fill(dt, count);
            static quint32 binPrintTime = 0;
            if(binPrintTime != sl.tai.sec) {
                binPrintTime = sl.tai.sec;
//                qInfo() << __PRETTY_FUNCTION__ << hPtr->bins.mid(0, 1000);
            }
        }
    }
}

void HistWorker::grabStatInfo(TqdcEvent &ev)
{
    // Grab stat info for INL results
    if(config.is_inl_enabled) {
        for (auto errWord: ev.errors) {
            auto &hptdcMap = inlResults[ev.index].errMask;
            const uint hptdcId = (errWord>>24) & 0xF;
            const quint16 errMask = errWord & 0x3FFF;
            hptdcMap[hptdcId] |= errMask;
        }
        if(!ev.fw_rev.isNull()) {
            inlResults[ev.index].fw_versions.insert(ev.fw_rev.toUInt());
        }
        if(!ev.bmc_rev.isNull()) {
            inlResults[ev.index].bmc_versions.insert(ev.bmc_rev.toUInt());
        }
        auto &minTemp = inlResults[ev.index].minTemp;
        auto &maxTemp = inlResults[ev.index].maxTemp;
        auto &tempStat = inlResults[ev.index].tempStat;
        for (auto temp_id: ev.temp.keys()) {
            const auto &temp = ev.temp[temp_id];
            auto &tempStatOne = tempStat[temp_id];
            tempStatOne.addTemp(temp);

            if(!tempStatOne.initialized)
                continue;
            if(!tempStatOne.temp_valid(temp)) {
                qWarning().noquote() <<
                    QString("[%1] Bad temp: temp_id=%2 mean=%3 temp=%4")
                        .arg(ev.index.getSerialStr())
                        .arg(temp_id)
                        .arg(tempStatOne.mean, 0, 'f', 2)
                        .arg(temp, 0, 'f', 2);
                continue;
            }
            if(minTemp.contains(temp_id))
                minTemp[temp_id] = std::min(minTemp[temp_id], temp);
            else
                minTemp[temp_id] = temp;
            if(maxTemp.contains(temp_id))
                maxTemp[temp_id] = std::max(maxTemp[temp_id], temp);
            else
                maxTemp[temp_id] = temp;
        }
    }
}

void HistWorker::errorsProcessing(TqdcEvent &ev)
{
    // Errors processing
    if(!ev.errors.empty() &&
       histogramms.contains(HistKey(HistId::errors)) &&
       isHistGroupEn(HistId::errors)) {
        auto &histErr = histogramms[HistKey(HistId::errors)];
        std::vector<int> errorsBits;
        errorsBits.resize(16, 0);
        for (auto errWord: ev.errors) {
            errWord &= 0x3FFF;
            for(ulong bit=0; errWord; ++bit){
                if(errWord&1){
                    ++errorsBits[bit];
                }
                errWord = (errWord>>1);
            }
        }
        for(ulong bit=0; bit<16; ++bit){
            if(errorsBits[bit])
                histErr.fill(bit, errorsBits[bit]);
        }
    }
}

void HistWorker::fillDoublePulse(HistKey hist_key,
                                 TqdcEvent &tdcStopEvent,
                                 QVector<TdcChHit> &stopHits,
                                 TqdcEvent &ev,
                                 QVector<TdcChHit> &tdcEvents)
{
    const TaiTime *stopTai = &tdcStopEvent.tai;
    const TaiTime *evTai = &ev.tai;
    TdcTtvxsMapType::const_iterator it;
    // TODO: Move look for stopTai to upper method
    QMap<DeviceIndex, TtvxsEvent>::const_iterator it1;

    if((it=tdcTtvxsMap.constFind(tdcStopEvent.index)) != tdcTtvxsMap.constEnd()) {
        it1 = curEv->ttvxsEv.find(it.value());
        if(it1 != curEv->ttvxsEv.constEnd()){
            stopTai = &it1.value().tai;
        }
    }
    if((it=tdcTtvxsMap.constFind(ev.index)) != tdcTtvxsMap.constEnd()) {
        it1 = curEv->ttvxsEv.find(it.value());
        if(it1 != curEv->ttvxsEv.constEnd()) {
            evTai = &it1.value().tai;
        }
    }

    if(!evTai->valid || !stopTai->valid)
        return;

    auto stopHitIt = stopHits.begin();
    auto &baseHist = histogramms[hist_key];
    while(stopHitIt != stopHits.end()) {
        TdcChHit &stopHit = *stopHitIt;
        for(auto stopEdge : {true, false}) { // lead, trail
            if(!(stopEdge ? config.stopEdgeLead : config.stopEdgeTrail))
                continue;
            if(!(stopEdge ? stopHit.leadingTime : stopHit.trailingTime))
                continue;
            const double stopTimeNs =
                    (stopEdge ? stopHit.hasLeadTimeInl : stopHit.hasTrailTimeInl) ?
                        (stopEdge ? stopHit.leadTimeInl : stopHit.trailTimeInl) :
                        (stopEdge ? stopHit.getLeadTimeNs(getInl(tdcStopEvent, stopHit))
                                  : stopHit.getTrailTimeNs(getInl(tdcStopEvent, stopHit)));

            double dtTai = 0; // in ns
            dtTai += (evTai->sec - stopTai->sec)*1e9;
            dtTai += evTai->ns - stopTai->ns;
            auto tdcHitIt = tdcEvents.begin();
            while(tdcHitIt != tdcEvents.end()) {
                TdcChHit &tdcHit = *tdcHitIt;
                for(auto curEdge : {true, false}) { // lead, trail
                    if(!(curEdge ? config.curEdgeLead : config.curEdgeTrail))
                        continue;
                    if(!(curEdge ? tdcHit.leadingTime : tdcHit.trailingTime))
                        continue;
                    const double hitTimeNs =
                            (curEdge ? tdcHit.hasLeadTimeInl : tdcHit.hasTrailTimeInl) ?
                                (curEdge ? tdcHit.leadTimeInl : tdcHit.trailTimeInl) :
                                (curEdge ? tdcHit.getLeadTimeNs(getInl(ev, tdcHit))
                                         : tdcHit.getTrailTimeNs(getInl(ev, tdcHit)));

                    double dt = dtTai+hitTimeNs-stopTimeNs;

                    shiftHistIfEmpty(&baseHist, dt);
                    baseHist.fill(dt);

                    HistKey inl_check_key = HistKey(HistId::inlCheck, ev.index, tdcHit.ch);
                    auto inlHist = histogramms[inl_check_key];
                    shiftHistIfEmpty(&inlHist, dt);
                    inlHist.fill(dt);
                }

                ++tdcHitIt;
            }
        }
        ++stopHitIt;
    }
}

QVector<double> &HistWorker::getInl(const TqdcEvent &ev, const TdcChHit &tdcEv)
{
    return getInl(ev.index, tdcEv.ch);
}

QVector<double> &HistWorker::getInl(const DeviceIndex &index, uint ch)
{
    return curDevCorrMap[index].inl[ch];
}

void HistWorker::recreateTaiHist(HistKey histKey, std::optional<TaiTime> reper)
{
    auto &reperSec = taiReperNew[histKey];
//    if(reperSec->value_or(TaiTime{}).sec != reper->value_or(0))
//        qInfo() << __PRETTY_FUNCTION__ << reperSec.value_or(0)
//                << "->" << reper.value_or(0);
    reperSec = reper;

    const auto winSec = (histKey.hist_id == HistId::tdcTaiTime)
            ? config.tai_ts_window : config.mscPlotWin;
    for(auto hIt = histogramms.begin(); hIt != histogramms.end(); ++hIt) {
        if(hIt.key() != histKey)
            continue;
        hIt->resize((winSec ? winSec*10 : 0), 0, winSec);
    }

    refreshCurrentPlot(true);
    emit taiReperUpdated(histKey, reperSec.value_or(TaiTime()));
}

void HistWorker::clearTaiHist(HistKey hKey)
{
    taiReperNew[hKey].reset();
    auto hIt = histogramms.find(hKey);
    if(hIt != histogramms.end()) {
        hIt->resize(0, 0, 0);
    }
    emit taiReperUpdated(hKey, TaiTime());
}

void HistWorker::shiftTaiHist(HistKey histKey, uint shift)
{
    auto &rep = taiReperNew[histKey];
    Q_ASSERT(rep.has_value());
    rep.value().sec += shift;
    qInfo() << __PRETTY_FUNCTION__
            << "histKey:" << histKey
            << "shift:" << shift
            << "newReper:" << rep.value_or(TaiTime()).sec;

    auto hIt = histogramms.begin();
    if(hIt != histogramms.end()) {
        int binShift = 1.*shift/hIt->bin_step;
        for(int b=0;b<std::min(binShift,hIt->n_bins); ++b)
            hIt->under_flow += hIt->bins[b];

        hIt->bins = hIt->bins.mid(binShift);
        hIt->bins.resize(hIt->n_bins);
        qInfo() << __PRETTY_FUNCTION__ << "for  ch:" << hIt.key().ch;
//        hIt->resize(0, 0, 0);
    }
    emit taiReperUpdated(histKey, rep.value_or(TaiTime()));
}

QMap<DeviceIndex, QMap<uint, uint> > HistWorker::getDevChCntMap(const HistId &hId)
{
    QMap<DeviceIndex, QMap<uint, uint>> devChCnt;
    auto chHistIt = histogramms.constFind(hId);
    if(chHistIt == histogramms.constEnd())
        return devChCnt;

    int binsCalculated = 0;
    auto &devChMap = (hId==HistId::tdcCh) ? tdcChMap : adcChMap;
    for(auto it=devChMap.cbegin(); it!=devChMap.cend(); ++it) {
        for(int ch=0; ch<=it->second; ++ch) {
            const auto &globCh = static_cast<int>(it->first+ch);
            const auto &chCnt = static_cast<uint>(chHistIt->bins.value(globCh));
            ++binsCalculated;
            if(chCnt)
                devChCnt[it.key()][ch] = chCnt;
        }
    }
    Q_ASSERT(binsCalculated==chHistIt.value().n_bins);
    return devChCnt;
}

void HistWorker::fillDevChHist(const HistId &hId, QMap<DeviceIndex, QMap<uint, uint> > input)
{
    auto hIt = histogramms.find(hId);
    if(hIt == histogramms.end())
        return;

    auto &devChMap = (hId==HistId::tdcCh) ? tdcChMap : adcChMap;
    for(auto inputIt=input.cbegin(); inputIt!=input.cend(); ++inputIt) {
        auto offsetIt = devChMap.constFind(inputIt.key());
        Q_ASSERT(offsetIt != devChMap.cend());
        if(offsetIt == devChMap.cend())
            continue;

        const auto &chOffset = offsetIt->first;
        const auto &devChCnt = inputIt.value();
        for(uint ch : devChCnt.keys()) {
            Q_ASSERT(ch<=offsetIt->second);
            hIt->fill(chOffset+ch, static_cast<int>(devChCnt[ch]));
        }
    }
}

void HistWorker::trashAllData()
{
    histSem->release(data_list.size());
    data_list.clear();
}

bool HistWorker::isHistGroupEn(HistId groupId) const
{
    return config.hist_groups.value(groupId).hist_group_en;
}
