/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceStatusTableViewFancy.h"

#include <QDebug>
#include <QHeaderView>
#include <QGuiApplication>
#include <QMenu>
#include <QMessageBox>

#include "mvc/DeviceEnableTableModel.h"
#include "ModularDeviceStatus/ModularDeviceStatusModel.h"
#include "device-discover/LldpDiscoveryTableModel.h"
#include "m-stream-output/MStreamStatisticTableModel.h"
#include "mvc/MergeColumnsTableModel.h"

DeviceStatusTableViewFancy::DeviceStatusTableViewFancy(QWidget *parent)
    : DeviceStatusTableView(parent)
    , cellPopupMenu(new QMenu(this))
    , headerPopupMenu(new QMenu(this))
{
    setContextMenuPolicy(Qt::CustomContextMenu);
    connect(this, &QTableView::customContextMenuRequested,
            this, &DeviceStatusTableViewFancy::cellPopup);

    horizontalHeader()->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(horizontalHeader(), &QHeaderView::customContextMenuRequested,
            this, &DeviceStatusTableViewFancy::horizontalHeaderPopup);
    setupDecorators();
}

void DeviceStatusTableViewFancy::setControlEnabled(bool en)
{
    controlEnabled = en;
}

void DeviceStatusTableViewFancy::setupDecorators()
{
    setFont({"Monospace"});
    horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    //setSelectionBehavior(QAbstractItemView::SelectRows);
    //setSelectionMode(QAbstractItemView::SingleSelection);
}

void DeviceStatusTableViewFancy::horizontalHeaderPopup(const QPoint &pos)
{
    auto index = indexAt(pos);
    if (! index.isValid())
        return;

    headerPopupMenu->clear();
    auto insertFunc = [=](const QString &name, void (DeviceStatusTableViewFancy::*pt2ConstMember)(), QVariant data = {}) {
        auto a = headerPopupMenu->addAction(name, this, pt2ConstMember);
        if(data.isValid())
            a->setData(data);
    };

    insertFunc("Hide column", &DeviceStatusTableViewFancy::hideColumnAction, index.column());
    insertFunc("Show all columns", &DeviceStatusTableViewFancy::showAllColumnsSetAction);
    insertFunc("Column set default", &DeviceStatusTableViewFancy::setColumnSetAction,
               QVariant::fromValue(getColumnSetDefault()));
    insertFunc("Column set #1", &DeviceStatusTableViewFancy::setColumnSetAction,
               QVariant::fromValue(getColumnSet1()));

    const QAbstractItemModel *prevModel = nullptr;
    for (int i=0; i<horizontalHeader()->count(); i++) {
        auto hidden = isColumnHidden(i);
        const auto &source = mergeModel->mapColumnToSource(i);
        if (!source.isValid())
            continue;
        QString modelName = source.model->objectName();
        if (source.model != prevModel) {
            prevModel = source.model;
            headerPopupMenu->addSection(modelName);
        }

        auto data = horizontalHeader()->model()->headerData(i, Qt::Horizontal);
        auto str = data.toString();
        if (str.isEmpty())
            continue;
        auto action2 = headerPopupMenu->addAction(str, this, &DeviceStatusTableViewFancy::toggleColumnAction);
        action2->setCheckable(true);
        action2->setChecked(!hidden);
        action2->setData(i);
    }
    headerPopupMenu->popup(mapToGlobal(pos));
}

DeviceIndexSet DeviceStatusTableViewFancy::getSelectedDeviceIndexes() const
{
    QSet<int> selectedRows;
    // assume all selectedIndexes belong to proxyModel
    const QAbstractItemModel *model = nullptr;
    const auto &selection = selectedIndexes();
    for (const auto &i : selection) {
        selectedRows.insert(i.row());
        if (!model)
            model = i.model();
        else
            Q_ASSERT(model == i.model());
    }

    DeviceIndexSet deviceIndexes;
    for (auto row : selectedRows) {
        if (!model) break;
        const auto &index = model->index(row, 1);
        Q_ASSERT(index.isValid());
        const auto &deviceIndex = deviceIndexOf(index);
        deviceIndexes.insert(deviceIndex);
    }
    return deviceIndexes;
}

void DeviceStatusTableViewFancy::cellPopup(const QPoint &pos)
{
    const auto &index = indexAt(pos);
    const auto &singleIndex = deviceIndexOf(index);
    auto deviceIndexes = getSelectedDeviceIndexes();
    if (deviceIndexes.empty() && !singleIndex.isNull())
        deviceIndexes.insert(singleIndex);
    if (deviceIndexes.empty())
        return;
    auto deviceCount = deviceIndexes.size();

    cellPopupMenu->clear();
    QString pluralS = deviceCount > 1 ? "s" : "";
    QString sectionStr = QString("%1 device%2").arg(deviceCount).arg(pluralS);

    cellPopupMenu->addSection(sectionStr);
    cellPopupMenu->addSeparator();
    auto action1 = cellPopupMenu->addAction("Device status...", this, &DeviceStatusTableViewFancy::actionDeviceStatus);
    action1->setData(QVariant::fromValue(singleIndex));
    cellPopupMenu->addSeparator();
    auto action2 = cellPopupMenu->addAction(QString("Reboot %1 device%2...").arg(deviceCount).arg(pluralS), this, &DeviceStatusTableViewFancy::actionDeviceReboot);
    action2->setData(QVariant::fromValue(deviceIndexes));
    if(!controlEnabled)
        action2->setEnabled(false);
    cellPopupMenu->addSeparator();
    cellPopupMenu->popup(mapToGlobal(pos));
}

void DeviceStatusTableViewFancy::hideColumnAction()
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return;
    int column = action->data().toInt();
    hideColumn(column);
}

void DeviceStatusTableViewFancy::showAllColumnsSetAction()
{
    for(auto c=0; c < model()->columnCount(); ++c)
        showColumn(c);
}

void DeviceStatusTableViewFancy::setColumnSetAction()
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return;
    auto columns = action->data().value<QSet<int>>();
    for(auto c=0; c < model()->columnCount(); ++c) {
        auto f = columns.contains(c) ? &DeviceStatusTableView::showColumn
                                     : &DeviceStatusTableView::hideColumn;
        (*this.*f)(c);
    }
}

void DeviceStatusTableViewFancy::showColumnAction()
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return;
    int column = action->data().toInt();
    showColumn(column);
}

void DeviceStatusTableViewFancy::toggleColumnAction()
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return;
    int column = action->data().toInt();
    if (isColumnHidden(column))
        showColumn(column);
    else
        hideColumn(column);
}

void DeviceStatusTableViewFancy::selectionChanged(const QItemSelection &selected, const QItemSelection &deselected)
{
    DeviceStatusTableView::selectionChanged(selected, deselected);
}

DeviceIndexSet DeviceStatusTableViewFancy::getActionDeviceIndexes() const
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return {};
    auto data = action->data();
    return data.value<DeviceIndexSet>();
}

DeviceIndex DeviceStatusTableViewFancy::getActionDeviceIndex() const
{
    auto action = qobject_cast<QAction *>(sender());
    if (!action) return {};
    auto data = action->data();
    return data.value<DeviceIndex>();
}

QSet<int> DeviceStatusTableViewFancy::getColumnSetDefault() const
{
    const auto &defMap = defaultColumnState();
    QSet<int> set;
    for(auto c=0; c < model()->columnCount(); ++c) {
        if(defMap.value(c, true))
            set << c;
    }
    return set;
}

QSet<int> DeviceStatusTableViewFancy::getColumnSet1() const
{
    return {
        mergeModel->getColumnFromSource<DeviceEnableTableModel>(
                    DeviceEnableTableModel::Columns::COL_ENABLE),
                mergeModel->getColumnFromSource<DeviceEnableTableModel>(
                    DeviceEnableTableModel::Columns::COL_DEVICE_ID),
                mergeModel->getColumnFromSource<DeviceEnableTableModel>(
                    DeviceEnableTableModel::Columns::COL_DEVICE_SERIAL),

                mergeModel->getColumnFromSource<LldpDiscoveryTableModel>(
                    LldpDiscoveryTableModel::Columns::COL_ONLINE),
                mergeModel->getColumnFromSource<LldpDiscoveryTableModel>(
                    LldpDiscoveryTableModel::Columns::COL_SLOT),

                mergeModel->getColumnFromSource<ModularDeviceStatusModel>(
                    ModularDeviceStatusModel::Columns::COL_DSB_FW),
                mergeModel->getColumnFromSource<ModularDeviceStatusModel>(
                    ModularDeviceStatusModel::Columns::COL_DSB_TEMP),
                mergeModel->getColumnFromSource<ModularDeviceStatusModel>(
                    ModularDeviceStatusModel::Columns::COL_TRIG_ACCEPT),
                mergeModel->getColumnFromSource<ModularDeviceStatusModel>(
                    ModularDeviceStatusModel::Columns::COL_TRIG_BLOCK),
                mergeModel->getColumnFromSource<ModularDeviceStatusModel>(
                    ModularDeviceStatusModel::Columns::COL_EV_NUM),

                mergeModel->getColumnFromSource<MStreamStatisticTableModel>(
                    MStreamStatisticTableModel::Columns::COL_RECV_DATA_RATE),
                mergeModel->getColumnFromSource<MStreamStatisticTableModel>(
                    MStreamStatisticTableModel::Columns::COL_EVENT_RATE),
    };
}

void DeviceStatusTableViewFancy::actionDeviceStatus()
{
    const auto &deviceIndex = getActionDeviceIndex();
    if (deviceIndex.isNull())
        return;
    emit showDeviceStatus(deviceIndex);
}

void DeviceStatusTableViewFancy::actionDeviceReboot()
{
    if(!controlEnabled)
        return;

    const auto &deviceIndexes = getActionDeviceIndexes();
    if (deviceIndexes.empty())
        return;

    auto cnt = deviceIndexes.size();
    auto msg = QString("Are you sure to want to reboot %1 device%2?")
            .arg(cnt).arg(cnt>1 ? "s" : "");
    auto b = QMessageBox::warning(this, "Reboot device(s)?", msg,
                                  QMessageBox::Yes | QMessageBox::No,
                                  QMessageBox::No);
    if(b != QMessageBox::Yes)
        return;

    for (const auto &deviceIndex : deviceIndexes)
        emit rebootDevice(deviceIndex);
}
