#include "EvNumChecker.h"

#include <QDebug>
#include <QTimer>
#include <QMetaType>

#include "base/DeviceIndexEnabled.h"
#include "base/DeviceId.h"

namespace {
//#define STRONG_NOT_IN_RANGE_SUPPRESSION
// [18:44:04] [GUI thread] "RunControl #ivan event number 0 not in range 33761..33761 "
// [18:44:07] [GUI thread] "RunControl #ivan event number 0 not in range 35638..35638 "
// with defined STRONG_NOT_IN_RANGE_SUPPRESSION 2nd log msg will be suppresed

const int DEVICE_TIMEOUT = 1000; // ms
const int SUBSYSTEM_TIMEOUT = 2000; // ms
const int CHECK_PERIOD = 3000; // ms
const bool DEBUG_INFO = false;

static const QString k_global("global");
}

EvNumChecker::EvNumChecker(QObject *parent) :
    QObject(parent),
    completed(true),
    checkEnabled(false),
    curCheckIndex(0),
    evNumCheckTimer(new QTimer(this)),
    checkExpireTimer(new QTimer(this))
{
    qRegisterMetaType<EvNumSet> ();
    connect(evNumCheckTimer, &QTimer::timeout, this, &EvNumChecker::startCheck);
    evNumCheckTimer->setSingleShot(true);
    evNumCheckTimer->setInterval(CHECK_PERIOD);

    connect(checkExpireTimer, &QTimer::timeout, this, &EvNumChecker::checkTimeout);
    checkExpireTimer->setSingleShot(true);
}

EvNumChecker::~EvNumChecker()
{
    stopCheck();
}

void EvNumChecker::startCheck()
{
    // Can be run in Local mode only
    stopCheck(false);
    checkEnabled = true;
    if(clientSet.isEmpty() && vmeMap.isEmpty()) return;

    ++curCheckIndex;
    emit getTrigEvNum(curCheckIndex);
    checkExpireTimer->start(DEVICE_TIMEOUT);
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__
                << " checkExpire started for " << DEVICE_TIMEOUT << "ms";
}

void EvNumChecker::set_dev_serials(const DeviceIndexSet &devSet)
{
    if(devSerials == devSet)
        return;

    devSerials = devSet;
    if(checkEnabled) {
        startCheck(); // restart check
    }
}

void EvNumChecker::setDeviceList(const DeviceIndexEnabledMap &new_devices)
{
    static const DeviceIdSet SUPPORTED_DEV_ID{
        DEVICE_ID_ADC64,
        DEVICE_ID_ADC64ECAL,
        DEVICE_ID_ADC64S2,
        DEVICE_ID_ADC64VE,
        DEVICE_ID_ADC64VE_XGE,
        DEVICE_ID_ADC64VE2,
        DEVICE_ID_ADC64WR,
        DEVICE_ID_TDC64VHLE,
        DEVICE_ID_TDC64VLE,
        DEVICE_ID_TDC72VXS,
        DEVICE_ID_TQDC16,
        DEVICE_ID_TQDC16VS_E,
        DEVICE_ID_U40VE_RC,
        DEVICE_ID_UT24VE,
    };
    DeviceIndexSet devSet;
    for(auto v : new_devices.getEnabled()) {
        if(SUPPORTED_DEV_ID.contains(v.getDevId()))
            devSet << v;
    }
    set_dev_serials(devSet);
}

void EvNumChecker::stopCheck(bool clearWarnSet)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__;
    checkEnabled = false;
    completed = true;
    evNumCheckTimer->stop();
    checkExpireTimer->stop();
    if(clearWarnSet)
        warnSet.clear();
}

void EvNumChecker::updateClientMap(const ClientIndexSet &set)
{
    if(set == clientSet)
        return;

    clientSet = set;
    if(checkEnabled)
        startCheck(); // restart check
}

void EvNumChecker::updateVmeMap(const QMap<DeviceIndex, QString> &map)
{
    bool hasChanges = (map.keys() != vmeMap.keys());
    vmeMap = map;
    if(hasChanges && checkEnabled) {
        startCheck(); // restart check
    }
}

void EvNumChecker::onRemoteEvNumCheckRequest(int checkIndex)
{
    if(completed)
        clearWarn(WARN_DOUBLE_REQUEST);
    else
        setWarn(WARN_DOUBLE_REQUEST, "EvNumChecker new request while previous one hasn't been completed");
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__ << "checkIndex:"<<checkIndex;

    curCheckIndex = checkIndex;
    if(devSerials.isEmpty() && clientSet.isEmpty() && vmeMap.isEmpty()) {
        emit completedEvNumCheck(curCheckIndex, EvNumSet());
    } else {
        completed = false;
        evNumbers.clear();
        clientEvNum.clear();
        emit evNumRequest(curCheckIndex);
        emit getTrigEvNum(curCheckIndex, false);
        checkExpireTimer->start(clientSet.isEmpty() ? DEVICE_TIMEOUT : SUBSYSTEM_TIMEOUT);
        if(DEBUG_INFO)
            qInfo() << typeid(*this).name() << __func__
                    << " checkExpire started for " << DEVICE_TIMEOUT << "ms";
    }
}

void EvNumChecker::gotTopLevelEvNum(int checkIndex, quint64 evNum)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__ << " checkIndex:"<<checkIndex
                << "evNum:"<<evNum;
    if(!checkEnabled) return;
    if(curCheckIndex != checkIndex) return;
    if(completed) {
        // 1st stage passed
        completed = false;
        trigEvNum = evNum;
        clearWarn(WARN_MISSING_INITIAL);
        clientEvNum.clear();
        evNumbers.clear();
        emit evNumRequest(curCheckIndex);
        checkExpireTimer->start(SUBSYSTEM_TIMEOUT);
    } else {
        // All collected
        performTopLevelCheck(checkIndex, evNum);
    }
}

void EvNumChecker::gotClientEvNum(int checkIndex, const ClientIndex &clientIndex, EvNumSet evNums)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__ << " checkIndex:"<<checkIndex
                << "clientIndex:"<<clientIndex
                << "evNum:"<<evNums;
    if(completed) return;
    if(curCheckIndex != checkIndex) return;
    clientEvNum[clientIndex] = evNums;
    checkAllCollected();
}

void EvNumChecker::gotVmeTrigEvNum(int checkIndex, const DeviceIndex &index, quint64 evNum)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__
                << " checkIndex:" << checkIndex
                << "vmeIndex:" << index
                << "evNum:" << evNum;
    if(completed) return;
    gotEvNum(checkIndex, index, evNum);
}

void EvNumChecker::gotDevEvNum(int checkIndex, const DeviceIndex &index, quint64 evNum)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__
                << " checkIndex:" << checkIndex
                << "index:" << index
                << "evNum:" << evNum;
    if(completed) return;
    gotEvNum(checkIndex, index, evNum);
}

void EvNumChecker::checkTimeout()
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__;
    if(completed) {
        // 1st stage
        setWarn(WARN_MISSING_INITIAL, "Failed to get initial event number from RC module");
    } else {
        QStringList messList;
        if(!checkAllCollected(true)) {
            // 2nd stage
            QStringList clStrList;
            for (auto clIndex : clientSet) {
                if(!clientEvNum.contains(clIndex))
                    clStrList << clIndex.toString();
            }
            if(!clStrList.isEmpty())
                messList << QString("Failed to get event number from client %1")
                                .arg(clStrList.join(' '));

            QStringList vmeStrList;
            for (auto i : vmeMap.keys()) {
                if(!evNumbers.contains(i))
                    vmeStrList << QString("#%1 %2").arg(i.toString()).arg(vmeMap[i]);
            }
            if(!vmeStrList.isEmpty())
                messList << QString("Failed to get event number from vmeTrigModule %1")
                                .arg(vmeStrList.join(' '));

            QStringList devStrList;
            for (auto i : devSerials) {
                if(!evNumbers.contains(i))
                    devStrList << i.getIdent();
            }
            if(!devStrList.isEmpty()) {
                messList << QString("Failed to get event number from devs %1")
                                .arg(devStrList.join(' '));
            }
        } else {
            // 3rd stage
            messList << "Failed to get final event number from RC module";
        }

        const QString mess = messList.join('\n');
        setWarn(WARN_TIMEOUT, mess);
        emit checkFail(mess);
    }

    if(checkEnabled)
        evNumCheckTimer->start();
}

void EvNumChecker::gotEvNum(int checkIndex, const DeviceIndex &index, quint64 evNum)
{
    if(curCheckIndex != checkIndex) return;
    evNumbers[index] = evNum;
    checkAllCollected();
}

bool EvNumChecker::checkAllCollected(bool info_only)
{
    for (auto clIndex : clientSet) {
        if(!clientEvNum.contains(clIndex))
            return false;
    }
    for (auto index : vmeMap.keys()) {
        if(!evNumbers.contains(index))
            return false;
    }
    for (auto index : devSerials) {
        if(!evNumbers.contains(index))
            return false;
    }
    if(info_only)
        return true;

    if(checkEnabled) {
        // Local mode: Get last evNum
        emit getTrigEvNum(curCheckIndex);
        checkExpireTimer->start(DEVICE_TIMEOUT);
    } else {
        // Send all collected events to remote
        EvNumSet list = evNumbers.values().toSet();
        for(auto clList : clientEvNum.values())
            list.unite(clList);
        emit completedEvNumCheck(curCheckIndex, list);
        checkExpireTimer->stop();
        completed = true;
    }
    clearWarn(WARN_TIMEOUT);
    return true;
}

void EvNumChecker::performTopLevelCheck(int checkIndex, quint64 evNum)
{
    bool checkSuccess = true;
    if(DEBUG_INFO){
        qInfo() << typeid(*this).name() << __func__;
        qInfo() << "All collected! range from"<<trigEvNum<<" to "<<evNum;
        qInfo() << "clientEvNum:"<<clientEvNum;
        qInfo() << "vmeEvNum:"<<vmeMap;
        qInfo() << "devSerials:"<<devSerials;
    }

    if(evNum < trigEvNum) {
        setWarn(WARN_EV_NUM_DECR, QString("RC trigEvNum decreased %1 -> %2").arg(trigEvNum).arg(evNum));
        checkSuccess = false;
    } else {
        clearWarn(WARN_EV_NUM_DECR);
    }
    for (auto clIndex : clientEvNum.keys()) {
        bool clientOk = true;
        for (const quint64 &clEvNum : clientEvNum[clIndex]) {
            if (trigEvNum > clEvNum || clEvNum > evNum) {
                setWarn(WARN_NOT_IN_RANGE, clIndex.toString(), QString("%1 has event number %2 not in range %3..%4 ")
                        .arg(clIndex.toString()).arg(clEvNum).arg(trigEvNum).arg(evNum));
                clientOk = false;
            }
        }
        if(clientOk) {
            clearWarn(WARN_NOT_IN_RANGE, clIndex.toString());
        } else {
            checkSuccess = false;
        }
    }
    for (auto index: vmeMap.keys()) {
        if(trigEvNum > evNumbers[index] || evNumbers[index] > evNum) {
            setWarn(WARN_NOT_IN_RANGE, index.toString(), QString("TMWR %1 event number %2 not in range %3..%4")
                    .arg(vmeMap[index]).arg(evNumbers[index]).arg(trigEvNum).arg(evNum));
            checkSuccess = false;
        } else {
            clearWarn(WARN_NOT_IN_RANGE, index.toString());
        }
    }
    Q_UNUSED(checkIndex)
//    if(checkSuccess) {
//        clearWarn(WARN_SUMMARY_ERR);
//    } else {
//        QString mess = QString("EventNumberChecker: %1; CheckIndex=%2 trig %3-%4")
//                .arg(checkSuccess ? "Success" : "Fail")
//                .arg(checkIndex).arg(trigEvNum).arg(evNum);
//        if(!clientEvNum.isEmpty()) {
//            mess.append("; clients:");
//            for (int i : clientEvNum.keys()) {
//                QStringList evStrList;
//                for (quint64 e : clientEvNum[i]) {
//                    evStrList << QString::number(e);
//                }
//                mess += QString(" %1:%2").arg(clientMap[i]).arg(evStrList.join(','));
//            }
//        }
//        if(!vmeMap.isEmpty()) {
//            mess.append("; TMWR:");
//            for (quint64 serial : vmeMap.keys()) {
//                mess += QString(" %1:%2").arg(vmeMap[serial]).arg(vmeMap[serial]);
//            }
//        }
//        if(!warnSet.contains(WARN_SUMMARY_ERR)) {
//            warnSet[WARN_SUMMARY_ERR];
//            qWarning() << mess;
//        }
//        emit checkFail(mess);
//    }
    completed = true;
    checkExpireTimer->stop();
    evNumCheckTimer->start();
}

void EvNumChecker::setWarn(WARN_ID id, const QString &subId, const QString &msg)
{
    Q_ASSERT(!msg.isNull());
    bool printLog = false;
#ifdef STRONG_NOT_IN_RANGE_SUPPRESSION
    if(id == WARN_NOT_IN_RANGE) {
        printLog = !warnSet[id].contains(subId);
        warnSet[id][subId] = msg;
    } else
#endif
    {
        auto &curMsg = warnSet[id][subId];
        if(curMsg != msg) {
            curMsg = msg;
            printLog = true;
        }
    }

    if(printLog) {
        if(!msg.isNull())
            qWarning() << msg;
    }
}

void EvNumChecker::setWarn(EvNumChecker::WARN_ID id, const QString &msg)
{
    setWarn(id, k_global, msg);
}

void EvNumChecker::clearWarn(EvNumChecker::WARN_ID id, const QString &subId)
{
    warnSet[id][subId].clear();
}

void EvNumChecker::clearWarn(WARN_ID id)
{
    clearWarn(id, k_global);
}
